//
//  Copyright (C) by S.Tomabech 1995,1997
//  All rights reserved.
//
//  solve Diophantine eq. x^2+d*y^2=p for prime p and  0< d <=p
//
//   d   (I) : coefficient
//   p   (I) : prime
//   x,y (O) : solution
//   return value   true  : solution exists
//                  false : solution does not exists
//
//  f p  0<d<=p ɂ Diophantine eq. x^2+d*y^2=p
//  ̉΂̉ (x,y) Ԃ,
//
#include "Mint.hpp"
#include "MintFunc.hpp" 
//   d   (I) : W
//   p   (I) : f
//   x,y (O) : 
//   ߂l  :  true  ݂
//              false ݂Ȃ
//

bool Cornacchia(const Mint& d,const Mint& p,Mint& x,Mint& y)
{
//  int k ;
  Mint l,a,b,x0 ;
  Mint c,r ;
  Mint p2=p>>1 ;

//Test_if_residue:

  if( d>p ) return false ;
  else if( d==p ) { x=0 ; y=1 ; return true ; }

  if( Kronecker(-d,p)== -1 ) return false ; // no solution

//Compute_square_root:
  x0=sqrt(-d,p) ;           // x0^2=-d mod p
  if( x0<=p2 ) x0=p-x0 ;    // replace by residue whose absolute value is greater than before
                            // Βl傫]Ɋ

  a=p ;
  b=x0 ;
  l=isqrt(p) ;
  while( b>l ) {
//Euclidean_algorithm:
#if 0
    a=mod(a,b) ;
    Swap(a,b) ;
#else 
    a%=b ;
    Swap(a,b) ;
#endif    
  }

//Test_solution:
  divide(p-b*b,d,c,r) ;
//  if( r!=0 ) return false ;
  if( !r.isZero() ) return false ;
#if 0  
  y=isqrt(c) ;
  if( c!=y*y ) return false ;
#else 
  y=SquareTest(c) ;
  if( sign(y)<0 ) return false ;
#endif  
  x=b ;
  return true ;
}


//  f p  0<d<=4*p ɂ Diophantine eq. x^2+d*y^2=4*p
//  ̉΂̉ (x,y) Ԃ,
//
//   d   (I) : W
//   p   (I) : f
//   x,y (O) : 
//   ߂l  :  true  ݂
//              false ݂Ȃ
//

bool Cornacchia4(Mint& d,Mint& p,Mint& x,Mint& y)
{
  unsigned long rem ;
  Mint x0,l,a,b,c,r,p4 ;
//  Fint t ;

  p4=p<<2 ;                   // p4=4*p
  if( d>p4 ) return false ;
  else if( d==p4 ) { x=0 ; y=1 ; return true ; }

  if( p==2 ) {
    if( d>=8 ) return false ;
    if( d==1 ) {                // 4+1*4=4*2
      x=2 ; y=2 ; return true ;
    }
    else if( d==2 ) {          // 0+2*4=4*2
      x=0 ; y=2 ; return true ;
    }
    else if( d==4 ) {          // 4+4*1=4*2
     x=2 ; y=1 ; return true ;
    }
    else if( d==7 ) {           // 1+7*1=4*2
      x=1 ; y=1 ; return true ;
    }
    else return false ;
  }

  rem=long(d)&3 ;
          // (4|4*p 4|d => 4|x^2) => (x/2)^2+(d/4)*y^2=p
  if( rem==0 ) {
    if( Cornacchia(d>>2,p,x,y) ) {
      x<<=1 ;
      return true ;
    }
    else return false ;
  }

  rem=long(d)&7 ;     // mod 8

  if( rem==3 ) ;
  else {            // d=1,2,5,6,7 mod 8
                    //
                    // ((d=2,6 mod 8 => d=even) => x even) => y even
                    // ( 4|d łȂƂɒ)
                    //   (x/2)^2+d*(y/2)^2=p
                    //
                    // ʂ x^2+d*y^2=4 mod 8
                    // x odd  y even Ȃ NG
                    // x even y odd Ȃ d=4 mod 8  NG
                    // x odd  y odd Ȃ 1+d=4 mod 8 Ȃ킿 d=3 mod 8
                    //  d=1,5,7 mod 8 ̏ꍇ x,y Ƃ even
    if( Cornacchia(d,p,x,y ) ) {
      x<<=1 ; y<<=1 ;
      return true ;
    }
    else return false ;

  }

  if( Kronecker(-d,p)==-1 ) return false  ;

  x0=sqrt(-d,p) ;
  if( long(x0+d)&1 ) x0=p-x0 ;
  a=p<<1 ;
  b=x0 ;
  l=isqrt(p4) ;

  while( b>l ) {
//Euclidean_algorithm:
#if 0
    a=mod(a,b) ;
    Swap(a,b) ;
#else 
    a%=b ;
    Swap(a,b) ;
#endif    
  }

  divide(p4-b*b,d,c,r) ;
//  if( r!=0 ) return false ;
  if( !r.isZero() ) return false ;
#if 0  
  y=isqrt(c) ;
  if( y*y!=c ) return false ;
#else 
  y=SquareTest(c) ;
  if( sign(y)<0 ) return false ;  
#endif  
  x=b ;
  return true ;

}
