//
//  Copyright (C) by S.Tomabechi 1993
//  All rights reserved.
//
//  calculate Euler's constant by Euler-Mclaurin method
//  Euler's constant is defined as
//               n-1
//   \gamma=lim(\sum(1/k-log(n))  n -> \infty
//               k=1
//
//    \gamma=0.57721566490...
//
//
#include <fstream.h>
#include <iostream.h>
#include <iomanip.h>
#include "Fint.hpp"
#include "Error.hpp"

Fint Euler(int precision)
{
  precision=GetPrecision()>precision ? GetPrecision():precision ;

  unsigned long j ;
  long sign(-1) ;
  unsigned long nBn ;
                                      // table of Bernoulli numbers
  static char filename[]="./bernoull.tbl" ;
  ifstream inf(filename,ios::in|ios::binary) ;     // binary file
  if( inf.bad() ) {
    cerr << endl << "Error : could not open file " << filename ;
    Error(file not found) ;
  }

  unsigned long nl(1) ;
  nl<<=10 ;       // n1=2^10

  Mint n(nl) ;
  Mint n2 ;
  long m(0) ;
  long ishift1(10),ishift2(0) ;

  (n2=n)*=n ;           // n2=n*n   2^20

  Fint a(precision+1,precision) ;
  Fint gamma ;
  a=1 ;
  for( gamma=0,j=1 ; j<nl ; j++ ) gamma+=a/j ;

  a=n ;

  gamma-=log(a) ;
  gamma+=1/a/2  ;

  Mint d ;
  Fint B1(0) ;
  Mint num ;
  B1[precision] ;

                                          // read Bernoulli number
  inf.read((char*)&nBn ,sizeof(nBn)) ;    // #Bernoulli numbers
  read(inf,num) ;
  read(inf,d) ;
  for( n=1, j=1 ; j<nBn; j++ ) {
    sign=-sign ;
    read(inf,num) ;                       // numerator of Bernoulli number
    B1=num ;
    read(inf,d) ;                         // denominator of Bernoulli number
    m+=2UL ;
    n*=n2 ;
    while( n.ndigit()>1 ) {               // avoid n becomes large
      n>>=ishift1;
      ishift2+=ishift1 ;
    }

    (d*=m)*= n  ;

    B1/=d ;
    B1>>=ishift2 ;
    if( B1.ndigit()<2 ) break ;
    if( sign>0 ) gamma+=B1 ;
    else gamma-=B1 ;
  }

 inf.close() ;
 return gamma[-(gamma.fdigit()-precision)] ;

}
