//
//  Copyright (C) by S.Tomabechi 1993,2001
//  All rights reserved.
//
//  exponential function
//
//   exp(x)=\sum x^n /n!
//
//   exp(x)=exp([x])*(exp(x-[x])    [x]=integral part of x
//
//   exp(-x)=1/exp(x)
//
//   2001/05/25 change computation for x<0
//
//   2001/06/15 change algorithm
//              using Bibary splitting method
//              reference Xavier Gourdon and Pascal Sebah 's
//                        Binary splitting method
//

#include "Fint.hpp"
#include "FintFunc.hpp"

#if 0     // this is old version
Fint old_exp(const Fint& a)
{
  static int fdigit=-1 ;     // precision of natural base at present
#if 0
#if 0
  if( sign(a)<0 ) {          // a<0,exp(a)=1/exp(-a) ;
    Fint t(a) ;
    return 1/exp(~t) ;
  }
#else
  if( sign(a)<0 ) {          // a<0,exp(a)=1/exp(-a) ;
    Mint ipart=Int(a) ;      // |a| 傫ƎvZs\Ȃ̂,
    Fint fpart(a-ipart) ;
    Fint t(a.fdigit()+1,a.fdigit()) ;
    t=1 ;
//    Fint t(a) ;
//    return 1/exp(~t) ;
    return power(exp(t),ipart)/exp(~fpart) ;
  }
#endif
#endif

#ifdef _MEMMAN_DEBUG
Mint::_MemMan.check(__FILE__,__LINE__) ;
#endif

                             // a>0
  Mint ipart=Int(a) ;        // integral part [a]
  Fint fpart(a-ipart) ;      // fractional part

  Fint t(a.fdigit()+1,a.fdigit()) ;
  t=1 ;
  Fint e(t) ;
  unsigned long n(1) ;

  if( sign(a)<0 ) ~fpart ; // avoid being alternative IɂȂ邱Ƃ
  while( t!=0 ) {            // calculate exp(a-[a])
    t*=fpart ;
    t/=n ;
    e+=t ;
    n++ ;
  }
  if( sign(a)<0 ) e.isInverse(e) ; // P 1/e Ƃق悢mȂ
                                   // exp(-x)=1/exp(x)

  if( fdigit<a.fdigit() ) {   // not enough precision
    Fint::_e=1 ;
    Fint::_e[a.fdigit()-Fint::_e.fdigit()] ;
    t=1 ;
    n=1 ;
    while( t!=0 ) {
      t/=n ;
      Fint::_e+=t ;
      n++ ;
    }
    fdigit=Fint::_e.fdigit() ;      // precision at present
  }

  e*=power(Fint::_e,ipart) ;

#ifdef _MEMMAN_DEBUG
e.check(__FILE__,__LINE__) ;
Mint::_MemMan.check(__FILE__,__LINE__) ;
#endif

  return e[a.fdigit()-e.fdigit()] ;

}

#endif


/* new version */
Fint exp(const Fint& a)
{
  static int fdigit=-1 ;     // precision of natural base at present
  static unsigned long K1=0,K2=0 ;
  static Mint P(1),Q(1) ;

#ifdef _MEMMAN_DEBUG
Mint::_MemMan.check(__FILE__,__LINE__) ;
#endif

                             // a>0
  Mint ipart=Int(a) ;        // integral part [a]
  Fint fpart(a-ipart) ;      // fractional part

  Fint t(a.fdigit()+1,a.fdigit()) ;
  t=1 ;
  Fint e(t) ;
  unsigned long n(1) ;

  if( sign(a)<0 ) ~fpart ; // avoid being alternative IɂȂ邱Ƃ
  while( t!=0 ) {            // calculate exp(a-[a])
    t*=fpart ;
    t/=n ;
    e+=t ;
    n++ ;
  }
  if( sign(a)<0 ) e.isInverse(e) ; // P 1/e Ƃق悢mȂ
                                   // exp(-x)=1/exp(x)

  if( fdigit<a.fdigit() ) {   // not enough precision
    Mint WP ;
    if( K2==0 ) {
      K2=64 ;
      Q=qfactorial(K1,K2) ;
      P=pfactorial(K1,K2) ;
    }
    while( ndigit(P)<=GetPrecision() ) {
      K1=K2 ;
      K2<<=1 ;
      WP=pfactorial(K1,K2);
      Q=Q*WP+qfactorial(K1,K2) ;
      P*=WP ;
    }
    Fint FQ,FP ;
    FQ.allocate(2*ndigit(Q),ndigit(Q)) ;
    FP.allocate(2*ndigit(P),ndigit(P)) ;
    FQ=Fint(Q)+1 ;
    FP=Fint(P) ;
    FQ>>=(32*ndigit(P)) ;
    FP>>=(32*ndigit(P)) ;
    Fint::_e.isDivF(FQ,FP) ;
//  Fint::_e=FQ/FP ;
    fdigit=Fint::_e.fdigit() ;      // precision at present



#if 0
    Fint::_e=1 ;
    Fint::_e[a.fdigit()-Fint::_e.fdigit()] ;
    t=1 ;
    n=1 ;
    while( t!=0 ) {
      t/=n ;
      Fint::_e+=t ;
      n++ ;
    }
    fdigit=Fint::_e.fdigit() ;      // precision at present
#endif
  }

  e*=power(Fint::_e,ipart) ;

#ifdef _MEMMAN_DEBUG
e.check(__FILE__,__LINE__) ;
Mint::_MemMan.check(__FILE__,__LINE__) ;
#endif

  return e[a.fdigit()-e.fdigit()] ;

}
