//
//  Copyright (C) by S.Tomabechi 1994-1997,1998,2000
//  All rights reserved.
//
//  multiplication of integers
//
//  use Mint::_uMtemp and Mint::_vMtemp as temporary variable
//
//  valid for even this=&a or this=&b
//
//  1997/04/25 implement Karatsuba's method
//
//  1997/06/10 implement multiplication by Fast Fourier Transformation
//
//  1997/11/09 a<0,b>0 ̂Ƃ a*=b ̌ʂ a>0 ƂȂC
//
//  1998/10/12 Karatsuba ̕@ŁA𕪊Ƃ̌
//             2 ̙pɂȂ悤ɏCBFFT Lɗpł悤ɂ邽߁B
//
//            ̕@́AʂȂ̂ŒP 1/2 ɂ@ɕύX
//            (2000/01/01)
//
#ifdef __BORLANDC__
#pragma inline
#endif

#ifndef __cplusplus
#error Must use C++ for the type Mint
#endif

#include "Mint.hpp"
#include "FFT.hpp"
#include "IntFuncs.hpp"
#include "Error.hpp"

static long k,l ;

Mint& Mint::isMul(const Mint& a,const Mint& b)
{
#ifdef _MEMMAN_DEBUG
a.check(__FILE__,__LINE__) ;
b.check(__FILE__,__LINE__) ;
#endif
  int bnDigit ;
  int anDigit ;

              // a=0 or b=0 ?
  __asm {
    mov  esi,a
    mov  edi,b
#ifdef __BORLANDC__
    mov  eax,[esi.nDigit]
    mov  ebx,[edi.nDigit]
#else
    mov  eax,[esi+NDIGIT]
    mov  ebx,[edi+NDIGIT]
#endif
    or   eax,eax
    jz   MultiplyZero
  }
Check_b:
   __asm {
    or   ebx,ebx
    jz   MultiplyZero
    cmp  eax,ebx
    jae  not_Swap_a_b
    xchg eax,ebx
    mov  a,edi
    mov  b,esi
  }
not_Swap_a_b:
  __asm {
    mov  anDigit,eax
    mov  bnDigit,ebx
    jmp  Multiply0
  }
MultiplyZero:     // {a|>=|b|
  __asm {
    mov edi,this
#ifdef __BORLANDC__
    mov  dword ptr [edi.nDigit],0
    mov  byte ptr [edi.nDim+3],1
#else
    mov  dword ptr [edi+NDIGIT],0
    mov  byte ptr [edi+SIGN],1
#endif
    jmp  RET1
  }

Multiply0: ;
              // multiplication of large integers by Karatsuba's method
  if( bnDigit>Karatsuba_digit ) {
    long n ;
             // multiplication by Fast Fourier Transformation
//FFT_digit=1<<20 ;
    if( bnDigit>FFT_digit ) {
      if( anDigit/bnDigit<2 ) {    // 傫𑵂
        n=anDigit ;
        k=1 ;
        l=0 ;
        while( n>k ) { k<<=1 ; l++ ; }     // index such that 2^k>=n
                                 // 傫Ƃ͔ɕ

#if 1
//        if( (k-n)>(n-k/2) ) {
        if( 4*k>5*n ) {
//cout << " k " << hex << k << " " << n << dec ;
          n=k>>1 ;
          n-- ;
          goto KaraTsuba ;
        }
#else
//cout << " "<< (double(anDigit)/k <1.0) << " " ;
//cout << endl << " k " << k ;
//cout << "anDigit " << anDigit ;
//cout << " " << double(anDigit)/k ;
        if( double(anDigit)/k<0.8 ) {
            n=k>>1 ;
// cout << " goto Kara " ;
//cout << " n " << n ;
            goto KaraTsuba ;
//          return isToom3(a,b) ;
        }
//cout << " "<< (double(anDigit)/k <1.0) << " " ;
// cout << " goto FFT " ;
#endif
        if( l<NUM_OF_FFT_SCHEME+_FFT_BIT_ ) {
          _fft[l-_FFT_BIT_].fft_mul(a,b,*this) ;
          return *this ;
        }
      }
//      else return isToom3(a,b) ;
    }
#if 0
    else if( bnDigit>Toom3_digit ) {
        return isToom3(a,b) ;
    }
#endif
//  if( bnDigit>=1280 ) return Toom(a,b) ;

    k=n=anDigit ;
    n>>=1 ;

/**********
    __asm {
      mov  eax,n               // eax <- n
      test eax,1               // even ?
      jz   POW2
      inc  eax
    }
POW2:
  __asm {
      mov edx,eax             // save n
      bsr ecx,eax
      xor ebx,ebx
      inc ebx
      shl ebx,cl              // the least power of 2 ebx=2^k s.t. ebx>n

      xor edx,ebx
      jz  SET_n               // n is power of 2
      shl ebx,1
    }
SET_n:
  __asm {
      mov n,ebx
    }
    n>>=1 ;
//    n-- ;
*****************/
KaraTsuba:
/*************
    if( n&1 ) n++ ;
    k=n ;
    l=1 ;
    while( n!=1 ) {
      n>>=1 ;
      l<<=1 ;
    }
    n=l ;
    k^=n ;
    if( k==0 ) ;
    else n<<=1 ;
***********/
                 // ̕ɕ
                 // a=a1*t+a0        t=(2^32)^n
                 // b=b1*t+b0
                 // a*b=p2*t^2+p1*t+p0
                 // p2=a1*b1 p0=a0*b0
                 // p1=(a1+a0)*(b1+b0)-p2-p0
    Mint p2(k,0),
         p1(k,0),
         p0(k,0) ;
    Mint a1(n,0) ;
    __asm {               // ̒
                          // ̊֐ł͌ exact łȂƈȂ̂
                          // ɒ
       mov  esi,a
       mov  ebx,n
#ifdef __BORLANDC__
       mov  edi,[esi]
#else
       mov  edi,[esi]
#endif
       cmp  anDigit,ebx
       jae  SearchZero1
       mov  ebx,anDigit  // case a<t
    }
SearchZero1:
    __asm {
       cmp  dword ptr [ebx*NBYTE+edi-NBYTE],0
       jne  FindNonZero1
       dec  ebx
       jnz  SearchZero1
    }
FindNonZero1:
    __asm {
#ifdef __BORLANDC__
      mov   [esi.nDigit],ebx
#else
      mov   [esi+NDIGIT],ebx
#endif
      mov   edi,dword ptr a1.digit

      mov   esi,a        // 98/01/11
#ifdef __BORLANDC__
      mov   bl,byte ptr [esi.nDim+3]
      mov   byte ptr [a1.nDim+3],bl
      mov   esi,[esi]
#else
      mov   bl,byte ptr [esi+SIGN]
      mov   byte ptr [a1+SIGN],bl
      mov   esi,[esi]
#endif

      mov   ebx,n
      lea   esi,[esi+ebx*NBYTE]
      mov   ecx,anDigit
      sub   ecx,ebx
      ja    NonZero1
#ifdef __BORLANDC__
      mov   dword ptr a1.nDigit,0      // =0
      mov   byte ptr [a1.nDim+3],1
#else
      mov   dword ptr [a1+NDIGIT],0
      mov   byte ptr [a1+SIGN],1
#endif
      jmp   KaraMul1
    }
NonZero1:
     __asm {
#ifdef __BORLANDC__
      mov   dword ptr a1.nDigit,ecx
#else
      mov   dword ptr [a1+NDIGIT],ecx
#endif
      dec   ecx
    }
Shift1:
    __asm {
      mov   eax,[ecx*NBYTE+esi]
      mov   [ecx*NBYTE+edi],eax
      dec   ecx
      jns   Shift1
    }

KaraMul1:
    if( &a==&b ) {
       p2.isMul(a1,a1) ;      // p2=a1^2
       a1.isAdd(a1,a) ;       // a1=a1+a0
       p1.isMul(a1,a1) ;      // (a1+a0)^2
       p0.isMul(a,a) ;
    }
    else {
      Mint b1(n,0) ;
      __asm {
         mov   esi,b
         mov   ebx,n
#ifdef __BORLANDC__
         mov   edi,[esi]
#else
         mov   edi,[esi]
#endif
         cmp   bnDigit,ebx
         jae   SearchZero2
         mov   ebx,bnDigit       // case b<t
       }
SearchZero2:
      __asm {
         cmp   dword ptr [ebx*NBYTE+edi-NBYTE],0
         jne   FindNonZero2
         dec   ebx
         jnz   SearchZero2
      }
FindNonZero2:
      __asm {
#ifdef __BORLANDC__
        mov   [esi.nDigit],ebx
        mov   edi,dword ptr b1.digit
#else
        mov   [esi+NDIGIT],ebx
        mov   edi,dword ptr [b1]
#endif
        mov   esi,b
#ifdef __BORLANDC__
        mov   bl,byte ptr [esi.nDim+3]
        mov   byte ptr [b1.nDim+3],bl
        mov   esi,dword ptr [esi]
#else
        mov   bl,byte ptr [esi+SIGN]
        mov   byte ptr [b1+SIGN],bl
        mov   esi,dword ptr [esi]
#endif

        mov   ebx,n
        lea   esi,[esi+ebx*NBYTE]
        mov   ecx,bnDigit
        sub   ecx,ebx
        ja    NonZero2
#ifdef __BORLANDC__
        mov   b1.nDigit,0
        mov   byte ptr [b1.nDim+3],1
#else
        mov   dword ptr [b1+NDIGIT],0
        mov   byte ptr [b1+SIGN],1
#endif
        jmp   KaraMul2
      }
NonZero2:
      __asm {
#ifdef __BORLANDC__
        mov   dword ptr [b1.nDigit],ecx
#else
        mov   dword ptr [b1+NDIGIT],ecx
#endif
        dec   ecx
      }
Shift2:
    __asm {
        mov   eax,[ecx*NBYTE+esi]
        mov   [ecx*NBYTE+edi],eax
        dec   ecx
        jns   Shift2
      }

KaraMul2:
      p2.isMul(a1,b1) ;
      p1.isMul(a1+=a,b1+=b ) ;
      p0.isMul(a,b) ;
      __asm {
        mov   eax,b
        mov   ecx,bnDigit
#ifdef __BORLANDC__
        mov   [eax.nDigit],ecx     // Ƃ̌ɖ߂
#else
        mov   [eax+NDIGIT],ecx
#endif
      }
    }
    __asm {
       mov   eax,a
       mov   ecx,anDigit
#ifdef __BORLANDC__
       mov   [eax.nDigit],ecx
#else
       mov   [eax+NDIGIT],ecx
#endif
    }

    (p1.isSub(p1,p2)).isSub(p1,p0) ;   // p1=p1-p2-p0

//  Swap(*this,p2[2*n]) ;  // *this <- p2[2*n]
    isShift32(p2,n<<1) ;
    isShiftAdd(*this,p1,n) ;
//  isAdd(*this,p1[n]) ;
#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif
    return isAdd(*this,p0) ;
  }

  int abnDigit(anDigit+bnDigit) ; // ς̍ő包
  int Count ;
  unsigned long* bi ;
  unsigned long* pa ;
  unsigned long* pb ;
  unsigned long Carry ;
  signed char this_sign ;

#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif

  if( this==&a ) {         // copy a
    _uMtemp=a ;
    pa=_uMtemp.digit ;
  }
  else pa=a.digit ;

#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif

  if( this==&b ) {         // copy b
    if( this!=&a ) { _vMtemp=b ; pb=_vMtemp.digit ; }
    else pb=pa ;
  }
  else pb=b.digit ;

#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif

  __asm {
    mov esi,a
#ifdef __BORLANDC__
    mov al,[esi.nDim+3]
    mov esi,b
    imul byte ptr [esi.nDim+3]
#else
    mov al,byte ptr [esi+SIGN]
    mov esi,b
    imul byte ptr [esi+SIGN]
#endif
    mov this_sign,al
    mov esi,this

#ifdef __BORLANDC__
    mov eax,[esi.nDim]
#else
    mov eax,[esi+NDIM]
#endif
    and eax,_DIGITS_
    cmp eax,abnDigit
    ja  SKIP_ALLOC
  }

//  if( ndim()<=abnDigit )
  allocate(abnDigit) ;  // g

SKIP_ALLOC:
#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif

  __asm {
    mov   ebx,pb
    mov   esi,pa
    mov   edi,this
#ifdef __BORLANDC__
    mov   edi,[edi]
#else
    mov   edi,[edi]
#endif

    mov   bi,ebx        // save index

    mov   ecx,anDigit
    xor   eax,eax
  }
clear:                  // for( j=0 ; j<ndigit ; j++ ) digit[j]=0 ;
  __asm {
    mov   [ecx*NBYTE+edi-NBYTE],eax
    dec   ecx
    jnz   clear
  }
OuterLoop:
  __asm {
    mov   ebx,bi
    mov   ecx,[ebx]     // ecx=b.digit[j]
  }
                        // case b.digit[j]!=0
InnerLoop:              // compute b.digit[j]*this
  __asm {
    mov   edx,anDigit
    mov   Count,edx
    xor   ebx,ebx
    mov   Carry,ebx
  }
Multiply:
  __asm {
      mov   eax,[ebx*NBYTE+esi]
      mul   ecx
      add   eax,Carry   //  propagate carry
      adc   edx,0       //  add carry
                        //  overflow is not occured here
      add   [ebx*NBYTE+edi],eax
      adc   edx,0       // オ
      mov   Carry,edx   // save oferflow
//      adc   Carry,edx
      inc   ebx
      dec   Count
      jnz   Multiply
      mov   [ebx*NBYTE+edi],edx
   }
end_InnerLoop:
  __asm {
    add     bi,NBYTE
    lea     edi,[edi+NBYTE]
    dec     bnDigit
    jnz     OuterLoop
   }

CheckDigits:
  __asm {

    mov   edi,this
    mov   al,this_sign
#ifdef __BORLANDC__
    mov   esi,[edi]
    mov   [edi.nDim+3],al
#else
    mov   esi,[edi]
    mov   [edi+SIGN],al
#endif
    mov   ebx,abnDigit
  }
IsZero:                       // search nonzero digit
  __asm {
    cmp   dword ptr [ebx*NBYTE+esi-NBYTE],0
    jne   Set
    dec   ebx
  }
Set:
OK:
  __asm {
#ifdef __BORLANDC__
    mov [edi.nDigit],ebx
#else
    mov [edi+NDIGIT],ebx
#endif
  }

RET1:
#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
check(__FILE__,__LINE__) ;
#endif
  return *this ;

}

#if 0
// ł͂Ȃ
Mint& Mint::Toom(const Mint& a,const Mint& b)
{
    int anDigit=a.nDigit ;
    int bnDigit=b.nDigit ;
//  if( bnDigit<768) return isMul(a,b) ;
    if( bnDigit<120 ) return isMul(a,b) ;
//  if( bnDigit<300 ) return isMul(a,b) ;
    long n=bnDigit/3 ;
    Mint p1(a) ;
    Mint p22(anDigit,0) ;
    Mint p21(anDigit,0) ;
    Mint p20(anDigit,0) ;
    Mint p11(anDigit,0) ;
    Mint p10(anDigit,0) ;
    Mint p00(anDigit,0) ;

    // a.nDigit=n ;
    __asm {
        mov esi,a
        mov ebx,n
#ifdef __BORLANDC__
        mov [esi.nDigit],ebx
#else
        mov [esi+NDIGIT],ebx
#endif
    }
    p1[-n] ;
    Mint p2(p1) ;
    p1.nDigit=n ;
    p2[-n] ;

    if( &a==&b ) {
            p22.isAdd(p2,p1) ;
        p21.Toom(p22,p22) ;
            p22.isAdd(p2,a) ;
        p20.Toom(p22,p22) ;
            p22.isAdd(p1,a) ;
        p10.Toom(p22,p22) ;
        p00.Toom(a,a);
        p11.Toom(p1,p1) ;
        p22.Toom(p2,p2) ;
    }
    else {
        Mint q1(b) ;
        __asm {
            mov esi,b
            mov ebx,n
#ifdef __BORLANDC__
            mov [esi.nDigit],ebx
#else
            mov [esi+NDIGIT],ebx
#endif
        }
        q1[-n] ;
        Mint q2(q1) ;
        q1.nDigit=n ;
        q2[-n] ;
        p22.Toom(p2,q2) ;
        p21.Toom(p20.isAdd(p2,p1),p11.isAdd(q2,q1)) ;   // p20,p11 ̓[NƂĎgp
        p20.Toom(p2+=a,q2+=b) ;
        p11.Toom(p1,q1) ;
        p10.Toom(p1+=a,q1+=b) ;
        p00.Toom(a,b) ;
        __asm {
            mov esi,b
            mov ebx,bnDigit
#ifdef __BORLANDC__
            mov [esi.nDigit],ebx
#else
            mov [esi+NDIGIT],ebx
#endif
        }

    }
#if 0
    *this = p00 ;
    p10.isSub(p10,p11) ;
    p10.isSub(p10,p00) ;
    isShiftAdd(*this,p10,n) ;
//  *this +=p10[n] ;
    p20.isAdd(p20,p11) ;
    p20.isSub(p20,p22) ;
    p20.isSub(p20,p00) ;
    isShiftAdd(*this,p20,n<<1) ;
//  *this +=p20[2*n] ;
    p21.isSub(p21,p22) ;
    p21.isSub(p21,p11) ;
    isShiftAdd(*this,p21,n*3) ;
//  *this +=p21[3*n] ;
//  *this += (p21.isSub(p21,p22).isSub(p21,p11))[3*n] ;
    isShiftAdd(*this,p22,n*4) ;
#else
//    this->operator[](p22,n*4) ;
//  *this=p22 ;
//  (*this)[n*4] ;
#if 0
    isShift32(p22,4*n) ;
    isAdd(*this,p00) ;
#else
    isShiftAdd(p00,p22,4*n) ;
#endif
//  *this = p00 ;

    p10.isSub(p10,p11) ;
    p10.isSub(p10,p00) ;
    isShiftAdd(*this,p10,n) ;
//  *this +=p10[n] ;
    p20.isAdd(p20,p11) ;
    p20.isSub(p20,p22) ;
    p20.isSub(p20,p00) ;
    isShiftAdd(*this,p20,n<<1) ;
//  *this +=p20[2*n] ;
    p21.isSub(p21,p22) ;
    p21.isSub(p21,p11) ;
    isShiftAdd(*this,p21,n*3) ;
//  *this +=p21[3*n] ;
//  *this += (p21.isSub(p21,p22).isSub(p21,p11))[3*n] ;
//  isShiftAdd(*this,p22,n*4) ;
#endif
//  *this +=p22[4*n] ;
    __asm {
        mov esi,a
        mov ebx,anDigit
#ifdef __BORLANDC__
        mov [esi.nDigit],ebx
#else
        mov [esi+NDIGIT],ebx
#endif
    }
    return *this ;
}
#endif


//
// ̐
// deg1 = degree+1
void isKaraMul(int deg1,Mint* p1,int deg2,Mint* p2,Mint* p3)
{
//    int deg3=deg1+deg2 ;
    int deg10,deg11,deg20,deg21 ;
    int j,k ;
    Mint* wp ;
    Mint* p11,*p10,*p21,*p20 ;
    Mint* ApB ;
    Mint* CpD ;
    Mint* ABCD ;
    Mint* A_C ;
    Mint* B_D ;
    Mint* WORK ;
    Mint* p ;
#if 0
cout << endl ;
for( j=deg1-1 ; j>=0 ; j-- ) cout << " " << p1[j] ;
cout <<  " * " << endl ;
for( j=deg2-1 ; j>=0 ; j-- ) cout << " " << p2[j] ;
cout << endl ;
#endif
    if( deg1<deg2 ) {
        wp=p1 ;
        p1=p2 ;
        p2=wp ;
        k=deg1 ;
        deg1=deg2 ;
        deg2=k ;
    }

//  for( j=0 ; j<=deg3 ; j++ ) p3[j].nDigit=0 ;
    if( deg2>20 ) {
        deg10=(deg1+1)>>1 ;
//        deg10=deg10>=deg2 ? deg2-1:deg10 ;
        deg11=deg1-deg10 ;
#if 1
        if( deg10>=deg2 ) {
            isKaraMul(deg10,p1,deg2,p2,p3) ;
            wp=new Mint[deg1+deg2] ;

            isKaraMul(deg11,p1+deg10,deg2,p2,wp) ;
            p3+=deg10 ;
            for( j=0 ; j<=(deg1-deg10)+(deg2-1) ; j++ ) {
                p3->isAdd(*p3,*(wp+j)) ;
                ++p3 ;
//              (p3+j+(deg10))->isAdd(*(p3+j+(deg10)),*(wp+j)) ;
            }
            delete [] wp ;
            return ;
        }
        else {
#endif
            p10=p1 ;
            p11=p1+deg10 ;
            p20=p2 ;
            deg21=deg2-deg10 ;
//            deg21=(deg2>deg10 ? deg2-deg10:0) ;
//            if( deg21!=0 ) p21=p2+deg10 ;
            p21=p2+deg10 ;

            // deg20=deg10 ;
            deg20=deg10 ;
//            deg20=deg2>=deg10 ? deg10:deg2 ;
//          A_C=p3+2*deg10 ;
            WORK=p3+2*deg10 ;
            A_C=new Mint[deg11+deg21] ;
    #if 0
            wp=A_C ;
            for( j=0 ; j<=deg11+deg21-2 ; j++ ) {
                wp->nDigit=0 ;
                wp++ ;
            }
    #endif
            isKaraMul(deg11,p11,deg21,p21,A_C) ;
            for( j=0 ; j<=deg11+deg21-2 ; j++ )
                WORK[j]=A_C[j] ;
#if 0
cout << endl << " A_C " ;
for( j=deg1+deg2-2 ; j>=0 ; j-- ) cout << " " <<    p3[j] ;
#endif
//          B_D=p3 ;
            WORK=p3 ;
            B_D=new Mint[deg10+deg20] ;
    #if 0
            wp=B_D ;
            for( j=0 ; j<=deg10+deg20-2 ; j++ ) {
                wp->nDigit=0 ;
                wp++ ;
//              B_D[j].nDigit=0 ;
            }
    #endif
            isKaraMul(deg10,p10,deg20,p20,B_D) ;
            for( j=0 ; j<=deg10+deg20-2 ; j++ )
                WORK[j]=B_D[j] ;

#if 0
cout << endl << " B_D " ;
for( j=deg1+deg2-2 ; j>=0 ; j-- ) cout << " " <<    p3[j] ;
#endif
            ApB=new Mint[deg10+1] ;
            CpD=new Mint[deg10+1] ;
//          for( j=0 ; j<deg10 ; j++ ) ApB[j].nDigit=0 ;
//          for( j=0 ; j<deg10 ; j++ ) CpD[j].nDigit=0 ;
            for( j=0 ; j<deg11 ; j++ ) {
                (ApB+j)->isAdd(*(p11+j),*(p10+j)) ;
            }
            for( j=deg11 ; j<deg10 ; j++ ) {
                ApB[j]=p10[j] ;
            }
    #if 0
            if( deg21>=0 ) {
                for( j=0 ; j<deg21 ; j++ ) {
                    (CpD+j)->isAdd(*(p21+j),*(p20+j)) ;
                }
                for( j=deg21 ; j<deg20 ; j++ ) {
                    CpD[j]=p20[j] ;
                }
            }
            else {
                k=deg21>0 ? deg21:0 ;
                for( j=0 ; j<k ; j++ ) {
                    (CpD+j)->isAdd(*(p21+j),*(p20+j)) ;
                }
                for( j=k ; j<deg20 ; j++ ) {
                    CpD[j]=p20[j] ;
                }
            }
    #endif
            for( j=0 ; j<deg21 ; j++ ) {
                (CpD+j)->isAdd(*(p21+j),*(p20+j)) ;
            }
            for( j=deg21 ; j<deg20 ; j++ ) {
                CpD[j]=p20[j] ;
            }

            WORK=p3+deg10 ;
            ABCD=new Mint[2*deg10] ;
    #if 0
            wp=ABCD ;
            for( j=0 ; j<2*deg10 ; j++ ) {
                wp->nDigit=0 ;
                wp++ ;
//              ABCD[j].nDigit=0 ;
            }
    #endif

            isKaraMul(deg10,ApB,deg10,CpD,ABCD) ;
#if 0
cout << endl << "ABCD " << flush ;
cout << endl ;
for( j=deg1+deg2-2 ; j>=0 ; j-- ) cout << " " <<    p3[j] ;
#endif
            delete [] ApB ;
            delete [] CpD ;
            for( j=0 ; j<=2*(deg10-1) ; j++ ) {
                (WORK+j)->isAdd(*(ABCD+j),*(WORK+j)) ;
                (WORK+j)->isSub(*(WORK+j),*(B_D+j)) ;
//              (WORK+j)->isSub(*(ABCD+j),*(B_D+j)) ;
            }
#if 0
cout << endl << "- B_D" << (*B_D) ;
for( j=deg1+deg2-1 ; j>=0 ; j-- ) cout << " " <<    p3[j] ;
#endif
//            if( deg21>0 ) {
                for( j=0 ; j<=deg11+deg21-2 ; j++ ) {
                    (WORK+j)->isSub(*(WORK+j),*(A_C+j)) ;
                }
//            }
            delete [] ABCD ;

#if 0
if( deg21>0 ) {
cout << endl << "- A_C"  << (*A_C) << " " ;
for( j=deg1+deg2-2 ; j>=0 ; j-- ) cout << " " <<    p3[j] ;
}
#endif
            delete [] A_C ;
            delete [] B_D ;

            return ;
#if 1
        }
#endif


    }
    for( j=0 ; j<deg1 ; j++,p1++ ) {
        wp=p3+j ;
        p20=p2 ;
        for( k=0 ; k<deg2 ; k++ ) {
            wp->isAdd(*wp,Mint::_Mtemp.isMul(*p1,*p20)) ;
            ++wp ;
            ++p20 ;
        }
    }
#if 0
cout << endl << " = " ;
for( j=deg1+deg2-2 ; j>=0 ; j-- ) cout << " " <<    p3[j] ;
cout << endl ;
#endif
}


void isKaraMul(int deg1,Mint* p1,int deg2,Mint* p2,Mint* p3,Mint& N)
{
//    int deg3=deg1+deg2 ;
    int deg10,deg11,deg20,deg21 ;
    int j,k ;
    Mint* wp ;
    Mint* p11,*p10,*p21,*p20 ;
    Mint* ApB ;
    Mint* CpD ;
    Mint* ABCD ;
    Mint* A_C ;
    Mint* B_D ;
    Mint* WORK ;
    Mint* p ;
#if 0
cout << endl ;
for( j=deg1-1 ; j>=0 ; j-- ) cout << " " << p1[j] ;
cout <<  " * " << endl ;
for( j=deg2-1 ; j>=0 ; j-- ) cout << " " << p2[j] ;
cout << endl ;
#endif
    if( deg1<deg2 ) {
        wp=p1 ;
        p1=p2 ;
        p2=wp ;
        k=deg1 ;
        deg1=deg2 ;
        deg2=k ;
    }

//  for( j=0 ; j<=deg3 ; j++ ) p3[j].nDigit=0 ;
    if( deg2>20 ) {
        deg10=(deg1+1)>>1 ;
//        deg10=deg10>=deg2 ? deg2-1:deg10 ;
        deg11=deg1-deg10 ;
#if 1
        if( deg10>=deg2 ) {
            isKaraMul(deg10,p1,deg2,p2,p3,N) ;
            wp=new Mint[deg1+deg2] ;

            isKaraMul(deg11,p1+deg10,deg2,p2,wp,N) ;
            p3+=deg10 ;
            for( j=0 ; j<=(deg1-deg10)+(deg2-1) ; j++ ) {
                p3->isAdd(*p3,*(wp+j)) ;
//                p3->isAdd(*p3,*wp) ;
                ++p3 ;
//                ++wp ;
//              (p3+j+(deg10))->isAdd(*(p3+j+(deg10)),*(wp+j)) ;
            }
            delete [] wp ;
            return ;
        }
        else {
#endif
            p10=p1 ;
            p11=p1+deg10 ;
            p20=p2 ;
            deg21=deg2-deg10 ;
//            deg21=(deg2>deg10 ? deg2-deg10:0) ;
//            if( deg21!=0 ) p21=p2+deg10 ;
            p21=p2+deg10 ;

            // deg20=deg10 ;
            deg20=deg10 ;
//            deg20=deg2>=deg10 ? deg10:deg2 ;
//          A_C=p3+2*deg10 ;
            WORK=p3+2*deg10 ;
            A_C=new Mint[deg11+deg21] ;
    #if 0
            wp=A_C ;
            for( j=0 ; j<=deg11+deg21-2 ; j++ ) {
                wp->nDigit=0 ;
                wp++ ;
            }
    #endif
            isKaraMul(deg11,p11,deg21,p21,A_C,N) ;
            for( j=0 ; j<=deg11+deg21-2 ; j++ )
                WORK[j]=A_C[j] ;
#if 0
cout << endl << " A_C " ;
for( j=deg1+deg2-2 ; j>=0 ; j-- ) cout << " " <<    p3[j] ;
#endif
//          B_D=p3 ;
            WORK=p3 ;
            B_D=new Mint[deg10+deg20] ;
    #if 0
            wp=B_D ;
            for( j=0 ; j<=deg10+deg20-2 ; j++ ) {
                wp->nDigit=0 ;
                wp++ ;
//              B_D[j].nDigit=0 ;
            }
    #endif
            isKaraMul(deg10,p10,deg20,p20,B_D,N) ;
            for( j=0 ; j<=deg10+deg20-2 ; j++ )
                WORK[j]=B_D[j] ;

#if 0
cout << endl << " B_D " ;
for( j=deg1+deg2-2 ; j>=0 ; j-- ) cout << " " <<    p3[j] ;
#endif
            ApB=new Mint[deg10+1] ;
            CpD=new Mint[deg10+1] ;
//          for( j=0 ; j<deg10 ; j++ ) ApB[j].nDigit=0 ;
//          for( j=0 ; j<deg10 ; j++ ) CpD[j].nDigit=0 ;
            for( j=0 ; j<deg11 ; j++ ) {
                (ApB+j)->isAdd(*(p11+j),*(p10+j)) ;
            }
            for( j=deg11 ; j<deg10 ; j++ ) {
                ApB[j]=p10[j] ;
            }
    #if 0
            if( deg21>=0 ) {
                for( j=0 ; j<deg21 ; j++ ) {
                    (CpD+j)->isAdd(*(p21+j),*(p20+j)) ;
                }
                for( j=deg21 ; j<deg20 ; j++ ) {
                    CpD[j]=p20[j] ;
                }
            }
            else {
                k=deg21>0 ? deg21:0 ;
                for( j=0 ; j<k ; j++ ) {
                    (CpD+j)->isAdd(*(p21+j),*(p20+j)) ;
                }
                for( j=k ; j<deg20 ; j++ ) {
                    CpD[j]=p20[j] ;
                }
            }
    #endif
            for( j=0 ; j<deg21 ; j++ ) {
                (CpD+j)->isAdd(*(p21+j),*(p20+j)) ;
            }
            for( j=deg21 ; j<deg20 ; j++ ) {
                CpD[j]=p20[j] ;
            }

            WORK=p3+deg10 ;
            ABCD=new Mint[2*deg10] ;
    #if 0
            wp=ABCD ;
            for( j=0 ; j<2*deg10 ; j++ ) {
                wp->nDigit=0 ;
                wp++ ;
//              ABCD[j].nDigit=0 ;
            }
    #endif

            isKaraMul(deg10,ApB,deg10,CpD,ABCD,N) ;
#if 0
cout << endl << "ABCD " << flush ;
cout << endl ;
for( j=deg1+deg2-2 ; j>=0 ; j-- ) cout << " " <<    p3[j] ;
#endif
            delete [] ApB ;
            delete [] CpD ;
            for( j=0 ; j<=2*(deg10-1) ; j++ ) {
                (WORK+j)->isAdd(*(ABCD+j),*(WORK+j)) ;
                (WORK+j)->isSub(*(WORK+j),*(B_D+j)) ;
//              (WORK+j)->isSub(*(ABCD+j),*(B_D+j)) ;
            }
#if 0
cout << endl << "- B_D" << (*B_D) ;
for( j=deg1+deg2-1 ; j>=0 ; j-- ) cout << " " <<    p3[j] ;
#endif
//            if( deg21>0 ) {
                for( j=0 ; j<=deg11+deg21-2 ; j++ ) {
                    (WORK+j)->isSub(*(WORK+j),*(A_C+j)) ;
                }
//            }
            delete [] ABCD ;

#if 0
if( deg21>0 ) {
cout << endl << "- A_C"  << (*A_C) << " " ;
for( j=deg1+deg2-2 ; j>=0 ; j-- ) cout << " " <<    p3[j] ;
}
#endif
            delete [] A_C ;
            delete [] B_D ;

            return ;
#if 1
        }
#endif


    }
    for( j=0 ; j<deg1 ; j++,p1++ ) {
        wp=p3+j ;
        p20=p2 ;
        for( k=0 ; k<deg2 ; k++ ) {
            wp->isAdd(*wp,Mint::_Mtemp.isMul(*p1,*p20)) ;
            (*wp)%=N ;
            ++wp ;
            ++p20 ;
        }
    }
#if 0
cout << endl << " = " ;
for( j=deg1+deg2-2 ; j>=0 ; j-- ) cout << " " <<    p3[j] ;
cout << endl ;
#endif
}

#if 0
int main(void)
{
    Mint p1[100] ;
    Mint p2[100] ;
    Mint p3[100] ;
    int deg1,deg2 ;
    int j ;


    for( j=0 ; j<100 ; j++ ) p3[j]=0 ;
    deg1=3 ;
    p1[0]=1 ;
    p1[1]=1 ;
    p1[2]=1 ;
    p1[3]=1 ;
    deg2=3 ;
    p2[0]=1 ;
    p2[1]=1 ;
    p2[2]=1 ;
    p2[3]=1 ;
    deg1=deg2=2 ;
    deg2=1 ;
//  deg2=0 ;
    isKaraMul(deg1+1,p1,deg2+1,p2,p3) ;
    cout << endl ;
    for( j=0 ; j<=deg1 ; j++ ) cout << " " << p1[j] ;
    cout << endl ;
    for( j=0 ; j<=deg2 ; j++ ) cout << " " << p2[j] ;
    cout << endl ;
    for( j=0 ; j<=deg1+deg2 ; j++ ) cout << " " << p3[j] ;

}
#endif
