//
//  Copyright (C) by S.Tomabechi 1993-1997
//  All rights reserved.
//
//  calculate residue a%b
//
//  RemType=0 residue with same signature of a
//  RemType=1 the smallest nonnegative residue  i.e. a-[a/b]*b
//            min { r>0 | r=a-q*b }
//  RemType=2 the smallest residue in absolute value
//
//  ]߂  ( a bŊƂ̗])
//
#ifdef __BORLANDC__
#pragma inline
#endif

#ifndef __cplusplus
#error Must use C++ for the type Mint
#endif

#define _ONLY_MINTDECL_
#include "Mint.hpp"
#include "Error.hpp"

Mint& Mint::isRem(const Mint& a,const Mint& b,int RemType)
{
#ifdef _MEMMAN_DEBUG
a.check(__FILE__,__LINE__) ;
b.check(__FILE__,__LINE__) ;
#endif

  int bnDigit ;

  __asm {
    mov eax,b
#ifdef __BORLANDC__
    mov edx,[eax.nDigit]
#else
    mov edx,[eax+NDIGIT]
#endif
    mov bnDigit,edx
    or  edx,edx                      // b=0 ?
    jnz NOT_ZERO
  }

  Error(zero divide) ;              // zero-divide

  int anDigit ;
  signed char sign ;                // signature of a
  bool bnotSame ;
NOT_ZERO:
  __asm {
    mov eax,a
#ifdef __BORLANDC__
    mov ecx,[eax.nDigit]
    mov dl,[eax.nDim+3]
#else
    mov ecx,[eax+NDIGIT]
    mov dl,[eax+SIGN]
#endif
    mov anDigit,ecx
    mov sign,dl
  }

  if( bnDigit==1 ) {                 // b is single precision
    if( (*digit=a.isRemUL(*b.digit,RemType))==0 )
      nDim&=_POSITIVE_ ;             // signature <- 1
    else if( RemType==SameSignRem )
      (nDim&=_DIGITS_)|=((unsigned long)sign)<<24 ;
    else if( RemType==PositiveRem ) nDim&=_POSITIVE_ ;
    else {
      if( *digit<=(((unsigned long)b)/2) )            // compare absolute values
        nDim&=_POSITIVE_ ;
      else {
        *digit=(*b.digit)-(*digit) ;
        nDim|=_NEGATIVE_ ;
      }
    }
    if( *digit ) nDigit=1 ;
    else nDigit=0 ;
#ifdef _MEMMAN
check(__FILE__,__LINE__) ;
#endif
    return *this ;
  }

  if( anDigit<bnDigit ) {
//cout << endl << "a " << a ;
//cout << endl << "b " << b ;
    if( bnotSame=(this!=&a) ) *this=a ;
    switch( RemType ) {
      case PositiveRem:                           // the least positive residue
        if( sign<0 ) {
          if( b.isPositive() ) *this+=b ;
          else *this-=b ;
        }
        break ;
      case AbsoluteRem:                           // the least absolute residue
        if( bnDigit==anDigit+1 ) {
          _uMtemp=b ;
          if( _uMtemp.isNegative() ) {
#ifdef __BORLANDC__
            __asm {
              mov byte ptr [@Mint@_uMtemp.nDim+3],1   // positive
            }
#else
            _uMtemp.nDim&=_POSITIVE_ ;
#endif
          }
          _vMtemp.isShiftr(_uMtemp,1) ;
#ifdef _MEMMAN_DEBUG
_vMtemp.check(__FILE__,__LINE__) ;
check(__FILE__,__LINE__) ;
#endif
          __asm {
            mov  eax,a
#ifdef __BORLANDC__
            mov  byte ptr [eax.nDim+3],1      // make a positive
#else
            mov  byte ptr [eax+SIGN],1
#endif
          }
          if( _vMtemp<a ) {
            if( sign>0 ) *this-=_uMtemp ;
            else *this+=_uMtemp ;
          }
          else {
            __asm {
            mov cl,sign 
            mov eax,a 
#ifdef __BORLANDC__
            mov  byte ptr [eax.nDim+3],cl
#else
            mov  byte ptr [eax+SIGN],cl
#endif
            }
            break ;
          }
          if( bnotSame ) {
            __asm {                            // reset original signature
               mov  eax,a
               mov  cl,sign
#ifdef __BORLANDC__
               mov  byte ptr [eax.nDim+3],cl
#else
               mov  byte ptr [eax+SIGN],cl
#endif
            }
          }
        }
        break ;
      default:
       break ;
    }
#ifdef _MEMMAN
check(__FILE__,__LINE__) ;
#endif
//cout << endl << "*this " << *this ;
    return *this ;
  }

  int shift(31) ;
  unsigned long q,res ;
  unsigned long v1,v2 ;
  unsigned long j4 ;
  unsigned long j_n4 ;
  int Count ;
  unsigned long *ui,*vi ;

  if( ndim()<anDigit+1 ) {       // area is not sufficient
    if( this==&a || this==&b ) reallocate(anDigit+1) ;
    else allocate(anDigit+1) ;
  }

#ifdef _MEMMAN
_MemMan.check(__FILE__,__LINE__) ;
#endif

  if( _vMtemp.ndim()<bnDigit ) _vMtemp.allocate(bnDigit) ;
#ifndef __BORLANDC__
  unsigned long* vptr ;
    vptr=Mint::_vMtemp.digit ;
#endif

//D1:
  __asm {
#ifdef __BORLANDC__
    mov   esi,dword ptr [@Mint@_vMtemp]
    mov   edi,this
    mov   edi,[edi]
#else
    mov   esi,vptr
    mov   edi,this
    mov   edi,[edi]
#endif

    mov   ui,edi   // save index
    mov   vi,esi

    mov   edi,b
#ifdef __BORLANDC__
    mov   edi,[edi]
#else
    mov   edi,[edi+DIGIT]
#endif
    mov   ebx,bnDigit
    bsr   ecx,[ebx*NBYTE+edi-NBYTE]
    sub   shift,ecx
    mov   ecx,shift
  }
  __asm {
    mov   eax,[edi+ebx*NBYTE-NBYTE]
  }
mult_vd:                      //  _vMtemp=b<<shift
  __asm {
    dec   ebx
    jz    EXIT1
    mov   edx,eax
    mov   eax,[edi+ebx*NBYTE-NBYTE]
    shld  edx,eax,cl
    mov   [esi+ebx*NBYTE],edx
    jmp   mult_vd
  }
EXIT1:
  __asm {
    shl   eax,cl
    mov   [esi],eax

    mov   edi,a
#ifdef __BORLANDC__
    mov   edi,[edi]
#else
    mov   edi,[edi+DIGIT]
#endif

    mov   esi,ui
    mov   ebx,anDigit
    xor   edx,edx
    mov   eax,[edi+ebx*NBYTE-NBYTE]
    shld  edx,eax,cl
    mov   [esi+ebx*NBYTE],edx
  }
mult_ud:                      //   a<<shift
  __asm {
    dec   ebx
    jz    EXIT2
    mov   edx,eax
    mov   eax,[edi+ebx*NBYTE-NBYTE]
    shld  edx,eax,cl
    mov   [esi+ebx*NBYTE],edx
    jmp   mult_ud
  }
EXIT2:
  __asm {
    shl   eax,cl
    mov   [esi],eax

  }
//D2:
  __asm {
    mov   ebx,bnDigit
    mov   edi,vi
    mov   eax,[ebx*NBYTE+edi-NBYTE]
    mov   v1,eax
    mov   eax,[ebx*NBYTE+edi-2*NBYTE]
    mov   v2,eax

    mov   eax,anDigit
    mov   j4,eax
    sub   eax,ebx
    shl   eax,2
    mov   j_n4,eax
  }

D3:
  __asm {
    mov   ebx,j4
    mov   esi,ui
    mov   edx,[ebx*NBYTE+esi] // dx=u[j]
    cmp   edx,v1
    jne   not_equal
    mov   eax,0ffffffffh         // u[j]==v1
    mov   q,eax
    mov   ecx,[ebx*NBYTE+esi-NBYTE]
  }
test0:
  __asm {
    add   ecx,v1
    jc    D4
    mov   res,ecx
    jmp   test1
  }
not_equal:
  __asm {
    mov   eax,[ebx*NBYTE+esi-NBYTE]
    div   v1
    mov   q,eax     // q=quotient
    mov   res,edx   // res=residue
  }
test1:
  __asm {
    mov   eax,v2
    mul   q
    mov   ecx,[ebx*NBYTE+esi-2*NBYTE]
    sub   ecx,eax
    mov   ecx,res
    sbb   ecx,edx
    jnc   D4
  }
q_1:
  __asm {
    dec   q
    jmp   test0
  }
D4:
  __asm {                       // u-q*v
    mov   edi,vi
    mov   ecx,bnDigit
    mov   esi,ui

    add   esi,j_n4
    mov   Count,ecx
    xor   ecx,ecx               // ecx=Overflow
    mov   ebx,ecx               // ebx=carry
  }
mult:
  __asm {
    mov   eax,[edi]
    lea   edi,[edi+NBYTE]      // DO NOT CHANGE carry flag
    mul   q
    add   eax,ecx
    adc   edx,0                 // add carry
    mov   ecx,edx               // save overflow
  }
//next1:
  __asm {
    add   eax,ebx
    adc   ecx,0
    sbb   [esi],eax
    mov   ebx,0
//    adc   ebx,0
    adc   ebx,ebx
    lea   esi,[esi+NBYTE]
    dec   Count
    jnz   mult
    mov   eax,ecx
    mov   ecx,0
    add   eax,ebx
//    adc   ecx,0
    adc   ecx,ecx
    sbb   [esi],eax
    adc   ecx,0
  }
//D5:
  __asm {
    jecxz D7
  }
//D6:                           // q is greater than real quotient
  __asm {                       // add v
    mov   ecx,bnDigit
    mov   edi,vi
    mov   esi,ui
    add   esi,j_n4
    xor   ebx,ebx
  }
Sub_loop:
  __asm {
    mov   eax,[ebx*NBYTE+edi]
    adc   [ebx*NBYTE+esi],eax
    inc   ebx
    dec   ecx
    jnz   Sub_loop
    jnc   D7
    inc   dword ptr [ebx*NBYTE+esi]
  }
D7:
  __asm {
    sub   j_n4,NBYTE
    jc    D8
    dec   j4
    jmp   D3
  }
D8:
  __asm {
    mov   esi,ui
    mov   ebx,bnDigit
    mov   ecx,shift
    or    ecx,ecx
    jnz  Renormal_0
    lea  esi,[esi+ebx*NBYTE-NBYTE]
    jmp  Search_nonZero
  }
Renormal_0:
  __asm {
    mov   eax,[esi]
  }
Renormal:         //   remainder>>shift
  __asm {
    dec   ebx
    jz    EXIT3
    mov   edx,eax
    mov   eax,[esi+NBYTE]
    shrd  edx,eax,cl
    mov   [esi],edx
    lea   esi,[esi+NBYTE]
    jmp   Renormal
  }
EXIT3:
  __asm {
    shr   eax,cl
    mov   [esi],eax

    mov   ebx,bnDigit
  }
Search_nonZero:
  __asm {
    cmp   dword ptr [esi],0
    jne   nonZero                         // 0 ?
    lea   esi,[esi-NBYTE]
    dec   ebx
    jnz   Search_nonZero
    jmp   SetZero
  }
nonZero:
  __asm {
    cmp   RemType,SameSignRem
    je    SetDigit
  }
  __asm {
    mov   esi,ui
    mov   edi,b
#ifdef __BORLANDC__
    mov   edi,[edi]
#else
    mov   edi,[edi]
#endif

    cmp   RemType,PositiveRem    // Is the least postive residue requierd ?
    je    PosRem
    mov   ecx,ebx                // #digits
    mov   ebx,bnDigit            // calculate the least absolute residue
    xor   edx,edx
  }
Compare:                         // comare with |b|/2
  __asm {
    mov   eax,[ebx*NBYTE+edi-NBYTE]  // |b|/2>|r| ?
    shrd  eax,edx,1
    cmp   eax,[ebx*NBYTE+esi-NBYTE]    //
    jb    Less                              // |b|/2<|r|
    ja    Greater                           // |b|/2>|r|
    mov   edx,[ebx*NBYTE+edi-NBYTE]
    dec   ebx
    jnz   Compare
  }
Greater:
  __asm {
    mov   ebx,ecx                           // #digits
    jmp   SetDigit                          // in case |b|/2>=|r|
  }
PosRem:
  __asm {
    cmp   sign,0
    jge   SetDigit
  }
Less:
  __asm {
    mov   ecx,bnDigit
    xor   ebx,ebx
  }
subloop:                                  // calculate |b|-|r|
  __asm {
    mov   eax,[ebx*NBYTE+edi]
    sbb   eax,[ebx*NBYTE+esi]
    mov   [ebx*NBYTE+esi],eax
    inc   ebx
    dec   ecx
    jnz   subloop
    neg   sign                     // negation
    mov   RemType,SameSignRem
    lea   esi,[esi+ebx*NBYTE-NBYTE]
    jmp   Search_nonZero                    // check #digits
  }
SetZero:
  __asm {
    mov   sign,1
    xor   ebx,ebx
  }
SetDigit:
  __asm {
    mov   edi,this

    mov   cl,sign
#ifdef __BORLANDC__
    mov   [edi.nDim+3],cl
    mov   [edi.nDigit],ebx
#else
    mov   [edi+SIGN],cl
    mov   [edi+NDIGIT],ebx
#endif

  }

//RET1:
#ifdef _MEMMAN_DEBUG
check(__FILE__,__LINE__) ;
//_vMtemp.check(__FILE__,__LINE__) ;
//_MemMan.check(__FILE__,__LINE__) ;
#endif
  return *this ;

}

