//
//  Copyright (C) by S.Tomabechi 1999
//  All rights reserved.
//
//  Montgomery   "A block Lanczos algorithm for finding dependencis over GF(2)" 1995
//
//  Selecting S_i and W_i
//
//  Cmt. Inputs : T=V_i^TAV and S_{i-1},where A (and hence T) is symmetric.]
//  Cmt. Outputs : Set S for diagonal of S_iS_i^T, and W_i^inv=S_i(S_i^TTS_i)^-1S_i^T
//  Construct an N*2N block matrix M, with T on the left and I_N on the right.
//  Cmt. Algorithm preforms row opertions on M. It may zero an entire row.
//  Number columns of T as c_1,c_2,...,c_N, with columns in S_{i-1} coming last.
//  Initialize S=empty. Cmt. S has the columns selected from V_i for W_i
//  do j=1 to N
//      do k=j to N while( M[c_j,c_j]=0 )
//           if( M[c_k,c_k]!=0 ) Exchange rows c_j and c_k of M
//      end while
//      if( M[c_j,c_k]!=0 ) then     Cmt. Use columns c_j of V_i in W_i
//           S <- S+{c_j}
//           Divide row c_j of M by M[c_j,c_j]  Cmt. A no-op over GF(2)
//           Add multiples of row c_j to other rows of M, to zero rest of columns c_j
//      else     Cmt. No pivot elements found in column c_j
//                   Cmt. Column c_j ofb T is linear combination of earlier ones Skip it.
//          do k=j to N while( M[c_j,c_j+N]=0 )
//              if( M[c_k,c_j+N]!=0 ) Exchange rows c_j and c_k of M
//          end while
//          assert( M[c_j,c_j+N]!=0 )
//          Add multiples of row c_j to other rows,to zero rest of column c_j+N
//          Zero row c_j of M.   Cmt. Will be zero for rest of algorithm.
//      end if
//   Cmt. Column c_j will remain unchanged for the duration of the algorithm.
//  end do
//  Copy right half of M W_i^inv
//
//

#ifdef __BORLANDC__
#pragma inline
#endif

#include <stdlib.h>
#include "IntFuncs.hpp"
#include "LBit.hpp"
#include "fstream.h"
#include "io.h"
#include "Error.hpp"
#include "Bit.hpp"
#include "TBit.hpp"

#define WordSize 32

unsigned long Select_S_W(unsigned long& S,unsigned long* M1,unsigned long* M2)
{
    long j ;
    unsigned long k,l ;
    unsigned long M1_cj ;
    unsigned long M2_cj ;
    unsigned long c[32] ;
    unsigned long S1(0) ;
    unsigned long iRank(0) ;
    unsigned long hi(WordSize) ;
    unsigned long lo(0) ;

    __asm {
        mov ecx,WordSize
        dec ecx
        mov eax,S
        mov eax,[eax]               // eax <- S
    }

LOOP_0:
        __asm {
            bt  eax,ecx
            jnc _ELSE_
            dec hi
            mov ebx,hi
            mov c[4*ebx],ecx            // set indices corresponding to non-zero bit
            jmp CONT1
        }
_ELSE_:
        __asm {
            mov ebx,lo
            inc lo
            mov c[4*ebx],ecx
        }
CONT1:
    __asm {
        dec ecx
        jns LOOP_0
    }

//cout << endl << "C " ;
//for( j=0 ; j<WordSize ; j++ ) cout << " " <<c[j] ;

    for( j=0 ; j<WordSize ; j++ ) {
        __asm {
            mov esi,j
            mov k,esi

            mov ecx,c[4*esi]                // cj <- c[j]        ecx ͈ȉύXȂ

        }
LOOP_FIND_NON_ZERO1:
        __asm {
            mov edi,k
            cmp edi,WordSize
            jae NOT_FOUND1

            mov ebx,M1                      // search a row with non zero j-th column
            mov edi,c[4*edi]                    // index c_k
            bt  [ebx+4*edi],ecx     // check c_j-th bit
            jnc  NEXT_ROW1
                                        // in case bit is on
            cmp  ecx,edi                    // same index    ?
            je   FOUND1              // do not swap because k=j

            mov ebx,M2               // swap rows
            mov  edx,[ebx+4*ecx]
            xchg edx,[ebx+4*edi]
            mov [ebx+4*ecx],edx

            mov ebx,M1
            mov  edx,[ebx+4*ecx]
            xchg edx,[ebx+4*edi]            // Swap
            mov [ebx+4*ecx],edx

            jmp  FOUND1

        }
NEXT_ROW1:
        __asm {
            inc k
            jmp LOOP_FIND_NON_ZERO1

        }

NOT_FOUND1:
        __asm {
            jmp NO_PIVOT
        }
FOUND1:
        __asm {
            inc  iRank

            bts  S1,ecx

            mov  eax,M2
            mov  edi,[eax+4*ecx]
            mov  M2_cj,edi
            mov  ebx,M1
            mov  edi,[ebx+4*ecx]
            mov  M1_cj,edi

            mov  k,0

        }
LOOP2:
     __asm {
            mov edi,k
            cmp edi,WordSize
            jae CONT
            cmp edi,esi
            je  NEXT1

            mov edi,c[4*edi]

            bt  [ebx+4*edi],ecx
            jnc NEXT1

            mov edx,M1_cj
            xor [ebx+4*edi],edx
            mov edx,M2_cj
            xor [eax+4*edi],edx
        }

NEXT1:
        __asm {
            inc k
            jmp LOOP2

        }

NO_PIVOT:
        __asm {


            mov esi,j                                               //
            mov k,esi
            mov ecx,c[4*esi]     // cj <- c[j]

            mov ebx,M2
            mov eax,[ebx+4*ecx] ;
            mov M2_cj,eax                // M1_cj = cj-th row of M1

        }
LOOP_FIND_NON_ZERO2:
        __asm {
            mov edi,k
            cmp edi,WordSize
            jae NOT_FOUND2

            mov ebx,M2
            mov edi,c[4*edi]                     // index c_k
            bt  [ebx+4*edi],ecx
            jnc  NEXT_ROW2

            cmp  ecx,edi
            je   FOUND2                         // do not swap because k=j


            mov ebx,M1
            mov  edx,[ebx+4*ecx]
            xchg edx,[ebx+4*edi]            // Swap c_j row and c_k row
            mov [ebx+4*ecx],edx

            mov ebx,M2
            mov  edx,[ebx+4*ecx]
            xchg edx,[ebx+4*edi]            // Swap
            mov [ebx+4*ecx],edx

            jmp  FOUND2

        }
NEXT_ROW2:
        __asm {
            inc k
            jmp LOOP_FIND_NON_ZERO2

        }

NOT_FOUND2:
        __asm {
        }
FOUND2:
        __asm {

            mov  ebx,M2                          // copy row
            mov  edx,[ebx+4*ecx]
            mov  M2_cj,edx

            mov  eax,M1                          // copy row
            mov  edx,[eax+4*ecx]
            mov  M1_cj,edx

            mov k,0

        }
LOOP22:
     __asm {
            mov edi,k
            cmp edi,WordSize
            jae CONT


            mov edi,c[4*edi]
            bt  [ebx+4*edi],ecx
            jnc NEXT12

            mov edx,M2_cj
            xor [ebx+4*edi],edx
            mov edx,M1_cj
            xor [eax+4*edi],edx
        }

NEXT12:
        __asm {
            inc k
            jmp LOOP22

        }


CONT:;
#if 0            // for debug
        TBit B ;
        cout << endl ;
        for( k=0 ; k<WordSize ; k++ ) {
            B.SetBitString(WordSize,M1+k) ;
            cout << endl << B << "  " ;
            B.SetBitString(WordSize,M2+k) ;
            cout <<  B ;
        }
#endif //#if 0
    } // for( j=0 ; j<WordSize ; j++ )

    __asm {
        mov eax,S1                                                      // set S_i*S_i^T
        mov ebx,S
        mov [ebx],eax                               // *S <- S1
    }

    return iRank ;

}


#if 0
void main(void)
{
    unsigned long iRank ;
    unsigned long M1[WordSize],M2[WordSize] ;
    unsigned long j,k,l ;
    unsigned long seed=time(NULL) ;
    unsigned long c[WordSize] ;
    cout << endl << "seed " << seed ;
    TBit B1[WordSize],B2[WordSize] ;

    for( j=0 ; j<WordSize ; j++ ) {
        c[j]=j ;
        M1[j]=irandom(seed) ;
        B1[j].SetBitString(WordSize,&M1[j]) ;
        B2[j].SetBitString(WordSize,&M2[j]) ;
        M2[j]=0 ;
        B2[j].BitOn(j) ;
    }

    for( j=0 ; j<WordSize ; j++ ) {
        cout << endl << B1[j] << "  " << B2[j] ;
    }
    iRank=Select_S_W(c,M1,M2) ;

    cout << endl ;
    for( j=0 ; j<WordSize ; j++ ) {
        cout << endl << B1[j] << "  " << B2[j] ;
    }

    cout << endl ;
    for( j=0 ; j<WordSize ; j++ ) {
        cout << " " << c[j] ;
    }
    cout <<endl << "iRank " << iRank ;
}
#endif
