//
//  Copyright (C) by S.Tomabechi 1999
//  All rights reserved.
//
//  block vector class
//

#ifndef _BLOCK_VECTOR_HPP_
#define _BLOCK_VECTOR_HPP_

#include <iostream.h>

#define WordSize    32
#define WordByte    4
#ifndef __BORLANDC__
#define BV_VECTOR   0
#define BV_NDIM     4                       // offset of members
#endif

class SparseMatrix ;
class Lanczos ;

class block_vector
{
    friend class SparseMatrix ;
    friend class Lanczos ;
public:
    unsigned long* v ;
    unsigned long ndim ;

    block_vector() ;                                    // default constructor
    block_vector(const unsigned long _ndim) ;
    block_vector(const block_vector& v) ;               // copy constructor 
    ~block_vector() ;                                   // destructor
    block_vector& operator=(const block_vector& v) ;    // assignment
    block_vector& operator+=(const block_vector& v) ;   // add and assign
    block_vector  operator+(const block_vector& v) ;    // add 
    block_vector& isAdd(const block_vector& a,const block_vector& b) ;
    block_vector& isAND(const block_vector& a,const unsigned long c) ;
    unsigned long operator*(const unsigned long pv) ;
    bool isZero(void) ;
    bool operator==(const block_vector& b) ;
    bool operator!=(const block_vector& b) ;

    void SetRow(const unsigned long irow,const unsigned long elem)
    {
        v[irow]=elem ;
    }
    block_vector operator*(const block_vector& v) ;
    // a * b
    block_vector& isMul(const block_vector& a,const block_vector& b) ;
    // product m * b
    block_vector& isMul(const SparseMatrix& m,const block_vector& b) ;
    // product m^T * b
    block_vector& isMulT(const SparseMatrix& m,const block_vector& b) ;
    // inner product  a^T * b
    block_vector& isIpro(const block_vector& a,const block_vector& b) ;

    friend ::ostream& operator<<(::ostream& os,const block_vector& a) ;
    friend block_vector operator*(const SparseMatrix& s,const block_vector& v) ;
    friend block_vector Ipro(const block_vector& a,const block_vector& b) ;
    friend block_vector inverse(const block_vector& b) ;
    friend void Swap(block_vector& a,block_vector& b) ;

} ;

inline
bool block_vector::operator!=(const block_vector& b)
{
    return !(operator==(b)) ;
}

inline
block_vector& block_vector::operator+=(const block_vector& a)
{
    return isAdd(*this,a) ;
}

inline block_vector::block_vector():ndim(0),v(NULL)
{
}

#endif