//
//  Copyright (C) by S.Tomabechi 1999
//  All rights reserved.
//
//  block Lanczos method
//
//  1999/10/14 D,E,F ߂֐CCɓWJ
//
#include <algorithm>
#include <vector>
#include "SMatrix.hpp"
#include <stdlib.h>
#include "bvector.hpp"
#include "IntFuncs.hpp"
#include "time.h"
#include "TBit.hpp"
#include "Bit.hpp"
#include "Lanczos.hpp"
#include "Error.hpp"
//#include <fstream.h>

#include "CPUTime.hpp"

using namespace std ;

//void Gauss(unsigned long** M,long nrow,long ncol,long BitSize,unsigned long* ipiv) ;
unsigned long Gauss(unsigned long nBit,unsigned long nRow,unsigned long** BZ) ;
unsigned long Select_S_W(unsigned long& mask,unsigned long* M1,unsigned long* M2) ;

unsigned long Gauss(unsigned long nBit1,unsigned long nBit2,
                    unsigned long nRow,unsigned long** Z1,unsigned long** Z2) ;

#ifdef LANCZOS_TRANSPOSE
//////////////////////////////////////////////////////
//
//   A=B^T*B  B^T ɓWJ Lanczos @Kp
//
//

Lanczos::~Lanczos()
{
  if( bdelete ) delete BT ;
}

Lanczos::Lanczos(const unsigned long nrow,const unsigned long ncol):
                 BT(NULL),
                 V0(nrow),V(nrow),V1(nrow),V2(nrow),
                 AV(nrow),
                 Wi(WordSize),Wi1(WordSize),Wi2(WordSize),
                 VAV(WordSize),VAV1(WordSize),
                 VA2V(WordSize),VA2V1(WordSize),
                 D(WordSize),E(WordSize),F(WordSize),
                 VAV2(WordSize),X(nrow),/*Y(nrow),*/I32(WordSize),iter(0),bdelete(true)

#ifdef LANCZOS_DEBUG
                 ,SST(WordSize),SST1(WordSize)
#endif
{
  unsigned long j ;
  TBit B ;
  B.SetBitString(WordSize*WordSize,I32.v) ;

  for( j=0 ; j<WordSize ; j++ ) {    // identity matrix
    I32.v[j]=0 ;
    B.BitOn(j*WordSize+j) ;
  }

  BT=new SparseMatrix(nrow,ncol) ;
}

Lanczos::Lanczos(const SparseMatrix* _BT):
                 BT(_BT),
                 V0(_BT->nrow),V(_BT->nrow),V1(_BT->nrow),V2(_BT->nrow),
                 AV(_BT->nrow),
                 Wi(WordSize),Wi1(WordSize),Wi2(WordSize),
                 VAV(WordSize),VAV1(WordSize),
                 VA2V(WordSize),VA2V1(WordSize),
                 D(WordSize),E(WordSize),F(WordSize),
                 VAV2(WordSize),X(_BT->nrow),/*Y(nrow),*/I32(WordSize),iter(0),bdelete(false)
#ifdef LANCZOS_DEBUG
                 ,SST(WordSize),SST1(WordSize)
#endif
{
  unsigned long j ;
  TBit B ;
  B.SetBitString(WordSize*WordSize,I32.v) ;

  for( j=0 ; j<WordSize ; j++ ) {    // identity matrix
    I32.v[j]=0 ;
    B.BitOn(j*WordSize+j) ;
  }
}



/////////////////////////
//
//  try to find solutions
//
unsigned long Lanczos::Solve(std::vector<unsigned long*>& solution)
{
  static unsigned long seed=time(NULL) ;
  unsigned long j,k ;
  unsigned long l ;
//  long l ;
  unsigned long iRank=32 ;
  block_vector eta(WordSize) ;
  block_vector eta1(WordSize) ;

  block_vector Temp(WordSize) ;

  block_vector BV(BT->ncol) ;

  iter=0 ;

  // get randomized vector
  do {
    for( j=0 ; j<X.ndim ; j++ ) X.v[j]=irandom(seed) ;
    AV.isMul(*BT,BV.isMulT(*BT,X)) ;
    Temp.isIpro(X,AV) ;
    mask=0xFFFFFFFF ;
    Wi1=I32 ;
    iRank=Select_S_W(mask,Temp.v,Wi1.v) ;
//  cout << endl << "iRank " << iRank ;
  } while( iRank<30 ) ;

  unsigned long bound=BT->nrow/(WordSize>>1) + 10;

                                        //  V2=V0=A*X ;
  V0.isMul(*BT,BV.isMulT(*BT,X)) ;        // B^T*(B*V)
  V2=V0 ;

                           // calc i=1  V_1 ߂
  // AV=A*V2
  AV.isMul(*BT,BV.isMulT(*BT,V2)) ;       // A*B=B^T*(B*V)



  VAV2.isIpro(V2,AV) ;     // V2^T*AV

  VA2V.isIpro(AV,AV) ;     // V^T*A^T*A*V
  Temp=VAV2 ;
  Wi2=I32 ;

  mask=0xFFFFFFFF ;
  iRank=Select_S_W(mask,Temp.v,Wi2.v) ;
  iter++ ;

  eta.isIpro(V0,V0) ;
  X.isAdd(X,V.isMul(V0,Temp.isMul(Wi2,eta))) ;


//  MakeD(Wi2,VA2V,mask,VAV2,eta) ;
  {
    if( mask!=0xFFFFFFFF ) {
      D.isAND(VA2V,mask) ;
      eta.isAdd(D,VAV2) ;
    }
    else {
      eta.isAdd(VA2V,VAV2) ;
    }
    D.isMul(Wi2,eta) ;
    D.isAdd(D,I32) ;
  }
  V1.isAND(AV,mask) ;


//  V1=V1+V0*D  ;                              // compute V_1
  V1.isAdd(V1,AV.isMul(V0,D) ) ;         // V1 <- V1+V0*D
                        // using AV as work area

//  V1=AV*SST+V0*D  ;

//               // debug W_0*A*V_1=0 ?
//  AV=A*V1 ;                        // calc i=2  V_2
  AV.isMul(*BT,BV.isMulT(*BT,V1)) ;
  VAV1.isIpro(V1,AV) ;
  VA2V1.isIpro(AV,AV) ;

  if( VAV1.isZero() ) goto RET1 ;

  Temp=VAV1 ;
  Wi1=I32 ;

  mask1=mask ;
  iRank=Select_S_W(mask1,Temp.v,Wi1.v) ;
  iter++ ;

  eta1.isIpro(V1,V0) ;                  // eta1=V1^T*V0

//  X=X+V1*Wi1*VV ;            // Wi1=Winv_1
  V.isMul(V1,Temp.isMul(Wi1,eta1)) ;        // using V as work
  X.isAdd(X,V) ;

//  MakeD(Wi1,VA2V1,mask1,VAV1,eta1) ;      // D <- D_2
  {
    if( mask1!=0xFFFFFFFF ) {
      D.isAND(VA2V1,mask1) ;
      eta1.isAdd(D,VAV1) ;
    }
    else {
      eta1.isAdd(VA2V1,VAV1) ;
    }
    D.isMul(Wi1,eta1) ;
    D.isAdd(D,I32) ;

  }

//cout << endl << "D 2 " << D ;

//  MakeE(Wi2,VAV1,mask1) ;            //  E <- E_2  Wi2 = Wi_0
  {
    E.isMul(Wi2,VAV1) ;
    if( mask1!=0xFFFFFFFF ) E.isAND(E,mask1) ;
  }
  V.isAND(AV,mask1) ;

  //  V=V+V1*D+V2*E ;
  V.isAdd(V,AV.isMul(V1,D)) ; // V <- V +V1*D   AV [NƂĎgp
  V.isAdd(V,AV.isMul(V2,E)) ; // V <- V +V2*E

  mask=mask1 ;
  for( ;; iter++ ) {
                        // calculate AV=A*V ;
    AV.isMul(*BT,BV.isMulT(*BT,V)) ;
    VAV.isIpro(V,AV) ;

    if( VAV.isZero() ) {                // terminate ?
#ifdef LANCZOS_DEBUG
      if( !V.isZero() ) {
        cout << endl << " VAV=0 but V!=0 "   ;
      }
#endif
      break ;
    }

    // V_i^T*A*A*V_i
    VA2V.isIpro(AV,AV) ;
    Temp=VAV ;
    Wi=I32 ;

    iRank=Select_S_W(mask,Temp.v,Wi.v) ;
    if( iRank==0 ) break ;



    eta.isIpro(V,V0) ;
//    X=X+V*Wi*VV ;            // Wi1=W_1
    AV.isAND(AV,mask) ;          // sύXȂ

    // mask1=0xFFFFFFFF ̂Ƃ VAV1*Wi1=I32  F=0
    if( mask1!=0xFFFFFFFF ) {
//      MakeF(Wi2,VAV1,Wi1,VA2V1,mask,eta1) ;       // calc F_{i+1} Wi2=W_1
      {
        F.isMul(VAV1,Wi1) ;
        F.isAdd(F,I32) ;
        Temp.isMul(F,eta1) ;          // eta1=(VA2V1*SST1+VAV1)
        F.isMul(Wi2,Temp) ;
        F.isAND(F,mask) ;
      }
      AV.isAdd(AV,V2.isMul(V2,F)) ; // WV <- WV +V2*F
    }

    V2.isMul(V,Temp.isMul(Wi,eta)) ;       // V2 ̓[N
    X.isAdd(X,V2) ;

//    MakeD(Wi,VA2V,mask,VAV,eta) ;                   // calc D_{i+1}
    {
      if( mask!=0xFFFFFFFF ) {
        D.isAND(VA2V,mask) ;
        eta.isAdd(D,VAV) ;
      }
      else {
        eta.isAdd(VA2V,VAV) ;
      }
      D.isMul(Wi,eta) ;
      D.isAdd(D,I32) ;
    }


//    MakeE(Wi1,VAV,mask) ;         // calc E_{i+1}
    {
      E.isMul(Wi1,VAV) ;
      if( mask!=0xFFFFFFFF ) E.isAND(E,mask) ;
    }

    // A*V_i+V_i*D_{i+1}+V_{i-1}*E_{i+1}+V_{i-2}*F_{i+1}

                        //    AV=AV*SST+V*D+V1*E+V2*F ;
    AV.isAdd(AV,V2.isMul(V,D)) ;      // WV <- WV +V*DAV2 ̓[NƂĎgp
    AV.isAdd(AV,V2.isMul(V1,E)) ;     // WV <- WV +V1*E  ???

    if( iter>bound ) {
      cout << endl << "loop over " ;
      return true ;
    }

    Swap(V2,V1) ;       // V2 <- V1
    Swap(V1,V) ;        // V1 <- V
    Swap(V,AV) ;        // V<- WV ;
    Swap(Wi2,Wi1) ;     // Wi2<-Wi1
    Swap(Wi1,Wi) ;      // Wi1 <- Wi
    Swap(VA2V1,VA2V) ;  // VA2V1 <- VA2V
    Swap(VAV2,VAV1) ;   // VAV2 <- VAV1
    Swap(VAV1,VAV) ;    // VAV1 <- VAV
#ifdef LANCZOS_DEBUG
    Swap(SST1,SST) ;    // SST1 <- SST
#endif
    mask1=mask ;
    Swap(eta1,eta) ;    //eta1 <- eta

  }

RET1: ;

  unsigned long len1=(BT->nrow+31)>>5;
  unsigned long len2=(BT->ncol+31)>>5;
  unsigned long* sol ;
  unsigned long* Bsol ;
//  X.isAdd(X,Y) ;    // X <- X+Y


//  V1.isMul(*BT,V2.isMulT(*BT,X)) ;  // V1 <- A*X
  V1.isMul(*BT,BV.isMulT(*BT,X)) ;  // V1 <- A*X

  unsigned long* BZ[2*WordSize] ;
  unsigned long* Z[2*WordSize] ;

  // AX=0 ̉߂
  // s (AX,X) ɂ AX |o
  // find non-trivaial solutions of AX=0 
  // eliminate matrix (AX,X) 
  for( l=0,j=0 ; j<WordSize ; j++,l++ ) {
    sol=new unsigned long[len2] ;
    memset(sol,0x00,len2*sizeof(unsigned long)) ;
    BZ[l]=sol ;
    for( k=0 ; k<BT->ncol ; k++ ) {
      if( BitIndex(V1.v,k*WordSize+j) ) BitOn(sol,k) ;
    }

    sol=new unsigned long[len1] ;
    Z[l]=sol ;
    memset(sol,0x00,len1*sizeof(unsigned long)) ;
    for( k=0 ; k<BT->nrow  ; k++ ) {
      if( BitIndex(X.v,k*WordSize+j) ) BitOn(sol,k) ;
    }
  }

  // s (AV,V) ɂ AV |o
  // eliminate (AV,V) 
  if( !V.isZero() ) {

//    AV.isMul(*BT,V2.isMulT(*BT,V)) ;
    AV.isMul(*BT,BV.isMulT(*BT,V)) ;
    for( j=0 ; j<WordSize ; j++,l++ ) {
      sol=new unsigned long[len2] ;
      memset(sol,0x00,len2*sizeof(unsigned long)) ;
      BZ[l]=sol ;
      for( k=0 ; k<BT->ncol ; k++ ) {
        if( BitIndex(AV.v,k*WordSize+j) ) BitOn(sol,k) ;
      }
      sol=new unsigned long[len1] ;
      memset(sol,0x00,len1*sizeof(unsigned long)) ;
      Z[l]=sol ;
      for( k=0 ; k<BT->nrow ; k++ ) {
        if( BitIndex(V.v,k*WordSize+j) ) BitOn(sol,k) ;
      }
    }

  }

//  cputime.Stop() ;


  iRank=Gauss(BT->ncol,BT->nrow,l,BZ,Z) ;

  // Z[iRank],Z[iRank+1],,,  AX=0 ̉
  // Z[iRank],Z[iRank+1],,, are solutions of AX=0 
  for( j=0 ; j<iRank; j++ ) delete [] Z[j] ;
  // l-iRank  AX=0 ̉̌
  // number of solutions AX=0 is l-iRank 
  for( ; j<l ; j++ ) {
    solution.push_back(Z[j]) ;
  }
  for( j=0 ; j<l ; j++ ) delete [] BZ[j] ;


//    cout << endl << "nullity " << iRank ;

  // BX=0 ̉ AX=0 ̉狁߂

  // find solutions of BX=0 from the solutions of AX=0 
  // A=B*B^T
  unsigned long** ZA=new unsigned long*[solution.size()] ;
  unsigned long** ZB=new unsigned long*[solution.size()] ;

  // vectors in Ker A
  for( j=0 ; j<solution.size() ; j++ )
    ZA[j]=solution[j] ;

  // B*ZA[j] ߂
  // compute B*ZA[j] 
  for( j=0 ; j<solution.size() ; j++ ) {
    Bsol=new unsigned long[(BT->ncol+31)/32] ;
    Mul(solution[j],*BT,Bsol) ;
    ZB[j]=Bsol ;
  }

// l  Ker B ɂxNǧ  debug
#ifdef LANCZOS_DEBUG
  for( l=0,j=0 ; j<solution.size() ; j++ ) {
    for( k=0 ; k<(BT->ncol+31)/32 ; k++ ) {
      if( ZB[j][k]!=0 ) goto AAA ;
    }
    l++ ;
  AAA:
  }
#endif

  // ͂߂ k  BX=0 ̉ɂ͂Ȃ肦Ȃ
  k=Gauss(BT->ncol,BT->nrow,solution.size(),ZB,ZA) ;


// debug
#ifdef LANCZOS_DEBUG
cout << endl << "post #BX!=0  init #sol BX!=0 " << k << " " << (solution.size()-l) ;
  for( j=k ; j<solution.size() ; j++ ) {
    Mul(ZA[j],*BT,Bsol) ;
    for( l=0 ; l<(BT->ncol+31)/32 ; l++ ) {
      if( ZB[j][l]!=0 ) { cout << endl << " not solution " ; break ; }
    }
  }
#endif

  for( j=0 ; j<solution.size() ; j++ ) {
    delete [] ZB[j] ;
  }
  delete [] ZB ;
/***************
  unsigned long** ZZ=new unsigned long*[solution.size()-k] ;
  for( j=k ; j<solution.size() ; j++ ) {
    ZZ[j-k]=ZA[j] ;
  }
  delete [] ZA ;
*********************/

//  for( j=0 ; j<solution.size() ; j++ ) ZZ[j]=solution[j] ;

  // Ker B^T ̃xNg ZA[k],... œƗȂ̂߂
  iRank=Gauss(BT->nrow,0,solution.size()-k,ZA+k,NULL) ;
  reverse(ZA+k,ZA+solution.size()) ;

  for( j=0 ; j<solution.size() ; j++ ) {
    solution[j]=ZA[j] ;
  }
  delete [] ZA ;

  // 擪 iRank Ɨȉ
  reverse(solution.begin(),solution.end()) ;
  reverse(solution.begin(),solution.begin()+iRank) ; // ̏͂ȂĂ悢
  return iRank ;

/*************
  for( j=k ; j<solution.size() ; j++ ) {
    solution[j]=ZZ[j] ;
  }
  for( ; solution.size()!=iRank ;) {
    delete [] solution.back() ;
    solution.pop_back() ;
  }

  delete [] ZZ ;
****************/

//  for( j=0 ; j<solution.size() ; j++ ) {


#ifdef LANCZOS_DEBUG
  // for debug {ɉɂȂĂ邩mF
  for( j=0 ; j<solution.size() ; j++ ) {
     unsigned long* ww=new unsigned long[len2] ;
     Mul(solution[j],*BT,ww) ;
     cout << endl ;
     for( k=0 ; k<len1 ; k++ )
       cout << hex << solution[j][k] ;
     cout << endl ;
     for( k=0 ; k<len2 ; k++ )
       cout << hex << ww[k] ;
     cout << dec ;
     for( k=0 ; k<len1 ; k++ ) {
       if( solution[j][k]!=0 ) break ;
     }
     if( k<len1 ) ;
     else {
       solution.erase(solution.begin()+j,solution.begin()+j+1) ;
       j-- ;
     }
     delete [] ww ;
  }
#endif


//cout << endl << "gaind " << solution.size() << " solution #iterate" << jj << " " << cputime ;

//  if( l!=0 ) return true ;
//  else return false ;

}






#else        // sparse matrix is represented by B





//////////////////////////////////////////////////////
//
//   A=B^T*B  B ɓWJ Lanczos @Kp
//
//

Lanczos::~Lanczos()
{
  if( bdelete ) delete B ;
}

Lanczos::Lanczos(const unsigned long nrow,const unsigned long ncol):
                 B(NULL),
                 V0(ncol),V(ncol),V1(ncol),V2(ncol),
                 AV(ncol),
                 Wi(WordSize),Wi1(WordSize),Wi2(WordSize),
                 VAV(WordSize),VAV1(WordSize),
                 VA2V(WordSize),VA2V1(WordSize),
                 D(WordSize),E(WordSize),F(WordSize),
                 VAV2(WordSize),X(ncol),/*Y(nrow),*/I32(WordSize),iter(0),bdelete(true)

#ifdef LANCZOS_DEBUG
                 ,SST(WordSize),SST1(WordSize)
#endif
{
  unsigned long j ;
  TBit BB ;
  BB.SetBitString(WordSize*WordSize,I32.v) ;

  for( j=0 ; j<WordSize ; j++ ) {    // identity matrix
    I32.v[j]=0 ;
    BB.BitOn(j*WordSize+j) ;
  }

  B=new SparseMatrix(nrow,ncol) ;
}

Lanczos::Lanczos(const SparseMatrix* _B):

                 B(_B),
                 V0(_B->ncol),V(_B->ncol),V1(_B->ncol),V2(_B->ncol),
                 AV(_B->ncol),
                 Wi(WordSize),Wi1(WordSize),Wi2(WordSize),
                 VAV(WordSize),VAV1(WordSize),
                 VA2V(WordSize),VA2V1(WordSize),
                 D(WordSize),E(WordSize),F(WordSize),
                 VAV2(WordSize),X(_B->ncol),/*Y(nrow),*/I32(WordSize),iter(0),bdelete(false)
#ifdef LANCZOS_DEBUG
                 ,SST(WordSize),SST1(WordSize)
#endif
{
  unsigned long j ;
  TBit BB ;
  BB.SetBitString(WordSize*WordSize,I32.v) ;

  for( j=0 ; j<WordSize ; j++ ) {    // identity matrix
    I32.v[j]=0 ;
    BB.BitOn(j*WordSize+j) ;
  }
}



/////////////////////////
//
//  try to find solutions
//
unsigned long Lanczos::Solve(std::vector<unsigned long*>& solution)
{
  static unsigned long seed=time(NULL) ;
  unsigned long j,k ;
  unsigned long l ;
//  long l ;
  unsigned long iRank=32 ;
  block_vector eta(WordSize) ;
  block_vector eta1(WordSize) ;

  block_vector Temp(WordSize) ;

  block_vector BV(B->nrow) ;

  iter=0 ;

  // get randomized vector whose rank with respect to metric B^T*B 
  // is at least 30  
  // Thanks to Prof. Yuji KIDA
  do {
    for( j=0 ; j<X.ndim ; j++ ) X.v[j]=irandom(seed) ;
    AV.isMulT(*B,BV.isMul(*B,X)) ;       // A*B=B^T*(B*V)
    Temp.isIpro(X,AV) ;     // V2^T*AV
//    AV.isMul(*BT,BV.isMulT(*BT,X)) ;
//    Tmp.isIpro(X,AV) ;
    mask=0xFFFFFFFF ;
    Wi1=I32 ;
    iRank=Select_S_W(mask,Temp.v,Wi1.v) ;
//  cout << endl << "iRank " << iRank ;
  } while( iRank<WordSize-2 ) ;

//  for( j=0 ; j<X.ndim ; j++ ) X.v[j]=irandom(seed) ;

  // max number of iteration
  unsigned long bound = B->ncol/(WordSize>>1) + 10;

                                        //  V2=V0=A*X ;
//  V0.isMul(*BT,BV.isMulT(*BT,X)) ;        // B^T*(B*V)
  V0.isMulT(*B,BV.isMul(*B,X)) ;
  V2=V0 ;

                           // compute V_1 
  // AV=A*V2
//  AV.isMul(*BT,BV.isMulT(*BT,V2)) ;       // A*B=B^T*(B*V)
  AV.isMulT(*B,BV.isMul(*B,V2)) ;       // A*B=B^T*(B*V)



  VAV2.isIpro(V2,AV) ;     // V2^T*AV

  VA2V.isIpro(AV,AV) ;     // V^T*A^T*A*V
  Temp=VAV2 ;
  Wi2=I32 ;

  mask=0xFFFFFFFF ;
  iRank=Select_S_W(mask,Temp.v,Wi2.v) ;
  iter++ ;

  eta.isIpro(V0,V0) ;       // V0^T*V0 inner product
  X.isAdd(X,V.isMul(V0,Temp.isMul(Wi2,eta))) ;


//  MakeD(Wi2,VA2V,mask,VAV2,eta) ;
  {
    if( mask!=0xFFFFFFFF ) {
      D.isAND(VA2V,mask) ;
      eta.isAdd(D,VAV2) ;
    }
    else {
      eta.isAdd(VA2V,VAV2) ;
    }
    D.isMul(Wi2,eta) ;
    D.isAdd(D,I32) ;
  }
  V1.isAND(AV,mask) ;


//  V1=V1+V0*D  ;                              // calculate V_1
  V1.isAdd(V1,AV.isMul(V0,D) ) ;         // V1 <- V1+V0*D
                        // AV is work

//  V1=AV*SST+V0*D  ;

//               // debug W_0*A*V_1=0 ?
//  AV=A*V1 ;                        // calc i=2  V_2
//  AV.isMul(*BT,BV.isMulT(*BT,V1)) ;
  AV.isMulT(*B,BV.isMul(*B,V1)) ;
  VAV1.isIpro(V1,AV) ;
  VA2V1.isIpro(AV,AV) ;

  if( VAV1.isZero() ) goto RET1 ;

  Temp=VAV1 ;
  Wi1=I32 ;

  mask1=mask ;
  iRank=Select_S_W(mask1,Temp.v,Wi1.v) ;
  iter++ ;

  eta1.isIpro(V1,V0) ;                  // eta1=V1^T*V0

//  X=X+V1*Wi1*VV ;            // Wi1=Winv_1
  V.isMul(V1,Temp.isMul(Wi1,eta1)) ;        // V is work
  X.isAdd(X,V) ;

//  MakeD(Wi1,VA2V1,mask1,VAV1,eta1) ;      // D <- D_2
  {                                         // compute D 
    if( mask1!=0xFFFFFFFF ) {
      D.isAND(VA2V1,mask1) ;
      eta1.isAdd(D,VAV1) ;
    }
    else {
      eta1.isAdd(VA2V1,VAV1) ;
    }
    D.isMul(Wi1,eta1) ;
    D.isAdd(D,I32) ;             // D=D+I

  }

//cout << endl << "D 2 " << D ;

//  MakeE(Wi2,VAV1,mask1) ;            //  E <- E_2  Wi2 = Wi_0
  {                                    // compute E 
    E.isMul(Wi2,VAV1) ;
    if( mask1!=0xFFFFFFFF ) E.isAND(E,mask1) ;
  }
  V.isAND(AV,mask1) ;

  //  V=V+V1*D+V2*E ;
  V.isAdd(V,AV.isMul(V1,D)) ; // V <- V +V1*D   AV is work
  V.isAdd(V,AV.isMul(V2,E)) ; // V <- V +V2*E

  mask=mask1 ;
  for( ;; iter++ ) {
                        // calculate AV=A*V ;
//    AV.isMul(*BT,BV.isMulT(*BT,V)) ;
    AV.isMulT(*B,BV.isMul(*B,V)) ;
    VAV.isIpro(V,AV) ;

    if( VAV.isZero() ) {                // terminate ?
#ifdef LANCZOS_DEBUG
      if( !V.isZero() ) {
        cout << endl << " VAV=0 but V!=0 "   ;
      }
#endif
      break ;
    }

    // V_i^T*A*A*V_i
    VA2V.isIpro(AV,AV) ;
    Temp=VAV ;
    Wi=I32 ;

    iRank=Select_S_W(mask,Temp.v,Wi.v) ;
    if( iRank==0 ) break ;



    eta.isIpro(V,V0) ;
//    X=X+V*Wi*VV ;            // Wi1=W_1
    AV.isAND(AV,mask) ;          

    // mask1=0xFFFFFFFF ̂Ƃ VAV1*Wi1=I32  F=0
    if( mask1!=0xFFFFFFFF ) {
//      MakeF(Wi2,VAV1,Wi1,VA2V1,mask,eta1) ;       // calc F_{i+1} Wi2=W_1
      {                                             // compute F
        F.isMul(VAV1,Wi1) ;
        F.isAdd(F,I32) ;
        Temp.isMul(F,eta1) ;          // eta1=(VA2V1*SST1+VAV1)
        F.isMul(Wi2,Temp) ;
        F.isAND(F,mask) ;
      }
      AV.isAdd(AV,V2.isMul(V2,F)) ; // WV <- WV +V2*F
    }

    V2.isMul(V,Temp.isMul(Wi,eta)) ;       // V2 ̓[N
    X.isAdd(X,V2) ;

//    MakeD(Wi,VA2V,mask,VAV,eta) ;                   // calc D_{i+1}
    {
      if( mask!=0xFFFFFFFF ) {
        D.isAND(VA2V,mask) ;
        eta.isAdd(D,VAV) ;
      }
      else {
        eta.isAdd(VA2V,VAV) ;
      }
      D.isMul(Wi,eta) ;
      D.isAdd(D,I32) ;
    }


//    MakeE(Wi1,VAV,mask) ;         // calc E_{i+1}
    {
      E.isMul(Wi1,VAV) ;
      if( mask!=0xFFFFFFFF ) E.isAND(E,mask) ;
    }

    // A*V_i+V_i*D_{i+1}+V_{i-1}*E_{i+1}+V_{i-2}*F_{i+1}

                        //    AV=AV*SST+V*D+V1*E+V2*F ;
    AV.isAdd(AV,V2.isMul(V,D)) ;      // WV <- WV +V*DAV2 is work
    AV.isAdd(AV,V2.isMul(V1,E)) ;     // WV <- WV +V1*E  ???

    if( iter>bound ) {
      cout << endl << "loop over " ;
      return true ;
    }

    Swap(V2,V1) ;       // V2 <- V1
    Swap(V1,V) ;        // V1 <- V
    Swap(V,AV) ;        // V<- WV ;
    Swap(Wi2,Wi1) ;     // Wi2<-Wi1
    Swap(Wi1,Wi) ;      // Wi1 <- Wi
    Swap(VA2V1,VA2V) ;  // VA2V1 <- VA2V
    Swap(VAV2,VAV1) ;   // VAV2 <- VAV1
    Swap(VAV1,VAV) ;    // VAV1 <- VAV
#ifdef LANCZOS_DEBUG
    Swap(SST1,SST) ;    // SST1 <- SST
#endif
    mask1=mask ;
    Swap(eta1,eta) ;    //eta1 <- eta

  }

RET1: ;

  unsigned long len1=(B->ncol+31)>>5;
  unsigned long len2=(B->nrow+31)>>5;
  unsigned long* sol ;
  unsigned long* Bsol ;
//  X.isAdd(X,Y) ;    // X <- X+Y


//  V1.isMul(*BT,V2.isMulT(*BT,X)) ;  // V1 <- A*X
//  V1.isMulT(*B,V2.isMul(*B,X)) ;  // V1 <- A*X

  V1.isMulT(*B,BV.isMul(*B,X)) ;  // V1 <- A*X

  unsigned long* BZ[2*WordSize] ;
  unsigned long* Z[2*WordSize] ;

  // AX=0 ̉߂
  // s (AX,X) ɂ AX |o
  
  // find solutions of AX=0 by eliminating (AX,X)  A=B^T*B 
  for( l=0,j=0 ; j<WordSize ; j++,l++ ) {
    sol=new unsigned long[len2] ;
    memset(sol,0x00,len2*sizeof(unsigned long)) ;
    BZ[l]=sol ;
    for( k=0 ; k<B->nrow ; k++ ) {
      if( BitIndex(V1.v,k*WordSize+j) ) BitOn(sol,k) ;
    }

    sol=new unsigned long[len1] ;
    Z[l]=sol ;
    memset(sol,0x00,len1*sizeof(unsigned long)) ;
    for( k=0 ; k<B->ncol  ; k++ ) {
      if( BitIndex(X.v,k*WordSize+j) ) BitOn(sol,k) ;
    }
  }

  if( !V.isZero() ) {

//    AV.isMul(*BT,V2.isMulT(*BT,V)) ;
//    AV.isMulT(*B,V2.isMul(*B,V)) ;
    AV.isMulT(*B,BV.isMul(*B,V)) ;
    for( j=0 ; j<WordSize ; j++,l++ ) {
      sol=new unsigned long[len2] ;
      memset(sol,0x00,len2*sizeof(unsigned long)) ;
      BZ[l]=sol ;
      for( k=0 ; k<B->nrow ; k++ ) {
        if( BitIndex(AV.v,k*WordSize+j) ) BitOn(sol,k) ;
      }
      sol=new unsigned long[len1] ;
      memset(sol,0x00,len1*sizeof(unsigned long)) ;
      Z[l]=sol ;
      for( k=0 ; k<B->ncol ; k++ ) {
        if( BitIndex(V.v,k*WordSize+j) ) BitOn(sol,k) ;
      }
    }

  }

//  cputime.Stop() ;

  iRank=Gauss(B->nrow,B->ncol,l,BZ,Z) ;

  // Z[iRank],Z[iRank+1] are contained in Ker(A) 
  for( j=0 ; j<iRank; j++ ) delete [] Z[j] ;
  // l-iRank is number of solutions AX=0
  for( ; j<l ; j++ ) {
    solution.push_back(Z[j]) ;
  }
  for( j=0 ; j<l ; j++ ) delete [] BZ[j] ;


//    cout << endl << "nullity " << iRank ;

  // BX=0 ̉ AX=0 ̉狁߂
  
  // find solutions of BX=0 from the solutions of AX=0 
  // Ker(B) is subset of Ker(A) (A=B^T*B) 
  unsigned long** ZA=new unsigned long*[solution.size()] ;
  unsigned long** ZB=new unsigned long*[solution.size()] ;

  // vectors in Ker A
  for( j=0 ; j<solution.size() ; j++ )
    ZA[j]=solution[j] ;

  // B*ZA[j] ߂
  for( j=0 ; j<solution.size() ; j++ ) {
    Bsol=new unsigned long[(B->nrow+31)/32] ;

//    Mul(solution[j],*BT,Bsol) ;
    Mul(*B,solution[j],Bsol) ;


    ZB[j]=Bsol ;
  }

// l  Ker B ɂxNǧ  debug
#ifdef LANCZOS_DEBUG
  for( l=0,j=0 ; j<solution.size() ; j++ ) {
    for( k=0 ; k<(B->nrow+31)/32 ; k++ ) {
      if( ZB[j][k]!=0 ) goto AAA ;
    }
    l++ ;
  AAA:
  }
#endif

  // ͂߂ k  BX=0 ̉ɂ͂Ȃ肦Ȃ
  
  // ZA[0],...,Z[k-1] are not contained in Ker(B)
  k=Gauss(B->nrow,B->ncol,solution.size(),ZB,ZA) ;


// debug
#ifdef LANCZOS_DEBUG
cout << endl << "post #BX!=0  init #sol BX!=0 " << k << " " << (solution.size()-l) ;
  for( j=k ; j<solution.size() ; j++ ) {
//    Mul(ZA[j],*BT,Bsol) ;
    Mul(*B,ZA[j],Bsol) ;
    for( l=0 ; l<(B->nrow+31)/32 ; l++ ) {
      if( ZB[j][l]!=0 ) { cout << endl << " not solution " ; break ; }
    }
  }
#endif

  for( j=0 ; j<solution.size() ; j++ ) {
    delete [] ZB[j] ;
  }
  delete [] ZB ;
/***************
  unsigned long** ZZ=new unsigned long*[solution.size()-k] ;
  for( j=k ; j<solution.size() ; j++ ) {
    ZZ[j-k]=ZA[j] ;
  }
  delete [] ZA ;
*********************/

//  for( j=0 ; j<solution.size() ; j++ ) ZZ[j]=solution[j] ;

  // Ker B^T ̃xNg ZA[k],... œƗȂ̂߂
  // select independent vectors from Ker(B^T)={ZA_k} 
  iRank=Gauss(B->ncol,0,solution.size()-k,ZA+k,NULL) ;
  reverse(ZA+k,ZA+solution.size()) ;

  for( j=0 ; j<solution.size() ; j++ ) {
    solution[j]=ZA[j] ;
  }
  delete [] ZA ;

  // 擪 iRank Ɨȉ
  reverse(solution.begin(),solution.end()) ;
  reverse(solution.begin(),solution.begin()+iRank) ; // ̏͂ȂĂ悢
  return iRank ;

/*************
  for( j=k ; j<solution.size() ; j++ ) {
    solution[j]=ZZ[j] ;
  }
  for( ; solution.size()!=iRank ;) {
    delete [] solution.back() ;
    solution.pop_back() ;
  }

  delete [] ZZ ;
****************/

//  for( j=0 ; j<solution.size() ; j++ ) {


#ifdef LANCZOS_DEBUG
  // for debug {ɉɂȂĂ邩mF
  // check whether solutions X satisfy B^T*X=0 or not 
  for( j=0 ; j<solution.size() ; j++ ) {
     unsigned long* ww=new unsigned long[len2] ;
//     Mul(solution[j],*BT,ww) ;
    Mul(*B,solution[j],ww) ;
     cout << endl ;
     for( k=0 ; k<len1 ; k++ )
       cout << hex << solution[j][k] ;
     cout << endl ;
     for( k=0 ; k<len2 ; k++ )
       cout << hex << ww[k] ;
     cout << dec ;
     for( k=0 ; k<len1 ; k++ ) {
       if( solution[j][k]!=0 ) break ;
     }
     if( k<len1 ) ;
     else {
       solution.erase(solution.begin()+j,solution.begin()+j+1) ;
       j-- ;
     }
     delete [] ww ;
  }
#endif


//cout << endl << "gaind " << solution.size() << " solution #iterate" << jj << " " << cputime ;

//  if( l!=0 ) return true ;
//  else return false ;

}

#endif


unsigned long Sel_S_W(unsigned long** M,long nrow,long ncol,long BitSize,long* ipiv) ;




/****************************
unsigned long size=32   ;
#include <algorithm>
#include <vector>
void ReadMatrix(char* filename,SparseMatrix*& sm) ;

void main(void)
{
  cin >> size ;

  unsigned long j,k,m ;
  unsigned long nelem ;
  TBit B,B1 ;
  unsigned long seed=(long)time(NULL) ;
//  cout << endl << "seed  " << seed ;
  unsigned long index[100] ;
  SparseMatrix M(size+100,size) ;
  block_vector bvec(size) ;
  block_vector vv ;
  unsigned long v ;

  SparseMatrix tM,p,q ;
  vector<unsigned long>::iterator it ;
  for( j=0 ; j<size+100 ; j++ ) {
    do {
      nelem=irandom(seed) ;
      nelem=20 ;
    } while( nelem<=0 ) ;
    m=irandom(seed)%(size) ;
    index[0]=m ;
    for( k=1 ;  k<nelem ; k++ ) {

      m=irandom(seed)%(size) ;
      it=find(index,index+k,m) ;
      if( it>=index+k )
        index[k]=m ;
      else k-- ;
    }
    sort(index,index+nelem) ;
    M.SetRowVector(j,nelem,index) ;
  }

//  SparseMatrix tM,*M,p ;
//  ReadMatrix("d:/mint/work/smat.in",M) ;

//  tM=t(*M) ;




  vector<unsigned long*> sol ;

  unsigned long* w=new unsigned long[(M.ncol+31)/32] ;

  for(;;) {
    {
    Lanczos L(M.nrow,M.ncol) ;


//    L.A=p ;
    *L.BT=M ;
//    M.SGauss() ;
//    break ;
    if( L.Solve(sol) ) ;
    for( j=0 ; j<sol.size() ; j++ ) {
      cout << endl << "SOLUTION " <<  hex ;
      for( k=0 ; k<(M.nrow+31)/32 ; k++ ) {
        cout << sol[j][k] ;
      }
      Mul(sol[j],M,w) ;
      cout << endl << hex ;
      for( k=0 ; k<(M.ncol+31)/32 ; k++ ) {
        cout << w[k] ;
      }
    }
//    while( !sol.empty() ) {
//      delete [] sol.back() ;
//      sol.pop_back() ;
//    }
    cout << dec ;
    unsigned long memory ;
//    for( memory=0,j=0 ; j< p.nrow ; j++ ) memory+=p.w_row[j] ;
//    cout << endl << "mem of p " <<memory ;
    for( memory=0,j=0 ; j< M.nrow ; j++ ) memory+=M.w_row[j] ;
    cout << endl << "mem of M " << (4*memory+1021)/1024 << " K byte"  ;
//break ;
    cout << endl << sol.size() << " data gained " ;
    if( sol.size()>=17 ) break ;
    }
  }
  delete [] w ;
}
**********/

/************************
void ReadMatrix(char* filename,SparseMatrix*& sm)
{
  ifstream is(filename) ;
  unsigned long nrow,ncol ;
  unsigned long entry[100] ;
  unsigned long nentry ;
  unsigned long j ;
  long w,w1 ;

  is >> ncol >> nrow ;
  sm=new SparseMatrix(nrow,ncol) ;

  for( j=0 ; j<nrow ; j++ ) {
    for( w1=-1,nentry=0 ;;) {
      is >> w ;

      if( w1!=-1 && w==0 ) {
        sm->SetRowVector(j,nentry,entry) ;
        break ;
      }
      else {
        if( w1!=w ) {
          entry[nentry++]=w ;
          w1=w ;
        }
      }
    }
  }
}
*********************/

void Lanczos::statics(::ostream& os)
{
  os << " #iteration " << iter ;

}


#ifdef LANCZOS_DEBUG
void Lanczos::MakeSST(unsigned long mask,block_vector& SST)
{
  unsigned long wary[32] ;
  long j,k ;
  long l ;
  TBit Bmask ;
  TBit B ;
  // S_i s߂
  Bmask.SetBitString(WordSize,&mask) ;
  for( j=0 ; j<WordSize ; j++ ) {
    SST.v[j]=0 ;
    if( !Bmask[j] ) continue ;
    B.SetBitString(32,SST.v+j) ;
    B.BitOn(j) ;
  }
}

void Lanczos::MakeST(unsigned long mask,block_vector& S)
{
  long j,k ;
  long l ;
  TBit Bmask ;
  TBit B ;
  // S_i s߂
//  block_vector S_i(iRank) ;
  Bmask.SetBitString(WordSize,&mask) ;
  for( l=0,j=0 ; j<WordSize ; j++ ) {
    S.v[j]=0 ;

    if( !Bmask[j] ) continue ;

    B.SetBitString(32,S.v+l) ;
    B.BitOn(j) ;
    l++ ;
  }


}

void Lanczos::MakeS(unsigned long mask,block_vector& S)
{
  unsigned long wary[32] ;
  long j,k ;
  long l ;
  TBit Bmask ;
  TBit B ;
  // S_i s߂
  Bmask.SetBitString(WordSize,&mask) ;
  for( l=0,j=0 ; j<WordSize ; j++ ) {
    S.v[j]=0 ;

    if( !Bmask[j] ) continue ;

    B.SetBitString(32,S.v+j) ;
    B.BitOn(l) ;
    l++ ;
  }


}

#endif   //  #ifdef LANCZOS_DEBUG

#if 0         // MakeXX ̓Rg

//  D=I32+Wi*(VA2V*SST+VAV) ;
void Lanczos::MakeD(block_vector& Wi,block_vector& VA2V,unsigned long mask,
                    block_vector& VAV,block_vector& eta)
{
  if( mask!=0xFFFFFFFF ) {
    D.isAND(VA2V,mask) ;
    eta.isAdd(D,VAV) ;
  }
  else {
    eta.isAdd(VA2V,VAV) ;
  }
  D.isMul(Wi,eta) ;
  D.isAdd(D,I32) ;
}

// Winv_{i-1}*V_i^T*A*V_i*S_i*S_i^T
void Lanczos::MakeE(block_vector& Wi1,block_vector& VAV,unsigned long mask)
{
  E.isMul(Wi1,VAV) ;
  if( mask!=0xFFFFFFFF ) E.isAND(E,mask) ;
}

//  F=Wi2*(I32+VAV1*Wi1)*(VA2V1*SST1+VAV1)*SST ;
void Lanczos::MakeF(block_vector& Wi2,block_vector& VAV1,block_vector& Wi1,
                    block_vector& VA2V1,
                    unsigned long mask,block_vector& eta1)
{
  F.isMul(VAV1,Wi1) ;
  F.isAdd(F,I32) ;
  F.isMul(F,eta1) ;         // eta1=(VA2V1*SST1+VAV1)
  F.isMul(Wi2,F) ;
  F.isAND(F,mask) ;
}
#endif   // #if 0

