//  PPSIQS Ver 1.1
//
//  Copyright (C) by S.Tomabechi 2001
//  All rights reserved.
//
//   factorization by PPSIQS
//
//      small divisors must be removed  before using PPSIQS
//

#define SIQS
#ifdef __BORLANDC__
#define WIN32
#endif

#ifdef __BORLANDC__
#include <algorith>
#else
#include <algorithm>
#endif

#include <vector>

#ifdef  WIN32
#include <io.h>
#endif

#include <fstream.h>
#include <stdlib.h>

#define _TEST
#include "Mint.hpp"        // multiple precision integer
#include "Fint.hpp"        // real number
#include "LBit.hpp"
#include <math.h>
#include "Lanczos.hpp"      // block Lanczos method
#include "SMatrix.hpp"      // sparse matrix
#include "PPMPQS.hpp"


#ifdef WIN32
#  ifdef __BORLANDC__
#    include <constrea.h>
#    include <conio.h>
#    include <classlib\StdTempl.h>
#    pragma option -Jg
#  else
     int wherex(void) ;
     int wherey(void) ;
     void clrscr(void) ;
     void gotoxy(int x,int y) ;
#    include <conio.h>
#    include <memory.h>
#  endif
#endif

#include <signal.h>
int  bhit=0 ;
inline int isBreak(void)
{
  return bhit ;
}

//  receive signal
void signal_kbhit(int dmy)      //signal handler when Ctrl+Pause ( Ctrl+Z ) is pressed.
{
  bhit=1 ;
  cout << endl << "wait a few minutes" << flush << endl ;
}

#include "Error.hpp"

#ifdef _MEMMAN
#include "MemMan.hpp"
MemMan Mem ;
#endif

#include "CPUTime.hpp"

using namespace std ;

typedef long LONG2[2] ;
typedef long LONG3[3] ;
typedef unsigned short USHORT4[4] ;
typedef unsigned long ULONG2[2] ;

#define MIN_SIEVE 100
//#define MIN_SIEVE 0

#define MaxPrimes (1000000)
extern const unsigned long MaxSieveResult=512;
#define MaxSearchLoop        256 // 64
#define MaxCycle             1024 // 512

#define MaxSieveUnit (0x00060000UL)

#define SUCCESS_ACCESS 0
#define ACCESS_NONE 0
//#define COMB_DIFF   3
#define COMB_DIFF   1
extern unsigned long SieveUnit=MaxSieveUnit ; // to be multiple of 4
void Sieve(PPrimes* Primes,unsigned long nPrimes,unsigned long nPrimes2,
           unsigned long nPowers);
void SieveInit(unsigned long ncomb,unsigned long nPrimes,
            PPrimes* Primes,unsigned long* Bmodp,Mint& b,Mint& c
            ) ;

void SieveShift(unsigned long nPrimes,PPrimes* Primes,unsigned long* Bmodp,long sign,
        Mint& b,Mint& c) ;
void SieveShiftp(unsigned long nPrimes,PPrimes* Primes,unsigned long* Bmodp) ;
void SieveShiftm(unsigned long nPrimes,PPrimes* Primes,unsigned long* Bmodp);

void SetTable(Mint& a,unsigned long ncomb,unsigned long* Bmodp[],
        unsigned long nPrimes,PPrimes* Primes);
void SieveDone(unsigned long ncomb,PPrimes* Primes) ;
//void nextB(unsigned long n,Mint& b) ;
bool select_a(Mint&a,Mint& mina,Mint& maxa,Mint& N,unsigned long n,unsigned long ment);

void DivideByBase(unsigned long nPrimes,PPrimes* Primes);
bool BitDecomp(long nPrimes,unsigned long* Primes,Mint& q,unsigned long* v2);
unsigned long KS(const Mint& N,const unsigned long nPrimes) ;  // Knuth Sch\"oppel function
//bool powerdetect(Mint& n,Mint& p,unsigned long& e) ;
Mint SquareTest(const Mint& n) ;


Mint _fx ;
Mint _a ;
Mint _b ;
long _DivResult ;
long _nResult ;
char SieveArray[MaxSieveUnit] ;   // work array for sieving
unsigned long _A[20] ;
unsigned long _B[20] ;
unsigned long _OFF ;
unsigned long nSieveResult=MaxSieveResult ;
unsigned long SieveResult[MaxSieveResult+1] ;
unsigned long Lmax ;
unsigned long Pmax ;
unsigned long Pmax2[2] ;
unsigned long Lmax2[2] ;
unsigned long Pmax4 ;

char Cutoff ;
unsigned long SieveWidth ;
unsigned long SieveWidth2 ;

unsigned long lbyte ;

unsigned long nPrimes ;
unsigned long nPrimes2 ;
unsigned long nPowers ;
unsigned long cache_size=128 ;
long _Kr_tab3[]={0,-1,1} ;              //  -(n/3)   Jacobi symbol

static Mint y ;
Mint& pp=y ;
#ifdef __BORLANDC__
constream con ;
#endif


static long MergeDuplicatedData(long nF) ;
bool BitDecomp(long nPrimes,Mint& T) ;

int PrimeDecomp(long nPrimes,unsigned long* Primes,
                                unsigned long* index,Mint& q) ;
unsigned long Gauss(unsigned long nBit1,unsigned long nBit2,
                    unsigned long nRow,unsigned long** Z1,unsigned long** Z2) ;
void Gauss(::iostream& mpqsmat,unsigned long** M,long mRow,
                     long nRow,long nCol,long size) ;
void Gauss(unsigned long** M,long nRow,long nCol,long size,unsigned long* ipiv) ;
//void SieveInit(unsigned long nPrimes,unsigned long nPrimes2) ;
void SetMatrix(::fstream& mpqsfx,unsigned long* ifx,
               unsigned long nX,unsigned long nF,
               unsigned long nPrimes,
               unsigned long* prime);

char* parse_in(istream& is) ;
Mint parse(const char* exp) ;
bool select_a(Mint&a,Mint& maxa,Mint& mina,Mint& N,unsigned long n,unsigned long ment) ;

void select_a(Mint&a,Mint& maxa,Mint& mina,Mint& N) ;
void  nextB(unsigned long n,Mint& b);

//
// work files
//
static char primefile[]="./SIQPRM.SIQ" ;
static ::fstream mpqsprime ;

static char rootfile[]="./SIQROOT.SIQ" ;
static ::fstream mpqsroot ;

static char fxfile[]="./SIQSFX.SIQ" ;
static ::fstream mpqsfx ;

static char ixfile[]="./SIQSFXIX.SIQ" ;
static ::fstream mpqsfxix ;

static char ppfxfile[]="./SIQSPPFX.SIQ" ;
static ::fstream mpqsppfx ;

static char pairfile[]="./SIQSPAIR.SIQ" ;
static ::fstream mpqspair ;

static char pppfile[]="./SIQSPPP.SIQ" ;
static ::fstream mpqsppp ;

static char errfile[]="./SIQS.LOG" ;
static ::fstream mpqslog ;

static char mpqsred[]="./SIQS-RED.SIQ" ;
static ::fstream mpqs_red ;

static char mpqscol[]="./SIQS-COL.SIQ" ;
static ::fstream mpqs_col ;

static char mpqsrow[]="./SIQS-ROW.SIQ" ;
static ::fstream mpqs_row ;

static char elemfile[]="./SIQSELEM.SIQ" ;
static ::fstream elem ;

static char parmfile[]="./SIQSPARM.IN" ;
static ::fstream parms ;

static char ansfile[]="./SIQSANS.SIQ" ;
static ::fstream mpqs_ans ;

static char ansfile2[]="./SIQSANS2.SIQ" ;
static ::fstream mpqs_ans2 ;
//
// file structure
//
//  mpqsfx   : fully factored (smooth) data
//             1. value of quadratic polynomial
//             2. sqaure part (a*x+b)^2-N mod N
//
//  mpqsfxix : index ( file pointer ) to mpqsfx
//
//  mpqsppfx : partially factored data
//             1 large primes p1,p2
//             2. value of quadratic polynomial
//             3. square part
//
//  mpqsppp  : index ( file pointer ) to mpqsppfx
//             1. large primes p1,p2             (2 * 4 bytes)
//             2. file pointer fo mpqsppfx       (4 bytes)
//
//             ꂸɎc p1,p2    ( 4 oCg)
//             mpqsppfx ւ̃t@C|C^ ( 4 oCg)
//             (p1,p2) mpqsppfx ֓ǂ݂ɍŝ
//             ʓ|Ȃ̂Ń_uĎĂ.
//
//  mpqspair : smooth relations generated from relations with large primes
//             1. file pointers to mpqsppp
//                file pointer < 0 : index to mpqsppfx
//                file pointer > 0 : index to mpqspair
//
//              (p)^A(pp)^ ꂽ֌W
//              ɂȂ̃f[^ւ̃t@C|C^ (4oCg*2)
//               񎟊֐lA͏ĂȂD
//              ̂Ƃ mpqsppfx ւ̃|C^
//              ̂Ƃ mpqspair ւ̃|C^
//
//  mpqssqrt : roots of x^2=p mod N ( not used now)
//



Mint _t ;
static Mint pN ;
static Mint sq ;
static TBit U,V,W ;
int nLoop_rho ;
unsigned long _p1,_p2 ;
static long LargePrime(long needF,long& nLPP) ;

long npp[10] ;
int  posy ;
Mint const2(2) ;
char _Copyright[]="PPSIQS Ver 1.1 by S.Tomabechi 2001" ;

char lg1[MAX_COMB] ;
unsigned long ip[MAX_COMB+1] ;
unsigned long pq[MAX_COMB+1] ;
unsigned long gap[MAX_COMB+1] ;
unsigned long pbase ;
unsigned long base ;
Mint BS[MAX_COMB+1] ;
unsigned long* Bmodp[MAX_COMB+1] ;
unsigned long step[MAX_COMB+1] ;
#ifdef SIQS_DEBUG
unsigned long Cmodp[MAX_COMB+1] ;
#endif
int nmachine=1 ;
int machine_no=0 ;
PPrimes* Primes ;


#ifdef PROFILE_TEST
unsigned long cycles[MaxCycle+1] ;
unsigned long duplicate ;
#endif
///////////////////////////////////////////
//
//  compute Knuth-Schr\"oppel function
//
unsigned long KS(const Mint& N,const unsigned long nPrimes)
{
    unsigned long p,q ;
    unsigned long pX ;
    double fMax ;
    unsigned long pMax ;
    Mint pN ;
    unsigned long j ;
    double f ;

    p=1 ;
    pX=p ;
    fMax=0.0 ;
    pMax=(ilog2(N)+1)/3 ;
    if( pMax>40 ) pMax=40 ;  // up to 40

    for(;;) {

        for(;;) {
            pN=p*N ;
            if( (long(pN)&7)==1 ) break ;            // =1 mod 8 ?
            else p=nextprm(p) ;
        }

        f=2*log(2.0)-log(double(p))/2 ;
        q=2 ;
        for( j=2 ; j<nPrimes ; j++ ) {
            do {                           // search prime q s.t. N is a quadratic residue mod q
                q=nextprm(q)    ;
            } while( Kronecker(pN,q)<0 ) ;
            f+=(Kronecker(pN,q)+1)*log(double(q))/q ;
        }
        if( f>fMax ) { fMax=f ; pX=p ; }
        if( p<pMax ) p=nextprm(p) ;
        else break ;
    }
    return pX ;

}

////////////////////////////////////////////
//
//  define factor base
//
//  nPrimes (I) number of primes
//  N       (I) composite number to be factored
//  Primes  (I) array of primes. to be allocated
//  maxPrimes (O) max prime of factor base
void SetPrimes(const unsigned long nPrimes,const Mint& N,
               PPrimes* Primes,unsigned long& maxPrime)
{
    // decide factor base
    // select primes q s.t. pN is a quadratic residue mod q
    unsigned long p,q,wp ;
    unsigned long j ;
    wp=p=2 ;
    for( j=2 ; j<nPrimes ; j++ ) {
        for(;;) {
            q=nextprm(wp) ;
            if( Kronecker(N,q)==-1 ){           // qaudratic residue ?
                wp=q ;                          // not so
                continue ;
            }
            else break ;                        // qaudratic residue
        }
        Primes[j].diff=(unsigned short)(q-p) ;  // set difference of prime and previous prime
//      Primes[j].dmy=1 ;
        wp=p=q ;
#ifdef SIQS_DEBUG
        if( prmdiv(q)!=1 ) {
            cout << endl << "not prime " << q ;
            Error(not prime) ;
        }
#endif
    }
    maxPrime=p ;

}

////////////////////////////////////////////////////
//
//  set square roots  x^2=N mod p
//
//  nPrimes (I) number of primes
//  N       (I) composite number to be factored
//  pX      (I) multiplier ( Knuth-Schr\"oppel function )
//  Primes  (I/O) array of primes. to be allocated
//
void SetSquareRoot(const unsigned long nPrimes,const Mint& N,const unsigned long pX,
                   PPrimes* Primes)
{
    unsigned long q ;
    unsigned long j ;
    q=2 ;
    for( j=2 ; j<nPrimes ; j++ ) {
        q+=(unsigned long)(Primes[j].diff) ;

#ifdef SIQS_DEBUG
        if( Kronecker(N,q)==-1 ) {
            cout << endl << q << flush ;
            Error(not square) ;
        }
#endif
        if( q==pX ) Primes[j].Root=0 ;    // q | N  i.e. (N/q)=0
        else Primes[j].Root=sqrt(N,q) ;   // square root x such that x^2=N mod q

#ifdef SIQS_DEBUG
        if( q!=pX ) {
            if( sqrt(N,q)!=sqrt(N,long(q)) ) {
                sqrt(N,long(q)) ;
                sqrt(N,q) ;
                cout << endl << j << q ;
                Error( long sqrt error) ;
            }
            if( (N-Mint(Primes[j].Root)*Primes[j].Root)%q!=0 ) {
                cout << endl << "sqrt Mint " << sqrt(N,Mint(q)) ;
                cout << endl << "sqrt long " << sqrt(N,long(q)) ;
                cout << endl << "sqrt ulong " << sqrt(N%q,q) ;
                cout << endl << "N mod q " << (N%q) ;

                cout << endl << j << " " << q ;
                cout << endl << (N-Mint(Primes[j].Root)*Primes[j].Root)%q ;
                Error( sqrt error) ;
            }
        }
#endif
    }
}

////////////////////////////////////////////////////
//
//  set logarithm
//
//  nPrimes (I) number of primes
//  pX      (I) multiplier ( Knuth-Schr\"oppel function )
//  Scale   (I) scale factor
//  Primes  (I/O) array of primes. to be allocated
//
void Setlog(const unsigned long nPrimes,const unsigned long pX,const double Scale,
                   PPrimes* Primes)
{

    unsigned long q ;
    unsigned long j ;
    q=0 ;
    for( j=1 ; j<nPrimes ; j++ ) {
        q+=(unsigned long)(Primes[j].diff) ;
        if( q==pX ) Primes[j].logP=char(Scale*log(double(q))/2) ;  // q is multiplier
        else Primes[j].logP=char(Scale*log(double(q))) ;
    }
}


void SetPrimesPowers(const unsigned long nPrimes,
                     const Mint& pN,
                     const unsigned long pX,         // multiplier
                     const double Scale,
                     const unsigned long PPmax,
                     const unsigned long PPPmax,
                     const unsigned long offset,
                     PPrimes* Primes,
                     unsigned long& nPowers)
{
    unsigned long p,q ;
    unsigned long j ;
    unsigned long p1,p2 ;
    double f ;
    Mint x ;
    q=2 ;
    nPowers=0 ;
    for( j=2 ; j<nPrimes ; j++ ) {
        q+=(unsigned long)(Primes[j].diff) ;
        if( q>PPPmax ) break ;
        if( q==pX ) continue ; // pX is the multiplier

        p=q ;
        p1=p ;
        p*=q ;
        f=Scale*log(double(p)) ;
        p2=sqrt(pN,q) ;
        while( p<PPmax ) {
            x=(pN-p2*Mint(p2))/p1 ;
            p2=mod(p2+x*invmodp(2*p2,q)*_p1,p) ;
            f=Scale*log(double(p))-f ;
            if( f<1.0 ) continue    ;
            Primes[offset+nPowers].logP=(char)f ;
            Primes[offset+nPowers].Root=p2 ;
            Primes[offset+nPowers].diff=(unsigned short)p ;
            nPowers++ ;
            p1=p ; p*=q ;
        }

    }

}



void Set2Powers(const unsigned long nPrimes,
                     const Mint& pN,
                     const double Scale,
                     const unsigned long PPmax,
                     PPrimes* Primes,
                     unsigned long& nPrimes2)
{
    unsigned long q ;
    unsigned long j ;
    unsigned long p,p1,p2 ;
    double f ;

    q=2 ;
    p=q ;
    p1=p ;
    p*=q ;
    f=Scale*log(double(p)) ;
    p2=1 ;
    nPrimes2=0 ;
    for(;;) {
        for( j=0 ; j<2 ; j++ ) {
            p1=p ;
            p*=q ;
            p2=mod((pN-p2*Mint(p2))/p1,q) ;
            p2+=p2*p1 ;
        }
        if( p>=PPmax ) break ;
        f=Scale*log(double(p))-f ;
        if( f<1.0 ) continue    ;
        Primes[nPrimes2].logP=(char)f ;
        Primes[nPrimes2].Root=p2 ;
        Primes[nPrimes2].diff=(unsigned short)p ;
        nPrimes2++ ;
    }

}


////////////////////////////////
//
// write factor base to file
//
void WritePrimes(const unsigned long nPrimes,const unsigned long nPowers,
                 const unsigned long nPrimes2,
                 PPrimes* Primes)
{
    unsigned long j ;
    unsigned long p ;
    mpqsprime.open(primefile,ios::out|ios::binary) ;
    mpqsprime.write((char*)&nPrimes,sizeof(unsigned long)) ;
    mpqsprime.write((char*)&nPowers,sizeof(unsigned long)) ;
    mpqsprime.write((char*)&nPrimes2,sizeof(unsigned long)) ;
    p=2 ;
    for( j=2 ; j<nPrimes ; j++ ) {
      p+=(unsigned long)(Primes[j].diff) ;
      mpqsprime.write((char*)&p,sizeof(p)) ;
    }
    Primes+=nPrimes ;
    for( j=0 ; j<nPowers ; j++ ) {
      p=(unsigned long)(Primes[j].diff) ;
      mpqsprime.write((char*)&p,sizeof(p)) ;
    }
    Primes+=nPowers ;
    for( j=0 ; j<nPrimes2 ; j++ ) {
      p=(unsigned long)(Primes[j].diff) ;
      mpqsprime.write((char*)&p,sizeof(p)) ;
    }

    mpqsprime.close() ;
}

////////////////////////////////
//
// write square roots to file
//
void WriteSquareRoots(const unsigned long nPrimes,const unsigned long nPowers,
                 const unsigned long nPrimes2,
                 PPrimes* Primes)
{
    unsigned long j ;
    unsigned long p ;
    mpqsroot.open(rootfile,ios::out|ios::binary) ;
    mpqsroot.write((char*)&nPrimes,sizeof(unsigned long)) ;
    mpqsroot.write((char*)&nPowers,sizeof(unsigned long)) ;
    mpqsroot.write((char*)&nPrimes2,sizeof(unsigned long)) ;
    p=2 ;
    for( j=2 ; j<nPrimes ; j++ ) {
      p=Primes[j].Root ;
      mpqsroot.write((char*)&p,sizeof(p)) ;
    }
    Primes+=nPrimes ;
    for( j=0 ; j<nPowers ; j++ ) {
      p=Primes[j].Root ;
      mpqsroot.write((char*)&p,sizeof(p)) ;
    }
    Primes+=nPowers ;
    for( j=0 ; j<nPrimes2 ; j++ ) {
      p=Primes[j].Root ;
      mpqsroot.write((char*)&p,sizeof(p)) ;
    }

    mpqsroot.close() ;
}

////////////////////////////////
//
// read square roots from file
//
void ReadSquareRoots(const unsigned long nPrimes,const unsigned long nPowers,
                 const unsigned long nPrimes2,
                 PPrimes* Primes)
{
    unsigned long j ;
    unsigned long p ;
    mpqsroot.open(rootfile,ios::in|ios::binary) ;
    mpqsroot.read((char*)&p,sizeof(unsigned long)) ;
    mpqsroot.read((char*)&p,sizeof(unsigned long)) ;
    mpqsroot.read((char*)&p,sizeof(unsigned long)) ;
    p=2 ;
    for( j=2 ; j<nPrimes ; j++ ) {
      mpqsroot.read((char*)&p,sizeof(p)) ;
      Primes[j].Root=p ;
    }
    Primes+=nPrimes ;
    for( j=0 ; j<nPowers ; j++ ) {
      mpqsroot.read((char*)&p,sizeof(p)) ;
      Primes[j].Root=p ;
    }
    Primes+=nPowers ;
    for( j=0 ; j<nPrimes2 ; j++ ) {
      mpqsroot.read((char*)&p,sizeof(p)) ;
      Primes[j].Root=p ;
    }

    mpqsroot.close() ;
}



////////////////////////////////
//
// read factor base from file
//
void ReadPrimes(unsigned long& nPrimes,unsigned long& nPowers,
                 unsigned long& nPrimes2,
                 PPrimes*& Primes,
                 unsigned long& maxPrime)
{
    unsigned long j ;
    unsigned long p ;
    unsigned long q ;
    PPrimes* Ptemp ;
    mpqsprime.open(primefile,ios::in|ios::binary) ;
    mpqsprime.read((char*)&nPrimes,sizeof(unsigned long)) ;
    mpqsprime.read((char*)&nPowers,sizeof(unsigned long)) ;
    mpqsprime.read((char*)&nPrimes2,sizeof(unsigned long)) ;
    Ptemp=Primes=new PPrimes[nPrimes+nPowers+nPrimes2] ;

    Ptemp[0].diff=-1 ;
    Ptemp[1].diff=2 ;
    p=2 ;
    for( j=2 ; j<nPrimes ; j++ ) {
        mpqsprime.read((char*)&q,sizeof(unsigned long)) ;
        Ptemp[j].diff=(unsigned short)(q-p) ;
        p=q ;
    }
    Ptemp+=nPrimes ;
    maxPrime=p ;
    for( j=0 ; j<nPowers ; j++ ) {
        mpqsprime.read((char*)&q,sizeof(unsigned long)) ;
        Ptemp[j].diff=(unsigned short)q ;
    }
    Ptemp+=nPowers ;
    for( j=0 ; j<nPrimes2 ; j++ ) {
        mpqsprime.read((char*)&q,sizeof(unsigned long)) ;
        Ptemp[j].diff=(unsigned short)q ;
    }

    mpqsprime.close() ;
}

void ReadPrimes(unsigned long& nPrimes,
               unsigned long*& primes)
{
    unsigned long p ;
    mpqsprime.open(primefile,ios::in|ios::binary) ;
    mpqsprime.read((char*)&nPrimes,sizeof(unsigned long)) ;
    mpqsprime.read((char*)&p,sizeof(unsigned long)) ;
    mpqsprime.read((char*)&p,sizeof(unsigned long)) ;
    primes=new unsigned long[nPrimes] ;

    primes[0]=(unsigned long)(-1) ;
    primes[1]=2 ;
    mpqsprime.read((char*)(primes+2),(nPrimes-2)*sizeof(unsigned long)) ;
    mpqsprime.close() ;
}



int main(int argc,char* argv[])
{

#ifdef PROFILE_TEST
    long maxcurve=1000000000 ;
    unsigned long ndiv=0 ;
#endif
    double scaleFB ;
    unsigned long* prime ;
    vector<unsigned long*> sol ;
    long kLPP(0),lLPP(0),mLPP(0) ;
    Mint N ;
    Mint d ;
    Mint NN ;
    unsigned long* index ;
    long nF(0) ;
    bool bsame ;
    unsigned long ncomb ;
    char* input_str ;
    unsigned long lenN ;
    long j,k ;

    for( j=0 ; j<MAX_COMB ; j++ ) {
//        step[j]=prm(j+1) ;
//step[j]=j+1 ;
        step[j]=1 ;
    }

Input_Number:


    bsame=false ;
    nF=0 ;
    kLPP=0 ;
    lLPP=0 ;
    mLPP=0 ;

    if( argc>1 ) {                      // cache size is specified
        unsigned long cache ;
        cache=atoi(argv[1]) ;
        if( cache==0 ) ;
        else {
            if( cache>6*64 ) cache_size=3*cache_size ;   // greather than prepared array
            else cache_size=cache ;
        }
#ifdef PROFILE_TEST
        if( argc>2 ) {
//          maxcurve=atoi(argv[2]) ;
            ndiv=atoi(argv[2]) ;
        }
        else ndiv=0 ;
#endif

    }
    SieveUnit=cache_size*1024 ;
//    if( parms.bad() ) {
    ncomb=0 ;
    if( access(parmfile,ACCESS_NONE)!=SUCCESS_ACCESS ) {  // parameter file does not exist
        NN=0 ;
        d=0 ;
        ncomb=0 ;
    }
    else {
        // get previous parameters
        parms.open(parmfile,ios::in) ;
        parms >> NN ;
        parms >> d ;
    }

    cout << endl << "Input number ( input 0 to exit )" << endl ;
//  cin >> SieveArray ;
    input_str=parse_in(cin) ;
    N=parse(input_str) ;
    if( N==0 ) return 0 ;                // exit
    if( N<0 ) ~N ;
    if( Even(N) ) Error(input number is even) ;

    // primality test
    if( Rabin(N) ) {                                         // test primality
        cout << endl << N << " is probably prime " ;
        if( APRCL(N) ) {
            mpqslog.open(errfile,ios::out|ios::app) ;
            mpqslog << endl << " ========================= " ;
            mpqslog << endl << N << " is prime " ;
            mpqslog.close() ;
            goto Input_Number ;
        }
    }

    CPUTime cputime ;
    CPUTime CPULanczos ;
#ifdef PROFILE_TEST
    CPUTime cpu_sela ;
    CPUTime cpu_sieve ;
    CPUTime cpu_sieve_all ;
    CPUTime cpu_sieve_init ;
    CPUTime cpu_table ;
    CPUTime cpu_shift ;
    CPUTime cpu_try_div ;
#endif
long nLPP[3]={0,0,0} ;
memset(npp,'\0',10*sizeof(long)) ;

    if( NN!=0 && N==NN ) {      // check whether input number is as same as previous one.
        bsame=true ;            // yes  same number
        parms >> nF ;           // restart siveing
        parms >> kLPP ;
        parms >> lLPP ;
        lLPP=0 ;
        parms >> mLPP ;
        mLPP=0 ;
        parms >> nLPP[0] ;
        nF=nLPP[0] ;
        parms >> nLPP[1] ;
        parms >> nLPP[2] ;
        cputime.Reset() ;
        cputime.Start() ;
        parms >> cputime ;      // add duration time until now
        parms >> cache_size ;
        SieveUnit=cache_size*1024 ;
    }
    else {
        unsigned long e ;
        cputime.Start() ;
        if( NN!=0 ) parms.close() ;
        if( powerdetect(N,d,e) ) {
            if( d!=1 ) {
                cout << endl << "   is power of prime " << d ;
                mpqslog.open(errfile,ios::out|ios::app) ;
                mpqslog << endl << " ========================= by SIQS " ;

                char PC[2] ;
                mpqslog <<  endl << input_str <<" = " ;
//              if( d>N ) Swap(d,N) ;                            // sort
                // primality test
                PC[0]='P' ; mpqslog << "P" ;
                mpqslog << len10(d) << "^" << e  ;
#if 0
            if( APRCL(N) ) { PC[1]='P' ; mpqslog << "P" ; }
            else           { PC[1]='C' ; mpqslog << "C" ; }
            mpqslog << len10(N)  ;
#endif

                mpqslog << endl << PC[0] << len10(d) <<" = " << d ;
//          mpqslog << endl << PC[1] << len10(N) <<" = " << N ;

                mpqslog << endl << "cputime " << cputime ;
                mpqslog.close() ;
//              unlink(parmfile) ;
                delete [] input_str ;
                goto Input_Number ;
            }
        }
        d=SquareTest(N) ;
        if( d>0 ) {
            cout << endl << " is square of " << d ;
            mpqslog.open(errfile,ios::out|ios::app) ;
            mpqslog << endl << " ========================= by SIQS " ;

            char PC[2] ;
            mpqslog <<  endl << input_str <<" = " ;
//          if( d>N ) Swap(d,N) ;                            // sort
            // primality test
            if( APRCL(d) ) { PC[0]='P' ; mpqslog << "P" ; }
            else         { PC[0]='C' ; mpqslog << "C" ; }
            mpqslog << len10(d) << "^2 " ;
#if 0
            if( APRCL(N) ) { PC[1]='P' ; mpqslog << "P" ; }
            else           { PC[1]='C' ; mpqslog << "C" ; }
            mpqslog << len10(N)  ;
#endif

            mpqslog << endl << PC[0] << len10(d) <<" = " << d ;
//          mpqslog << endl << PC[1] << len10(N) <<" = " << N ;

            mpqslog << endl << "cputime " << cputime ;
            mpqslog.close() ;
            delete [] input_str ;
//          unlink(parmfile) ;
            goto Input_Number ;
        }
        d=0 ;                   // start from the beggining

    }

    unsigned long icomb ;
    long l ;
    bool found ;
    found=false ;
    unsigned long p,q,pX ;
#ifdef  WIN32
#  ifdef __BORLANDC__
    con.window(1,1,80,45) ;
    con.clrscr() ;
#  else
    clrscr() ;
#  endif
#else
    cout << "\e[1H\e[2J" ;  // clear screen  (for Linux)

#endif
    if( ilog2(N)<256 ) scaleFB=0.95 ;
    else scaleFB=1.0 ;
scaleFB=1.0 ;

    // decide size of factor base
#ifdef __BORLANDC__
    long nX ;
     nX=min(MaxPrimes,
            int(
    #if 1
               scaleFB*1.78
    #else
                0.95*1.78
    #endif
                *expl(
                     sqrtl(
                         logl((long double)N)
                             *logl(logl((long double)N))
                         /12
                        )
                 )
             ) /32*32
         ) ;

#else
    long nX ;
    nX=int(
    #if 0
               scaleFB*1.78
    #else
                0.95*1.78
    #endif
                 *exp(
                         sqrt(
                                log((double)N)
                                *log(log((double)N))
                                /12
                            )
                    )
            )/32*32;
    if( nX>MaxPrimes ) nX=MaxPrimes ;
#endif

#if 0
nX*=1.2 ;
#endif

    nX=nX<56 ? 56:nX ;

    lenN=len10(N) ;
    if( lenN<=30 ) {
        nX+=256 ;
    }

    nPrimes=nX-32 ;

    cout << _Copyright ;
    cout << endl << "C" << lenN << "=" << N ;

    Primes=new PPrimes[nPrimes+100] ;

    Primes[0].diff=(unsigned short)(-1) ;   // -1 and 2 must be contained in factor base
    Primes[0].logP=0 ;
    Primes[1].diff=2 ;


    // determine a multiplier so that get more primes of factor base
    // fȂ悤 N->pN multiplier ߂
    pX=KS(N,nPrimes) ;

    pN=pX*N ;                    // replace N by N * multiplier

    // decide factor base
    // select primes q s.t. pN is a quadratic residue mod q
    unsigned long maxPrime ;
    if( !bsame ) {
        SetPrimes(nPrimes,pN,Primes,maxPrime) ;
    }
    else {
        delete [] Primes ;
        ReadPrimes(nPrimes,nPowers,nPrimes2,Primes,maxPrime) ;
    }
    Pmax=maxPrime ;

    for( j=0 ; j<nPrimes ; j++ ) {
        Primes[j].dmy=1 ;
    }

#ifndef LINEAR_SIEVE
    q=0 ;               // do not sieve for the primes < MIN_SIEVE
    for( j=0 ; j<nPrimes ; j++ ) {
        q+=Primes[j].diff ;
        if( q<MIN_SIEVE ) Primes[j].dmy=0 ;
        else break ;
    }
#endif

    // determine parameters
#ifdef __BORLANDC__
    long double lnpN=logl((long double)pN) ;
    SieveWidth=(unsigned long)(40000*expl(0.66*logl(lnpN))) ;
#else
    double lnpN=log((double)pN) ;
    SieveWidth=(unsigned long)(40000*exp(0.66*log(lnpN))) ;
#endif
    if( ilog2(N)<224 ) k=4 ;
    else k=2 ;
//k=4 ;
k=2 ;
#if 1
    SieveWidth=2*(SieveWidth/SieveUnit/2+k-1)*SieveUnit/k ;
#else
    SieveWidth=(SieveWidth/SieveUnit/2+k-1)*SieveUnit/k ;
#endif
    #if 1
    if( lenN<80 ) {
        (SieveWidth>>=7)<<=2 ;
    }
    else if( lenN<90 ) (SieveWidth>>=5)<<=2   ;
    else (SieveWidth/=20)<<=2   ;
//    else (SieveWidth>>=6)<<=2   ;
//  if( ndiv!=0 ) ((SieveWidth/=ndiv)>>=2)<<=2 ;
//  else (SieveWidth>>=6)<<=2   ;
//cout << endl << "SieveWidth " << SieveWidth ;
    #else
//(SieveWidth>>=7)<<=2    ;   // 60ȉłOK
(SieveWidth>>=6)<<=2  ;   // 90 ł͂̂炢Kv
//(SieveWidth>>=4)<<=2  ;   // 90 ł͂̂炢Kv
    #endif


//(SieveWidth>>=3)<<=2  ;
#if 1
    if( SieveWidth<SieveUnit ) {
        SieveUnit=SieveWidth ;
    }
    SieveWidth=(SieveWidth+SieveUnit-1)/SieveUnit*SieveUnit ;
    SieveWidth2=SieveWidth<<1 ;
#else
    j=ilog2(SieveWidth)-3 ;     // 60 digit ȉł͍œK
    j=ilog2(SieveWidth)-4 ;
//  SieveWidth>>(j-1) ;
//  if( SieveWidth&1 ) j++ ;
    SieveWidth=1<<j ;
//  SieveWidth>>=
    if( SieveUnit>SieveWidth ) SieveUnit=SieveWidth ;
    SieveWidth2=SieveWidth<<1 ;
#endif

#if 1
    double MeanValue=log(double(SieveWidth)*SieveWidth*((long double)pN)/2) ;
#else
    double MeanValue=log(double(SieveWidth)*SieveWidth*sqrt(((long double)pN)/2)) ;
#endif

    double Scale=2.0*255/MeanValue ;

    long PPmax=isqrt(Pmax) ;
    if( PPmax<65536 ) PPmax=65536 ;
    long PPPmax=isqrt(PPmax) ;
    if( PPPmax<100 ) PPPmax=100 ;

    double Success ;
    double RealSuccess ;
    if( len(pN)<266 ) (Success=(0.15000-0.00019*len(pN)))/=1.05 ;
//  if( len(pN)<256 ) Success=0.825*(0.15000-0.00019*len(pN)) ;
//  if( len(pN)<256 ) Success=(0.15000-0.00019*len(pN))/1.4 ;
//    else (Success=0.11450-0.000143*len(pN))/=1.1 ;
    else (Success=0.11450-0.000143*len(pN)) ;
//    else (Success=0.092118-0.0001732*len(pN)) ;

//Success*=0.75;

    Mint x,w ;

    x=Mint(Pmax)*Pmax ;
    Pmax2[0]=(unsigned long)x ;
    Pmax2[1]=(unsigned long)x(-1) ;

    // these magic numbers are owed to Prof. Kida
#ifdef __BORLANDC__
    Cutoff=char((0.7075+0.0108*logl((long double)N))*log(Pmax)*Scale) ;
#else
    Cutoff=char((0.7075+0.0108*log((long double)N))*log(Pmax)*Scale) ;
#endif
    Lmax=(unsigned long)(Pmax*(1.28+0.40*lnpN)) ;
//Lmax<<=1 ;
#if 0
cout << endl << "lnpN " << lnpN ;
    Lmax=(unsigned long)(Pmax*(1.28+0.50*lnpN)) ;
#endif
    if( Lmax<Pmax ) Lmax=Pmax ;
#ifdef __BORLANDC__
    w=Int(exp(1.78*Fint(logl(Lmax)))) ;
//  w=Int(exp(2.00*Fint(logl(Lmax)))) ;
//  (w=Lmax)*=Lmax ;
#else
    w=Int(exp(1.78*Fint(log(Lmax)))) ;
#endif
    Lmax2[0]=(unsigned long)w ;
    Lmax2[1]=(unsigned long)w(-1) ;
    nLoop_rho=isqrt(Lmax)/100 ;    // max loop number of rho-method

    cout << endl << "#FactorBase " << nPrimes ;
    cout << " max FactorBase " << Pmax ;
    cout << endl << "max LargePrime " << Lmax ;
    cout << " Upper Bound " << w ;
    cout << endl << "SieveWidth " << SieveWidth  ;
#ifdef PROFILE_TEST
    cout << " SieveWidth2 " << SieveWidth2  ;
#endif
    cout << " SieveUnit " << SieveUnit ;
    cout << " CutOff " << (int)Cutoff ;
    cout << " multiplier " << pX ;

    // calculate square roots of the equation X*X=pN mod q
    if( !bsame ){
        SetSquareRoot(nPrimes,pN,pX,Primes) ;
    }
    else {
        ReadSquareRoots(nPrimes,nPowers,nPrimes2,Primes) ;
    }

    // set Scale*log_2(P)
    Setlog(nPrimes,pX,Scale,Primes) ;

    if( ilog2(pN)>256 && !bsame ) {
      SetPrimesPowers(nPrimes,
                     pN,
                     pX,         // multiplier
                     Scale,
                     PPmax,
                     PPPmax,
                     nPrimes,
                     Primes,
                     nPowers) ;

      Set2Powers(nPrimes,
                     pN,
                     Scale,
                     PPmax,
                     Primes+nPrimes+nPowers,
                     nPrimes2) ;

    }


    if( !bsame ) {
        WritePrimes(nPrimes,nPowers,nPrimes2,Primes) ;
        WriteSquareRoots(nPrimes,nPowers,nPrimes2,Primes) ;
    }


    // open work files
    mpqsfx.open(fxfile,ios::binary|ios::out|ios::in) ;
    if( mpqsfx.bad() ) {
        cerr << endl << "Error : could not open file " << fxfile ;
        Error(file not found) ;
    }

    mpqsfxix.open(ixfile,ios::binary|ios::in|ios::out) ;
    if( mpqsfxix.bad() ) {
        cerr << endl << "Error : could not open file " << ixfile ;
        Error(file not found) ;
    }

    mpqsppfx.open(ppfxfile,ios::binary|ios::in|ios::out) ;
    if( mpqsppfx.bad() ) {
        cerr << endl << "Error : could not open file " << ppfxfile ;
        Error(file not found) ;
    }

    mpqspair.open(pairfile,ios::binary|ios::in|ios::out) ;
    if( mpqspair.bad() ) {
        cerr << endl << "Error : could not open file " << pairfile ;
        Error(file not found) ;
    }
    mpqspair.seekp(8) ;

    mpqsppp.open(pppfile,ios::binary|ios::in|ios::out) ;
    if( mpqsppp.bad() ) {
        cerr << endl << "Error : could not open file " << pppfile ;
        Error(file not found) ;
    }

    if( d==0 ) {
        d=Int(sqrt(sqrt(Fint(pN))/SieveWidth*738/1000)) ;
        if( d<(unsigned long)Pmax+3 ) d=(unsigned long)Pmax+3 ;
        d=(d/4)*4+3 ;                    // d=3 mod 4

        mpqsppfx.write((char*)&_p1,sizeof(_p1)) ; // write dummy data
        mpqsppfx.write((char*)&_p2,sizeof(_p2)) ; // write dummy data
        write(mpqsppfx,_fx) ;
        write(mpqsppfx,sq) ;
    }
    else {
        parms >> Success ;

        parms >> l ;
        for( k=0,j=0 ; j<nF ; j++ ) {
            l=mpqsfx.tellp() ;
            read(mpqsfx,_fx) ;
            read(mpqsfx,sq) ;
#ifdef SIQS_DEBUG
if( !BitDecomp(_nPrimes,_fx) ) {
    PrimeDecomp(_nPrimes-1,Primes+1,(unsigned long*)SievePos,_fx) ;
    cout << endl << "j " << j ;
    cout << endl << "_fx " << _fx ;
    Error(bit decomposition error) ;
}

#endif
            k++ ;
            mpqsfxix.write((char*)&l,sizeof(l)) ;
        }
        nF=k ;

        parms >> l ;

        parms >> l ;
        mpqsppfx.read((char*)&_p1,sizeof(long)) ; // read dummy data
        mpqsppfx.read((char*)&_p2,sizeof(long)) ; // read dummy data
        read(mpqsppfx,_fx) ;
        read(mpqsppfx,sq) ;
        for( j=0 ; j<kLPP ; j++ ) {
            l=mpqsppfx.tellp() ;
            l=-l ;
            mpqsppfx.read((char*)&_p1,sizeof(long)) ;
            mpqsppfx.read((char*)&_p2,sizeof(long)) ;
            mpqsppp.write((char*)&_p1,sizeof(long)) ;
            mpqsppp.write((char*)&_p2,sizeof(long)) ;
            mpqsppp.write((char*)&l,sizeof(l)) ;

            read(mpqsppfx,_fx) ;
            read(mpqsppfx,sq) ;

        }

        parms >> l ;
        parms >> l ;
        parms >> l ;
        if( !parms.eof() ) SieveUnit=l ;
        parms >> base >> pbase ;
        parms >> ncomb ;
        for( j=0 ; j<=ncomb ; j++ ) {
            parms >> ip[j] ;
        }
        for( j=0 ; j<=ncomb ; j++ ) {
            parms >> pq[j] ;
        }
        parms.close() ;
    }


    //SIGBREAK
#ifdef WIN32
//cout << endl << "signal handeler " ;
    signal(SIGBREAK,signal_kbhit) ;
//    signal(SIGBREAK,signal_kbhit) ;
#else
    // set signal handler (Linux version)
    signal(SIGTSTP,signal_kbhit) ;
#endif

    Mint c,a2,d2 ;
    long newFF(0) ;
//  Mint const3(3) ;



//Mint bb ;
Mint a ;
Mint b ;
unsigned long nFF ;
    Mint maxa ;
    Mint mina ;
#ifdef PROFILE_TEST
unsigned long count_of_a(0) ;
#endif
#if 0
//  maxa=13*sqrt(2*Fint(pN))/SieveWidth/10 ;
//  mina=7*sqrt(2*Fint(pN))/SieveWidth2/10 ;
    maxa=10*sqrt(2*Fint(pN))/SieveWidth/8 ;
//maxa=28*sqrt(2*Fint(pN))/SieveWidth/7 ;
//  mina=78*sqrt(2*Fint(pN))/SieveWidth2/100 ;
mina=4*sqrt(2*Fint(pN))/SieveWidth2/5 ;

maxa=8*sqrt(2*Fint(pN))/SieveWidth/7 ;
//maxa=28*sqrt(2*Fint(pN))/SieveWidth/7 ;
//  mina=78*sqrt(2*Fint(pN))/SieveWidth2/100 ;
mina=7*sqrt(2*Fint(pN))/SieveWidth2/8 ;
#else
maxa=130*sqrt(2*Fint(pN))/SieveWidth/7 ;
mina=16*sqrt(2*Fint(pN))/SieveWidth2/8 ;
//mina=sqrt(2*Fint(pN))/SieveWidth ;
//SieveWidth<<=1     ;
//SieveWidth2<<=1    ;
mina=sqrt(Fint(pN))/SieveWidth ;
//mina=sqrt(2*Fint(pN))/SieveWidth2 ;
#endif
    if( ncomb!=0 ) goto INIT ;
//    base=10 ;
    base=5 ;
    //
    //  search the number of sutitable combinations
    //
    for(;;) {
        if( power(Pmax,ncomb)<mina ) { //  Pmax^ncomb must be greater than mina
            ncomb++ ;
            continue ;
        }
        pbase=2 ;
        for( j=2 ; j<nPrimes ; j++ ) {
            pbase+=(unsigned long)Primes[j].diff ;
            if( pbase<pX ) continue ;
//          if( power(pbase,ncomb)>mina ) {
            if( pbase*power(Pmax,ncomb-1)>mina ) {
                base=j ;
                break ;
            }
        }
        if( j>=nPrimes || pbase==2 ) {
            ncomb++ ;
//          continue ;
        }
        else break ;
//cout << endl << "ncomb " << ncomb ;
    }

#if 1
    for(;;) {
        pbase=2 ;
        for( j=2 ; j<nPrimes ; j++ ) {
            pbase+=(unsigned long)Primes[j].diff ;
            if( pbase<pX ) continue ;
            if( power(pbase,ncomb)>mina ) {
//          if( pbase*power(Pmax,ncomb-1)>mina ) {
                base=j ;
#if 0
if( base>100 )
#endif
                break ;
            }
        }
        if( j>=nPrimes || pbase==2 ) {
            ncomb++ ;
        }
        else break ;
//cout << endl << "ncomb " << ncomb ;
    }
#endif

#if 1
//  ncomb+=2 ;
//  ncomb+=2 ;
    if( lenN<=60 ) ncomb+=2 ;
//    else if( lenN<80 ) ncomb+=3 ;
    else if( lenN<=80 ) ncomb+=3 ;
//    else if( lenN<=90 ) ncomb+=4 ;
    else if( lenN<=91 ) ncomb+=4 ;
//    else if( lenN<90 ) ncomb+=2 ;
//    else ncomb+=4 ;
//    else if( lenN<105 ) ncomb+=5 ;
    else ncomb+=5 ;
//ncomb-- ;
//ncomb++ ;
//ncomb++ ;

//  ncomb++;
ABCD:
    for(;;) {
        pbase=2 ;
        for( j=2 ; j<nPrimes ; j++ ) {
            pbase+=(unsigned long)Primes[j].diff ;
            if( pbase<pX ) continue ;
            if( power(pbase,ncomb)>mina ) {
//          if( pbase*power(Pmax,ncomb-1)>mina ) {
                base=j ;
                break ;
            }
        }
        if( j>=nPrimes || pbase==2 ) {
            ncomb++ ;
        }
        else break ;
//cout << endl << "ncomb " << ncomb ;
    }
#endif



//cout << endl << "base pbase " << base << " " << pbase ;
cout <<  " ncomb " << ncomb ;
//cout << endl << "ncomb " << ncomb ;
//cout << endl << "ncomb " << ncomb ;
//cout << endl << "ncomb " << ncomb ;
//cout << endl << "ncomb " << ncomb ;
//cout << endl << "ncomb " << ncomb ;
//cout << endl << power(pbase,ncomb) ;
//cout << endl << mina ;

if( ncomb>20 ) Error(ncomb>1) ;
#if 0
        pbase=2 ;
        for( j=2 ; j<nPrimes ; j++ ) {
            pbase+=(unsigned long)Primes[j].diff ;
            if( pbase<pX ) continue ;
            else {
                base = j ;
                break ;
            }
        }
#endif
    ip[0]=base ;
    pq[0]=pbase ;
    gap[0]=0 ;
    for( j=1 ; j<ncomb ; j++ ) {
        gap[j]=1 ;
        ip[j]=ip[j-1]+1 ;
        pq[j]=pq[j-1]+(unsigned long)Primes[ip[j]].diff ;
    }
//    ip[ncomb]=nPrimes-1 ;
    for( j=0 ; j<machine_no ; j++ ) {
        gap[ncomb-1]++ ;
        ip[ncomb-1]++ ;
        pq[ncomb-1]+=(unsigned long)Primes[ip[ncomb-1]].diff ;
    }
    ip[ncomb]=nPrimes ;
//  for( j=0 ; j<100 ; j++ ) {
//cout << endl << endl << "ncomb " << ncomb ;
//cout << endl << " Cutoff " << int(Cutoff) ;
//Cutoff++ ;
//Cutoff++ ;
//Cutoff++ ;
#if 1
//    Cutoff+=0.30*ncomb ;
//    Cutoff+=ncomb ;
//    Cutoff+=2*0.50*ncomb ;
#endif
//    Cutoff+=+0.2*ncomb ;
//cout << endl << " Cutoff " << int(Cutoff) ;
//Cutoff=100 ;
INIT:
    gap[0]=1 ;
    for( j=1 ; j<=ncomb ; j++ ) {
        gap[j]=ip[j]-ip[j-1] ;      // ga[j]=ip[j]-ip[j-1]
    }
    a=pq[0] ;
    for( j=1 ; j<ncomb ; j++ ) a*=pq[j] ;
nPowers=0 ;
nPrimes2=0 ;
    for( j=0 ; j<=ncomb ; j++ ) {
        Bmodp[j]=new unsigned long[nPrimes] ;
    }

#if 1
#ifdef  WIN32
    cout << endl << "To stop sieving, press Ctrl + Pause" ;
#else
    cout << endl << "To stop sieving, press Ctrl + Z" ;
    cout << endl ;
#endif
#endif

#if 1
    cout << endl << endl;

#ifdef WIN32
# ifdef __BORLANDC__
    posy=con.rdbuf()->wherey() ;
# else
    posy=wherey() ;
# endif
#endif
#endif

#ifdef PROFILE_TEST
    cpu_sela.Reset() ;
    cpu_sieve.Reset() ;
    cpu_sieve_all.Reset() ;
    cpu_sieve_init.Reset() ;
    cpu_table.Reset() ;
    cpu_shift.Reset() ;
    cpu_try_div.Reset() ;
#endif
//  Cutoff+=+0.5*ncomb ;

#ifdef SIQS_DEBUG
q=2 ;
for( j=2 ; j<nPrimes ; j++ ) {

    q+=(unsigned long)Primes[j].diff ;
    if( ((pN-Primes[j].Root*Mint(Primes[j].Root))%q)!=0 ) {
        Error( not a root) ;
    }
}
#endif

///////////////////////////////////////////////////
//
//  sieving process
//


//SieveUnit=SieveWidth2 ;
    for(; ncomb<=MAX_COMB ;) {
#if 0
#ifdef PROFILE_TEST
cputime.Stop() ;
if( cputime.Time()>10*60 ) break ;
cputime.Start() ;
#endif
#endif
//bool select_a(Mint&a,Mint& maxa,Mint& mina,Mint& N,unsigned long n,unsigned long ment)

#ifdef PROFILE_TEST
        cpu_sela.Start() ;
#endif
        if( !select_a(a,mina,maxa,pN,ncomb,nPrimes) ) {
//cout << endl << "select_a false " ;
#if 0
            Cutoff-=+0.5*ncomb ;
#endif
            for( j=0 ; j<=ncomb ; j++ ) {
                delete [] Bmodp[j] ;
            }
            ncomb++ ;
            goto ABCD ;
        }
#ifdef PROFILE_TEST
        cpu_sela.Stop() ;
#endif
//cout << " " << a ;

#ifdef PROFILE_TEST
        maxcurve-- ;
        if( maxcurve<0 ) break ;
#endif
        b=BS[0] ;
        for( icomb=1 ; icomb<ncomb ; icomb++ ) b+=BS[icomb] ;
        c=(b*b-pN)/a ;
#if SIQS_DEBUG
        if( (b*b-a*c)!=pN ) {
            Error(not equal) ;
        }
#endif
#ifdef PRIFILE_TEST
cpu_sieve_init.Start() ;
cpu_sieve_all.Start() ;
#endif
        SetTable(a,ncomb,Bmodp,nPrimes,Primes);
#ifdef PRIFILE_TEST
cpu_sieve_init.Stop() ;
cpu_sieve_all.Stop() ;
#endif

#ifdef SIQS_DEBUG           // check result of SetTable
{
    for( k=0,j=2 ; j<nPrimes ; j++) {
        if( Primes[j].dmy==0 ) {
            if( j!=ip[k] ) {
                cout << endl <<" j k " << j << " " << k << " #" ;
                for( j=0 ; j<=ncomb ; j++ ) cout << ip[j] << " " ;
                Error(index is wrong) ;
            }
            k++ ;
        }
    }
    q=2 ;
    for( k=0,j=2 ; j<nPrimes ; j++ ) {
        q+=(unsigned long)Primes[j].diff ;
        if( Primes[j].dmy==0 ) continue ;
        if( Primes[j].ainv!=invmodp(a,q) ) {
            cout << endl <<  Primes[j].ainv << " " <<invmodp(a,q);
            Error(not a^-1 ) ;
        }
    }
}
#endif

#ifdef SIQS_DEBUG
{
    q=2 ;
    for( j=2 ; j<nPrimes ; j++ ) {
        unsigned long bq ;
        unsigned long abq ;

        q+=(unsigned long)Primes[j].diff ;
        if( Primes[j].dmy==0 ) continue ;
        if( Primes[j].ainv!=invmodp(a,q) ) {
            cout << endl <<  Primes[j].ainv << " " <<invmodp(a,q);
            Error(not a ) ;
        }
        bq=0 ;
        for( k=0 ; k<ncomb ; k++ )  {
            abq=mod(2*BS[k]*Primes[j].ainv,q) ;
            if( abq!=Bmodp[k][j] ) Error(a8a8a) ;
            bq+=Bmodp[k][j] ;
        }
        bq%=q ;
        if( bq!=isMul(2*Bmodp[ncomb][j],Primes[j].ainv,q) ) {
//  if( bq!=isMul(2*Bmodp[ncomb][j],a%q,q) ) {
            cout << endl << bq << " " << isMul(2*Bmodp[ncomb][j],Primes[j].ainv,q) ;
//      cout << endl << bq << " " << isMul(2*Bmodp[ncomb][j],a%q,q) ;
            cout << endl << Bmodp[ncomb][j] << " " << Primes[j].ainv << endl  ;
            for( k=0 ; k<=ncomb ; k++ ) cout << " " << Bmodp[k][j];
            Error(AASASA) ;
        }
        if( (Bmodp[ncomb][j])%q!=(b%q) ) {
    cout << endl << Bmodp[ncomb][j] << " " << b%q  << "  bq "  << bq << "          "  ;
    cout << endl ;
            for( k=0 ; k<=ncomb ; k++ ) cout << " " << BS[k]%q;
            Error(aaaa) ;
        }
    }
}
#endif

#ifdef SIQS_DEBUG
        if( ((b*b-pN)%a)!=0 ) Error(((b*b-pN)%a)!=0) ;
#endif
#ifdef SIQS_DEBUG
{
    q=2 ;
//cout << endl << "c " << c ;
    for( k=0,j=2 ; j<nPrimes ; j++) {
//cout << endl << "Primes[j].ainv " << Primes[j].ainv << " " << j ;
        if( Primes[j].dmy==0 ) {
            if( j!=ip[k] ) {
            cout << endl << j << " " << ip[k] ;
                Error(indec is wrong) ;
            }
            k++ ;
        }
    }
}
#endif
#ifdef PROFILE_TEST
cpu_sieve_init.Start() ;
cpu_sieve_all.Start() ;
#endif

         SieveInit(ncomb,nPrimes,Primes,Bmodp[ncomb],b,c) ;

#ifdef PROFILE_TEST
cpu_sieve_init.Stop() ;
cpu_sieve_all.Stop() ;
#endif


#ifdef SIQS_DEBUG
{
    q=2 ;
//cout << endl << "c " << c ;
    for( k=0,j=2 ; j<nPrimes ; j++) {
//cout << endl << "Primes[j].ainv " << Primes[j].ainv << " " << j ;
        if( Primes[j].dmy==0 ) {
            if( j!=ip[k] ) {
            cout << endl << j << " " << ip[k] ;
                Error(indec is wrong) ;
            }
            k++ ;
        }
    }

for( j=0 ; j<ncomb ; j++ ) {
    cout << endl << Cmodp[j] << " " << mod(-c,pq[j]) << " " << pq[j] ;
}
    for( k=0,j=2 ; j<nPrimes ; j++ ) {
        q+=(unsigned long)Primes[j].diff ;
        if( q<1 ) continue ;
        if( Primes[j].dmy==0 ) {
//          continue ;
            if( Cmodp[k]!=mod(-c,q) ) {
                cout << endl << Cmodp[k] << " " << mod(-c,q) << " " << q ;
cout << endl << " abs%q " << abs(c)%q ;
cout << endl << " c%q " << c%q << "sign(c) " << sign(c) ;
cout << endl << "primes.ainv " << Primes[j].ainv ;
cout << endl << "hoge hoge "  <<  hoge ;
                Error(Cmodp[ll]!=modp(-c,q)) ;
            }
            k++ ;
//          continue ;
        }
        if( Kronecker(b*b-a*c,q)<0 ) {
cout << endl << " a " << a ;
cout << endl << " b " << b ;
cout << endl << " c " << c ;
cout << endl << "b^2-ac " << ( b*b-a*c) ;
cout << endl << "pN " << pN ;
            Error(not square residue) ;
        }
        long qq ;

        qq=Primes[j].SievePos[0]    ;
        if( qq>SieveUnit ) {
            cout << endl << "qq SieveUnit " << qq << " " << SieveUnit ;
            Error(qq>SieveUnit) ;
        }
        qq+=(SieveWidth-SieveUnit) ;
        if( mod(a*qq*qq+2*b*qq+c,q)!=0 ) {
cout << endl << q << " " << qq  << " " << (Primes[j].SievePos[0]+(SieveWidth-SieveUnit)) << " "  ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq-2*b*qq+c,q);

            for(;;) {
                if( mod(a*qq*qq+2*b*qq+c,q)==0 ) break ;
                qq++ ;
            }
cout << endl << "real root " << qq ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
cout << endl << mod(a,q) << "x^2 + " << mod(2*b,q) << " x + " << mod(c,q) ;
cout << endl << "ainv " << Primes[j].ainv ;
            Error(not a root) ;
        }
        qq=Primes[j].SievePos[1]    ;
        qq+=(SieveWidth-SieveUnit) ;
        if( mod(a*qq*qq+2*b*qq+c,q)!=0 ) {
cout << endl << q << " " << qq  << " " << (Primes[j].SievePos[1]+(SieveWidth-SieveUnit)) << " ";
cout << endl << "real root " << qq << " " << mod(a*qq*qq-2*b*qq+c,q);
            for(;;) {
                if( mod(a*qq*qq+2*b*qq+c,q)==0 ) break ;
                qq++ ;
            }
cout << endl << "real root " << qq ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
cout << endl << mod(a,q) << "x^2 + " << mod(2*b,q) << " x + " << mod(c,q) ;
cout << endl << "ainv " << Primes[j].ainv ;
            Error(not a root) ;
        }
    }
}
#endif

//Sieve(Primes,nPrimes,nPrimes2,nPowers) ;
#ifdef PROFILE_TEST
cpu_sieve.Start() ;
cpu_sieve_all.Start() ;
#endif

        Sieve(Primes,nPrimes,nPrimes2,nPowers) ;

#ifdef PROFILE_TEST
cpu_sieve.Stop() ;
cpu_sieve_all.Stop() ;
#endif
//cout << endl << _nResult ;
        y=SieveWidth ; ~y ;
//cout << endl << "y Wid " << y << " " << SieveWidth ;
        for( ; _nResult>0 ; ) {
            _OFF=SieveResult[--_nResult] ;      // offset in seive array
            x=y+_OFF ;

            ((_fx.isMul(a,x)+=(b<<1))*=x)+=c ;      //  (a*x^2+b*x+c)
            if( _fx.isZero() ) continue ;
#ifdef PROFILE_TEST
cpu_try_div.Start() ;
#endif
            DivideByBase(nPrimes,Primes) ;   // divide by factor base
#ifdef PROFILE_TEST
cpu_try_div.Stop() ;
#endif

//cout << " " << _DivResult ;
            if( _DivResult<0 ) continue ;  // not smooth

            nLPP[_DivResult]++ ;
            found=true ;
            sq=mod((a*x+b),pN) ;       // sq=(ax+b)^2-N
            _fx*=a ;
            if ( _DivResult==0 ) {         // fully factored
                l=mpqsfx.tellp() ;
                mpqsfxix.write((char*)&l,sizeof(l)) ;
                write(mpqsfx,_fx) ;
                write(mpqsfx,sq) ;
                nF++ ;

#ifdef SIQS_DEBUG
{
unsigned long nPrimes ;
unsigned long* primes ;
ReadPrimes(nPrimes,primes) ;
if( !BitDecomp(nPrimes,primes,x=_fx,v2) ) {
    cout << endl << "d " << d  ;
    delete [] primes ;
    Error(bit decomposition error) ;
}
delete [] primes ;
}
#endif

            }
            else {                                   //  smooth but large prime(s)
                l=-mpqsppfx.tellp() ;
                mpqsppp.write((char*)&_p1,sizeof(_p1)) ;
                mpqsppp.write((char*)&_p2,sizeof(_p2)) ;
                mpqsppp.write((char*)&l,sizeof(l));  // file pointer to mpqsppfx

                mpqsppfx.write((char*)&_p1,sizeof(_p1)) ;
                mpqsppfx.write((char*)&_p2,sizeof(_p2)) ;
                write(mpqsppfx,_fx) ;
                write(mpqsppfx,sq) ;
                kLPP++ ;

#ifdef SIQS_DEBUG
{
x=_fx ;
(x/=_p1)/=_p2 ;
unsigned long nPrimes ;
unsigned long* primes ;
ReadPrimes(nPrimes,primes) ;
if( !BitDecomp(nPrimes,primes,x,v2) ) {
    cout << endl << "d " << d  ;
    DivideByBase(nPrimes,Primes) ;
    delete [] primes ;
    Error(bit decomposition error) ;
}
delete [] primes ;
}
#endif
            }
        }


//goto DISP ;

#ifdef SIQS_DEBUG
#if 1
q=2 ;
for( j=2 ; j<nPrimes ; j++ ) {
    q+=(unsigned long)Primes[j].diff ;
//  if( Primes[j].dmy==0 ) continue ;
long qq ;
//qq=long(Primes[j].SievePos[1])+long(q)    ;
qq=long(Primes[j].SievePos[1])+long(q)  ;
qq-=long(SieveWidth) ;
    if( mod(a*qq*qq+2*b*qq+c,q)!=0 ) {
cout << endl << q << " " << " " << long(Primes[j].SievePos[1])<< " " << qq ;
        Error(not a root) ;
    }
qq=long(Primes[j].SievePos[0])+long(q)  ;
qq-=long(SieveWidth) ;
    if( mod(a*qq*qq+2*b*qq+c,q)!=0 ) {
cout << endl << q << " " << " " << long(Primes[j].SievePos[0])<< " " << qq ;
        Error(not a root) ;
    }
}
#else
q=2 ;
for( j=2 ; j<nPrimes ; j++ ) {
    q+=(unsigned long)Primes[j].diff ;
    if( Primes[j].dmy==0 ) continue ;
unsigned long qq ;
qq=Primes[j].SievePos[1]    ;
qq+=(SieveWidth-SieveUnit) ;
    if( mod(a*qq*qq+2*b*qq+c,q)!=0 ) {
cout << endl << q << " " << qq ;
        Error(not a root) ;
    }
qq=Primes[j].SievePos[0]    ;
cout << endl << q << " " << qq ;
qq+=(SieveWidth-SieveUnit) ;
    if( mod(a*qq*qq+2*b*qq+c,q)!=0 ) {
cout << endl << q << " " << qq ;
        Error(not a root) ;
    }
}
#endif
#endif
#ifdef SIQS_DEBUG
        if( ((b*b-pN)%a)!=0 ) Error(((b*b-N)%a)!=0) ;
#endif
//cout << endl << b ;
#ifdef PROFILE_TEST
count_of_a++ ;
#endif
        for( icomb=1 ; icomb<(1<<(ncomb-1)) ; icomb++ ) {
#ifdef PROFILE_TEST
maxcurve-- ;
count_of_a++ ;
#endif ;
//          nextB(k,bb) ;
//cout << " " << b ;
//          b%=a ;
#ifdef SIQS_DEBUG
            if( ((b*b-pN)%a)!=0 ) Error(((b*b-N)%a)!=0) ;
#endif

//Sieve0(ncomb,b,Primes,nPrimes,nPrimes2,nPowers) ;
//bb=0 ;
    // Gray code
    unsigned long nu=icomb<<1 ;
    unsigned long nu2 ;
    unsigned long ind ;
    long ssign ;
    nu=BitScanForward(&nu,1) ;
    nu2=1<<nu ;
    ind=icomb/nu2 ;
//  cout << endl << "n nu n/ny " << n << " " << nu << " " << double(n)/nu << " " << n%nu ;
//  b=BS[nu-1] ;
    if( (icomb%nu2)!=0 ) ind++ ;

    if( ind&1 ) ssign=0 ;
    else ssign=1 ;
//cout << endl << " icomb ind nu ssign  " << icomb << " " << ind << " " << nu << " "<< ssign ;
#if 0
    unsigned long nu ;
    unsigned long nu2 ;
    unsigned long ind ;
    long ssign ;

    __asm {
        mov eax,n
//      mov edx,eax
        shl eax,1   // 2*n
//      mov n2,eax
        bsf ecx,eax    //
        mov nu,ecx

        mov ebx,1
        shl ebx,cl   // nu2=1<<nu

        mov eax,n
        xor edx,edx
        div ebx
        or  edx,edx
        jz  AAAA
        inc eax
    }
AAAA:
    __asm {
//      mov ind,eax
        mov ssign,0
        and eax,1
        jz NOT_EVEN
        dec ssign
NOT_EVEN:


    }
#endif
    if( ssign==1 ) b+=(BS[nu-1]<<1) ;
    else b-=(BS[nu-1]<<1) ;
    c=(b*b-pN)/a ;
#ifdef SIQS_DEBUG
{
    q=2+Primes[2].diff ;
    if( Kronecker(b*b-a*c,q)<0 ) {
        cout << endl << "not square residue " ;
        continue ;
//      Error(not square residue) ;
    }
        if( ((b*b-pN)%a)!=0 ) {
            cout << endl << "nu sign " << nu << " " << ssign ;
            cout << endl << "b " << b ;
            cout << endl << "a " << a ;
            cout << "BS " << endl ;
            for( j=0 ; j<ncomb ; j++ ) {
                cout << " " << BS[j] ;
            }
cout << endl << "k " << k ;
            Error(((b*b-pN)%a)!=0) ;
        }
}
#endif
#ifdef PROFILE_TEST
cpu_shift.Start() ;
cpu_sieve_all.Start() ;
#endif
//        SieveShift(nPrimes,Primes,Bmodp[nu-1],ssign,b,c) ;
        if( ssign!=0 ) SieveShiftp(nPrimes,Primes,Bmodp[nu-1]) ;
        else SieveShiftm(nPrimes,Primes,Bmodp[nu-1]) ;

#ifdef PROFILE_TEST
cpu_shift.Stop() ;
cpu_sieve_all.Stop() ;
#endif
#ifdef SIQS_DEBUG
q=2 ;
//cout << endl << "c " << c ;
for( k=0,j=2 ; j<nPrimes ; j++) {
//cout << endl << "Primes[j].ainv " << Primes[j].ainv << " " << j ;
    if( Primes[j].dmy==0 ) {
        if( j!=ip[k] ) {
            Error(aaaaa) ;
        }
        k++ ;
    }
}
//for( j=0 ; j<ncomb ; j++ ) {
//  cout << endl <<Cmodp[j] << " " << pq[j] ;
//}
for( k=0,j=2 ; j<nPrimes ; j++ ) {
    q+=(unsigned long)Primes[j].diff ;
    if( Primes[j].dmy==0 ) {
//      continue ;
        if( Cmodp[k]!=mod(-c,q) ) {
            cout << endl << Cmodp[k] << " " << mod(-c,q) << " " << q ;
cout << endl << " abs%q " << abs(c)%q ;
cout << endl << "primes.dmy " << Primes[j].dmy ;
            Error(Cmodp[ll]!=modp(-c,q)) ;
        }
        k++ ;
//      continue ;
    }
    if( Kronecker(b*b-a*c,q)<0 ) {
cout << endl << " a " << a ;
cout << endl << " b " << b ;
cout << endl << " c " << c ;
cout << endl << "b^2-ac " << ( b*b-a*c) ;
cout << endl << "pN " << pN ;
        Error(not square residue) ;
    }
long qq ;
qq=Primes[j].SievePos[0]    ;
qq+=(SieveWidth-SieveUnit) ;
    if( mod(a*qq*qq+2*b*qq+c,q)!=0 ) {
cout << endl << q << " " << qq ;
        for(;;) {
//break ;
            if( mod(a*qq*qq+2*b*qq+c,q)==0 ) break ;
cout << " " << qq ;
            qq++ ;
        }
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
cout << endl << mod(a,q) << "x^2 + " << mod(2*b,q) << " x + " << mod(c,q) ;
        Error(not a root) ;
    }
qq=Primes[j].SievePos[1]    ;
qq+=(SieveWidth-SieveUnit) ;
    if( mod(a*qq*qq+2*b*qq+c,q)!=0 ) {
cout << endl << q << " " << qq ;
        for(;;) {
            if( mod(a*qq*qq+2*b*qq+c,q)==0 ) break ;
            qq++ ;
        }
cout << endl << "real root " << qq ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
qq-- ;
cout << endl << "real root " << qq << " " << mod(a*qq*qq+2*b*qq+c,q);
cout << endl << mod(a,q) << "x^2 + " << mod(2*b,q) << " x + " << mod(c,q) ;
        Error(not a root) ;
    }
}
#endif
#ifdef PROFILE_TEST
cpu_sieve.Start() ;
cpu_sieve_all.Start() ;
#endif
            Sieve(Primes,nPrimes,nPrimes2,nPowers) ;

#ifdef PROFILE_TEST
cpu_sieve.Stop() ;
cpu_sieve_all.Stop() ;
#endif
            y=SieveWidth ; ~y ;
            for( ; _nResult>0 ; ) {
                _OFF=SieveResult[--_nResult] ;      // offset in seive array
                x=y+_OFF ;

                ((_fx.isMul(a,x)+=(b<<1))*=x)+=c ;      //  (a*x^2+b*x+c)
                if( _fx.isZero() ) continue ;
#ifdef PROFILE_TEST
cpu_try_div.Start() ;
#endif

                DivideByBase(nPrimes,Primes) ;   // divide by factor base
#ifdef PROFILE_TEST
cpu_try_div.Stop() ;
#endif

//cout << " " << _DivResult ;
                if( _DivResult<0 ) continue ;  // not smooth

                nLPP[_DivResult]++ ;
                found=true ;
                sq=mod((a*x+b),pN) ;       // sq=(ax+b)^2-N
                _fx*=a ;
                if ( _DivResult==0 ) {         // fully factored
                    l=mpqsfx.tellp() ;
                    mpqsfxix.write((char*)&l,sizeof(l)) ;
                    write(mpqsfx,_fx) ;
                    write(mpqsfx,sq) ;
                    nF++ ;

#ifdef SIQS_DEBUG
{
unsigned long nPrimes ;
unsigned long* primes ;
ReadPrimes(nPrimes,primes) ;
if( !BitDecomp(nPrimes,primes,x=_fx,v2) ) {
    cout << endl << "d " << d  ;
    delete [] primes ;
    Error(bit decomposition error) ;
}
delete [] primes ;
}
#endif

                }
                else {                                   //  smooth but large prime(s)
                    l=-mpqsppfx.tellp() ;
                    mpqsppp.write((char*)&_p1,sizeof(_p1)) ;
                    mpqsppp.write((char*)&_p2,sizeof(_p2)) ;
                    mpqsppp.write((char*)&l,sizeof(l));  // file pointer to mpqsppfx

                    mpqsppfx.write((char*)&_p1,sizeof(_p1)) ;
                    mpqsppfx.write((char*)&_p2,sizeof(_p2)) ;
                    write(mpqsppfx,_fx) ;
                    write(mpqsppfx,sq) ;
                    kLPP++ ;

#ifdef SIQS_DEBUG
{
x=_fx ;
(x/=_p1)/=_p2 ;
unsigned long nPrimes ;
unsigned long* primes ;
ReadPrimes(nPrimes,primes) ;
if( !BitDecomp(nPrimes,primes,x,v2) ) {
    cout << endl << "d " << d  ;
    DivideByBase(nPrimes,Primes) ;
    delete [] primes ;
    Error(bit decomposition error) ;
}
delete [] primes ;
}
#endif
                }
            }

#ifdef SIQS_DEBUG
#if 1
{
    q=2 ;
    for( j=2 ; j<nPrimes ; j++ ) {
        q+=(unsigned long)Primes[j].diff ;
//  if( Primes[j].dmy==0 ) continue ;
        long qq ;
//qq=long(Primes[j].SievePos[1])+long(q)    ;
        qq=long(Primes[j].SievePos[1])+long(q)  ;
        qq-=long(SieveWidth) ;
        if( mod(a*qq*qq+2*b*qq+c,q)!=0 ) {
cout << endl << q << " " << " " << long(Primes[j].SievePos[1])<< " " << qq << " " << k << " " << j << "/"  << nPrimes;
            Error(not a root) ;
        }
        qq=long(Primes[j].SievePos[0])+long(q)  ;
        qq-=long(SieveWidth) ;
        if( mod(a*qq*qq+2*b*qq+c,q)!=0 ) {
            cout << endl << q << " " << " " << long(Primes[j].SievePos[0])<< " " << qq << " " << k ;
            Error(not a root) ;
        }
    }
}
#else
q=2 ;
for( j=2 ; j<nPrimes ; j++ ) {
    q+=(unsigned long)Primes[j].diff ;
    if( Primes[j].dmy==0 ) continue ;
unsigned long qq ;
qq=Primes[j].SievePos[0]    ;
qq+=(SieveWidth-SieveUnit) ;
    if( mod(a*qq*qq+2*b*qq+c,q)!=0 ) {
cout << endl << q << " " << qq << flush ;
        cout << endl << " k " << k ;
        Error(not a root) ;
    }
qq=Primes[j].SievePos[1]    ;
qq+=(SieveWidth-SieveUnit) ;
    if( mod(a*qq*qq+2*b*qq+c,q)!=0 ) {
cout << endl << q << " " << qq << flush ;
        cout << endl << " k " << k ;
        Error(not a root) ;
    }
}
#endif
#endif
        }
DISP: ;
#ifdef PROFILE_TEST
cpu_sieve_init.Start() ;
cpu_sieve_all.Start() ;
#endif
        // revcover parameters
        SieveDone(ncomb,Primes);
#ifdef PROFILE_TEST
cpu_sieve_init.Stop() ;
cpu_sieve_all.Stop() ;
#endif
        if( found    /* && kLPP>lLPP */ ) {     // show status
            found=false ;
#ifdef  WIN32
#  ifdef __BORLANDC__
        con.rdbuf()->gotoxy(1,posy) ;
        cout << endl    << nF << "("
         << int(nF+newFF+Success*(kLPP-lLPP)) << ")/" << nX << " ";
#  else
        gotoxy(1,posy) ;
        cout << endl    << nF << "("
         << int(nF+newFF+Success*(kLPP-lLPP)) << ")/" << nX << " "  ;

#  endif
#else
        cout << "\e[1A\e[1G" ;          // move cursor up by 1 line (for Linux)
                                        //J[\s̐擪ɂĂ
cout << endl    << nF << "("
         << int(nF+newFF+Success*(kLPP-lLPP)) << ")/" << nX << " "  << flush ;
#endif
                                         // kLPP : # of partially factored data
            if( nF+newFF+Success*(kLPP-lLPP)>nX ) {
cout << endl <<  "search cycles "  << flush   ;
//                lLPP=kLPP-lLPP ;
                // generate smooth data from large primes
                if( nX>nF ) {
                    newFF=LargePrime(nX-nF,kLPP) ;
#if 0
                    Success/=(1.0-0.4) ;         // modify estimation
                    while( Success>=0.2 ) Success*=0.7 ;
                    lLPP=kLPP ;
                    mLPP=kLPP ;
#endif
                    if( newFF>0 ) {
                        nF=MergeDuplicatedData(nF+newFF) ;
//                      newFF=0 ;
                    }
                    else {
                        Success/=(1.0+0.05) ;   // modify estimantion
                        newFF=-newFF ;
//                      nFF=nF ;
                    }

                }
                if( nF>nX ) break ;        // found enough data
                else {
cout << endl << (nF+newFF) << " data gained."
// << (nF+newFF )
         << " try more. " << flush ;
            newFF=0 ;
#ifndef WIN32
        cout << "\e[1A\e[1G" ;          // move cursor up to 1 line  (for Linux)
        cout << "\e[1A\e[1G" ;
#endif
                }

            }
        }
        if( isBreak() ) break ;

    }
    for( j=0 ; j<=ncomb ; j++ ) {
        delete [] Bmodp[j] ;
    }

#ifdef PROFILE_TEST
cout << endl << "count_of_a  " << count_of_a << " " ;
cout << endl << "sieve       " << cpu_sieve << " " ;
cout << endl << "sieve_init  " << cpu_sieve_init<< " "  ;
cout << endl << "sieve_shift " << cpu_shift << " "  ;
#endif

//return 1 ;



    posy++ ;
//con.rdbuf()->gotoxy(1,posy) ;
    cout << endl << "(f)  type " << nLPP[0] << " total " << nF ;
    posy++ ;
    cout << endl << "(p)  type " << nLPP[1]
         << "( p_p " << npp[0] << ", p_p_pp " << npp[1]
         << ", p_p_p_p_pp_pp " << npp[4] <<  ")" ;
    posy++ ;
    cout << endl << "(pp) type " << nLPP[2]
         << "( pp_pp " << npp[2] << ", pp_pp_pp " << npp[3] << ", others "<< npp[9] <<")" << flush;
//cout << endl << "enter LargePrime "<< enterLP ;
q=0 ;
for( j=0 ; j<10 ; j++ ) q += npp[j] ;
//posy++ ;

    if(  kLPP!=0 ) RealSuccess=double(nX-nLPP[0])/kLPP ;

{
    parms.open(parmfile,ios::out) ;

    parms << endl << N ;
    parms << endl << d ;
    parms << endl << nF ;
    parms << endl << (nLPP[1]+nLPP[2]) ;
    parms << endl << lLPP ;
    parms << endl << mLPP ;
    parms << endl << nLPP[0] ;
    parms << endl << nLPP[1] ;
    parms << endl << nLPP[2] ;
    cputime.Stop() ;
    parms << endl << cputime ;
    parms << endl << cache_size ;

    parms << endl << Success ;
    l=mpqsfx.tellp() ;
    parms << endl << l ;

    l=mpqsfxix.tellp() ;
    parms << endl << l ;

    l=mpqsppfx.tellp() ;
    parms << endl << l ;

    l=mpqsppp.tellp() ;
    parms << endl << l ;

    l=mpqspair.tellp() ;
    parms << endl << l ;

    parms << endl << SieveUnit  << endl ;
    parms << endl << base << " " << pbase ;
    parms << endl <<  ncomb << endl ;
    for( j=0 ; j<=ncomb ; j++ ) {
        parms << ip[j] << " " ;
    }
    parms << endl ;
        for( j=0 ; j<=ncomb ; j++ ) {
        parms << pq[j] << " " ;
    }
    parms << endl ;

    parms.close() ;
}

    mpqsppfx.close() ;
    mpqspair.close() ;
    mpqsppp.close() ;
    mpqsfxix.close() ;
    delete [] Primes ;
    if( nF<nX ) goto RET ;


    if( nX>MaxPrimes ) nX=MaxPrimes ;

    unsigned long* ifx ;                // file pointers to all smooth data
    ifx=new unsigned long[nX] ;

    mpqsfxix.open(ixfile,ios::binary|ios::in) ;
    if( mpqsfxix.bad() ) {
        cerr << endl << "Error : could not open file " << ixfile ;
        Error(file not found) ;
    }
    mpqsfxix.read((char*)ifx,nX*sizeof(unsigned long)) ;
    mpqsfxix.close() ;

    ReadPrimes(nPrimes,prime) ;
    SetMatrix(mpqsfx,ifx,nX,nF,nPrimes,prime) ;
    delete [] prime ;
    mpqsfx.close() ;

    long heavy_col ;
posy++ ;
///////////////////////////////////////////////////////////////////
//
// block Lanczos method
//

    cout << endl << endl << "block Lanczos method " << flush ;

RETRY_LANCZOS:
    CPULanczos.Reset() ;

    SparseMatrix* SM ;                         // sparse matrix
    unsigned long* v ;                         // bit vector
    elem.open(elemfile,ios::binary|ios::in) ;
    SM=new SparseMatrix(nX,nPrimes) ;          // make sparse matrix from elements file
    read(elem,*SM,nX,nPrimes) ;
    elem.close() ;

    v=new unsigned long[(nX+31)/32] ;          // allocate bit vector
    memset(v,0x00,(nPrimes+31)/32*4) ;         // set zero as vector
    V.SetBitString(nPrimes,v) ;

    unsigned long* old_row ;
    double Mreduction ;
    unsigned long nSingleton ;

    old_row=new unsigned long[SM->nrow] ;
    for( j=0 ; j<SM->nrow ; j++ ) old_row[j]=j ;
    nSingleton=SM->nrow ;
    SM->DeleteSingleton(old_row) ;             // delete singlton rows
    nSingleton-=SM->nrow ;

#ifdef NFS_DEBUG
    cout << endl << "post deleting singleton " ;
    SM->statics(cout) ;
#endif

    Mreduction=SM->NumberOfNonZeroEntry() ;

    unsigned long* new_col ;
    new_col=new unsigned long[nPrimes] ;

    memset(new_col,0x00,nPrimes*sizeof(unsigned long)) ;

    ULONG2* weight_index ;
    weight_index=new ULONG2[nPrimes] ;
    for( j=0 ; j<nPrimes ; j++ ) {
        weight_index[j][0]=SM->w_col[j] ;       // set column weight
        weight_index[j][1]=j ;                  // original column number
    }

    QSortULD(nPrimes,8,0,(unsigned long*)weight_index) ;
    for( k=0,j=nPrimes-1 ; ; j-- ) {
        if( weight_index[j][0]!=0 ) break ;
        new_col[weight_index[j][1]]=SparseMatrix::isLight ;  // column to be used
        k++ ;
    }
    cout << endl << "#singleton " << nSingleton << " : zero-weight columns " << (k-nSingleton) ;
    for( k=0 ;; j-- ) {
        if( weight_index[j][0]!=1 ) break ;   // s񂩂珜
        new_col[weight_index[j][1]]=SparseMatrix::isDelete ;    // column to be deleted
        k++ ;
    }

    if( k!=0 ) cout << ": deleted columns " << k ;

    unsigned long denseColumn ;
#if 0
    if( SM->memory()>16*1024 ) denseColumn=20 ;
    else denseColumn=0 ;
#endif
    denseColumn=0 ;

    // delete some columns with heavy weight to save memory
    // but denseColumn=0 now
    for( j=0 ; j<denseColumn ; j++ ) {
        new_col[weight_index[j][1]]=SparseMatrix::isHeavy ; // columns with many non-zero entries
    }
    delete [] weight_index ;

    for( k=0,j=0 ; j<nPrimes ; j++ ) {
        if( new_col[j]<SparseMatrix::isDelete ) {
            new_col[j]=k ;
            k++ ;
        }
    }
    unsigned long delcol ;
    delcol=SM->DeleteColumn(new_col) ;
    Mreduction=Mreduction-delcol ;


    for( k=0,j=0 ; j<nPrimes ; j++ ) {
        if( new_col[j]==SparseMatrix::isHeavy ) {  // column with many non-zero entries
            new_col[j]=k ;
            k++ ;
        }
        else new_col[j]=SparseMatrix::isHeavy ;
    }
    heavy_col=k ;     // number of columns removed from matrix (= denseColumn )

    mpqs_col.open(mpqscol,ios::binary|ios::out) ;
    mpqs_col.write((char*)&nPrimes,sizeof(unsigned long)) ;
    mpqs_col.write((char*)new_col,nPrimes*sizeof(unsigned long)) ;
    mpqs_col.close() ;
    delete [] new_col ;

    SM->DeleteNullRow(old_row) ;       // delete rows without non-zero entry

    mpqs_row.open(mpqsrow,ios::binary|ios::out) ;
    mpqs_row.write((char*)&(SM->nrow),sizeof(unsigned long)) ;
    mpqs_row.write((char*)old_row,(SM->nrow)*sizeof(unsigned long)) ;
    mpqs_row.close() ;

    mpqs_red.open(mpqsred,ios::binary|ios::out) ;
    write(mpqs_red,*SM) ;             // write reduced matrix
    mpqs_red.close() ;
    mpqs_red.open(mpqsred,ios::binary|ios::in) ;

    // reload reduced matrix
#ifdef LANCZOS_TRANSPOSE
    read(mpqs_red,*SM,SM->nrow,SM->ncol) ;
#else
    readTransposed(mpqs_red,*SM,SM->ncol,SM->nrow) ;
#endif

    mpqs_red.close() ;
    unlink(mpqsred) ;
    delete [] SM->w_col ;   // column weight are not needed any more
    SM->w_col=NULL ;

    Lanczos* L ;
    L=new Lanczos(SM) ;     // make Lanczos scheme

    unsigned long Nentry ;
    double entry_per_row;

    unsigned long SMatrixMemory ;

#ifdef LANCZOS_TRANSPOSE
    SMatrixMemory=L->BT->memory() ;          // memeory size of matrix
    Nentry=L->BT->NumberOfNonZeroEntry() ;   // number of non-zero entries
    entry_per_row=L->BT->entry_per_row() ;
    Mreduction=Nentry/Mreduction*100.0 ;
    L->BT->statics(cout) ;
#else
    SMatrixMemory=L->B->memory() ;
    Nentry=L->B->NumberOfNonZeroEntry() ;
    entry_per_row=L->B->entry_per_row() ;
    Mreduction=Nentry/Mreduction*100.0 ;
    (L->B)->statics(cout) ;
#endif
    if( denseColumn!=0 ) {
        cout << " " << Mreduction << "% compressed " ;
    }

    unsigned long LanczosLoop ;
    unsigned long msol ;
    unsigned long reducedFF ;
    unsigned long nrel ;
    // search non-trivial relations

    CPULanczos.Start() ;
    for( LanczosLoop=1 ;; LanczosLoop++ ) {
        msol=L->Solve(sol) ;            // search solutions s.t. B*X=0  B=A*A^T
        cout << endl ;
        L->statics(cout) ;
        cout <<  " gained " << msol << " solutions " ;
        if( msol>heavy_col) break ;
    }
    CPULanczos.Stop() ;
cout << " " << CPULanczos ;

#ifdef LANCZOS_TRANSPOSE
    reducedFF=L->BT->nrow ;
#else
    reducedFF=L->B->ncol ;
#endif
    delete L ;
    delete SM ;
    nrel=sol.size() ;


    mpqs_ans.open(ansfile,ios::binary|ios::out) ;
    mpqs_ans.write((char*)&reducedFF,sizeof(unsigned long)) ;
    for( j=0 ; j<nrel ; j++ ) {
        mpqs_ans.write((char*)sol[j],((reducedFF+31)/32)*sizeof(unsigned long)) ;
    }
    mpqs_ans.close() ;
    // 擪 msol ̌ ق͕̂̂KvȂ
    while( msol<sol.size() ) {
        delete [] sol.back() ;
        sol.pop_back() ;
    }

    unsigned long** QcMat ;
    QcMat=(unsigned long**)SieveArray ;
    unsigned long** PseudoSol ;
    PseudoSol=QcMat+1024 ;
    for( j=0 ; j<msol ; j++ ) {
        QcMat[j]=new unsigned long[(heavy_col+31)/32] ;
        memset(QcMat[j],0x00,(heavy_col+31)/32*4) ;
    }


    U.SetBitString(heavy_col,v) ;
    mpqs_col.open(mpqscol,ios::binary|ios::in) ;
    mpqs_col.read((char*)&l,sizeof(unsigned long)) ;
    new_col=new unsigned long[nPrimes] ;
    mpqs_col.read((char*)new_col,nPrimes*sizeof(unsigned long)) ;
    mpqs_col.close() ;

    elem.open(elemfile,ios::binary|ios::in) ;

    // reconstruct matrix from all columns
    // s񂩂珜 column gݓ
    unsigned long irow ;
    unsigned long n ;
    for( irow=0,j=0 ; j<reducedFF ; j++ ) {
        memset(v,0x00,(heavy_col+31)/32*4) ;
        for(;; irow++ ) {
            elem.read((char*)&n,sizeof(unsigned long)) ;  // # non-zero entries
            if( old_row[j]==irow ) {
                irow++ ;
                break ;
            }
            for( k=0 ; k<n ; k++ )
                elem.read((char*)&l,sizeof(unsigned long)) ;  // nonzero entry number
        }
        for( k=0 ; k<n ; k++ ) {
            elem.read((char*)&l,sizeof(unsigned long)) ;  // nonzero entry number
            if( new_col[l]<SparseMatrix::isDelete ) {
                U.BitOn(new_col[l]) ;
            }
        }
        for( k=0 ; k<msol ; k++ ) {
            if( BitIndex(sol[k],j) ) {
                V.SetBitString(heavy_col,QcMat[k]) ;
                if( heavy_col!=0 ) V^=U ;
            }
        }
    }
    delete [] old_row ;
    elem.close() ;

    nrel=sol.size() ;
    for( j=0 ; j<msol ; j++ ) PseudoSol[j]=sol[j] ;

    unsigned long nsol ;
    // find real dependencies
    // create solutions of A^T*X=0 from solutions of (A*A^T)*X=0
    if( heavy_col!=0 ) nsol=Gauss(heavy_col,reducedFF,msol,QcMat,PseudoSol) ;
    else nsol=0 ;

#if 0
cout << endl << "Rank " << nsol ;
for( j=nsol  ; j<sol.size() ; j++ ) {
  V.SetBitString(heavy_col,QcMat[j]) ;
   cout << endl << V ;
}
#endif

    for( j=0 ; j<sol.size() ; j++ ) delete [] QcMat[j] ;
    // write solutions to file
    mpqs_ans2.open(ansfile2,ios::binary|ios::out) ;
    mpqs_ans2.write((char*)&reducedFF,sizeof(unsigned long)) ;

    for( j=0 ; j<sol.size() ; j++ ) sol[j]=PseudoSol[j] ;
    reverse(sol.begin(),sol.end()) ;
    for( j=0 ; j<nsol ; j++ ) {
        delete [] sol.back() ;
        sol.pop_back() ;
    }

    nsol=sol.size() ;


#ifdef MPQS_CHECK_SOLUTION                 // check solutions
    {
        int j,jj ;
        for( k=0,j=0 ; j<nPrimes ; j++ ) {
            if( new_col[j]!=SparseMatrix::isDelete ) {
                new_col[j]=k ;
                k++ ;
            }
        }
        mpqs_row.open(mpqsrow,ios::binary|ios::in) ;
        mpqs_row.read((char*)&j,sizeof(unsigned long)) ;
        old_row=new unsigned long[reducedFF] ;
        mpqs_row.read((char*)old_row,reducedFF*sizeof(unsigned long)) ;
        mpqs_row.close() ;
        V.SetBitString(nPrimes,(unsigned long*)SieveResult) ;
        U.SetBitString(nPrimes,v) ;
        unsigned long m ;
        for( l=0 ; l<nsol ; l++ ) {
            elem.open(elemfile,ios::binary|ios::in) ;
            memset(SieveResult,0x00,(nPrimes+31)/32*4) ;
            for( irow=0,jj=0 ; jj<reducedFF ; jj++ ) {
                memset(v,0x00,(nPrimes+31)/32*4) ;             // clear U
                for( ;; irow++ ) {
                    elem.read((char*)&n,sizeof(unsigned long)) ;        // # entries
                    if( old_row[jj]==irow ) {
                        irow++ ;
                        break ;
                    }
                    else {
                        for( k=0 ; k<n ; k++ ) {
                            elem.read((char*)&m,sizeof(unsigned long)) ;    // nonzero entry index
                        }
                    }
                }
                for( k=0 ; k<n ; k++ ) {
                    elem.read((char*)&m,sizeof(unsigned long)) ;    // nonzero entry index
                    if( new_col[m]!=SparseMatrix::isDelete ) {
                        U.BitOn(new_col[m]) ;
                    }
                }
                if( BitIndex(sol[l],jj) ) V^=U ;
            }
            elem.close() ;
            for( k=0 ; k<(nPrimes+31)/32 ; k++ ) {
                if( SieveResult[k]!=0 ) {
//                  cout << endl << V ;
                    delete [] sol[l] ;
                    sol[l]=NULL ;
                    Error(why) ;
                    cout << k << "why " ;
                    break ;
                }
            }
        }
        delete [] old_row ;
    }
#endif


    delete [] new_col ;
    j=nsol ;
    unsigned long realsol ;
    for( realsol=0,k=0 ; k<sol.size() ; k++ ) {
        if( sol[k] ) realsol++ ;
    }
    nsol=realsol ;
    mpqs_ans2.write((char*)&nsol,sizeof(unsigned long)) ;
    for( j=0 ; j<sol.size() ; j++ ) {
        if( sol[j] ) mpqs_ans2.write((char*)sol[j],((reducedFF+31)/32)*sizeof(unsigned long)) ;
    }
    mpqs_ans2.close() ;
    for( j=0 ; j<sol.size() ; j++ ) {
        if( sol[j] ) delete [] sol[j] ;
    }

    while( !sol.empty() ) sol.pop_back() ;

    sol.resize(0) ;

    mpqs_row.open(mpqsrow,ios::binary|ios::in) ;
    mpqs_row.read((char*)&j,sizeof(unsigned long)) ;
    old_row=new unsigned long[reducedFF] ;
    mpqs_row.read((char*)old_row,reducedFF*sizeof(unsigned long)) ;
    mpqs_row.close() ;

    mpqs_ans2.open(ansfile2,ios::binary|ios::in) ;
    mpqs_ans2.seekp(0) ;
    mpqs_ans2.read((char*)&reducedFF,sizeof(unsigned long)) ;
    mpqs_ans2.read((char*)&nsol,sizeof(unsigned long)) ;

    lbyte=((reducedFF+31)>>5)<<2 ;

////////////////////////////////////////////////////////

    mpqsfxix.open(ixfile,ios::binary|ios::in) ;
    mpqsfxix.read((char*)ifx,nX*sizeof(long)) ;
    mpqsfxix.close() ;

    mpqsfx.open(fxfile,ios::binary|ios::in) ;

    V.SetBitString(reducedFF,v) ;

#ifdef  WIN32
#ifdef __BORLANDC__
posy=con.rdbuf()->wherey() ;
#else
posy=wherey() ;
#endif
#endif

#ifdef SIQS_DEBUG
//  mpqsfx.open(fxfile,ios::binary|ios::in) ;
    ReadPrimes(nPrimes,prime) ;
    unsigned long* iindex ;
    iindex=new unsigned long[nPrimes] ;
    for( j=0 ; j<reducedFF ; j++ ) {
        mpqsfx.seekp(ifx[old_row[j]]) ;
        read(mpqsfx,_fx) ;
        read(mpqsfx,sq) ;
        if( !PrimeDecomp(nPrimes-1,prime+1,iindex+1,_fx) ){
          Error(PrimeDecomp error) ;
        }
    }
//  mpqsfx.close() ;
    for( j=0 ; j<reducedFF ; j++ ) {
      cout << " " << old_row[j] ;
    }
    delete [] iindex ;
    delete [] prime ;
#endif


//////////////////////////////////////////
//
//  final stage ( find a non-trivial factor )
//

    // set factor base
    ReadPrimes(nPrimes,prime) ;
    index=new unsigned long[nX] ;
    for( l=0 ; l<nsol ; l++ ) {
        mpqs_ans2.read((char*)v,lbyte) ;   // read a solution ( bit vector )
        x=1 ;

        for( k=0 ; k<nPrimes ; k++ ) index[k]=0 ;
        for( k=0 ; k<reducedFF ; k++ ) {
            if( V[k] ) {
                mpqsfx.seekp(ifx[old_row[k]]) ;
                read(mpqsfx,_fx) ;
                read(mpqsfx,sq) ;                    // square part
                (x*=sq)%=N  ;                        // x=mod(x*sq,pN)
                if( _fx<0 ) ~_fx ;
                                        // primary decomposition by factor base
                if( !PrimeDecomp(nPrimes-1,prime+1,index+1,_fx) ){
#if 1
                    break ;
#else
                    Error(Prime decomposition error) ;
#endif
                }
            }
        }
        for( y=1,k=1 ; k<nPrimes ; k++ ) {
            p=prime[k] ;
            if( index[k]!=0 ) {

#ifdef SIQS_DEBUG
if( index[k]&1 ) Error(not square) ;
#endif
                q=index[k]>>1 ;
                y*=power(sq=p,_fx=q,N) ; // multiply by p^q mod N
                y%=N ;
            }
        }

        d=gcd(x-y,N) ;
        if( d!=1 && d!=N && d!=0 ) goto RET2 ; // find a factor
        cout << endl << "failed " ;
        cout << " gcd " << d << " " ;
    }

  delete [] prime ;                       // failed to find a factor
  delete [] index ;
  delete [] old_row ;
  delete [] v ;
  mpqs_ans2.close() ;
  mpqsfx.close() ;
  goto RETRY_LANCZOS ;

RET2:
  delete [] prime ;
  delete [] index ;
  delete [] ifx ;
  delete [] old_row ;
  delete [] v ;

RET: ;
    mpqs_ans2.close() ;
#if 0
    mpqsfx.seekp(0) ;
    for( j=0 ; j<nF ; j++ ) {
        read(mpqsfx,_fx) ;
        read(mpqsfx,_fx) ;
    }
#endif
    mpqsfx.close() ;
    if( nF>nX ) {
        cout << endl << "found  factor " << d ;
        cout << endl << "cputime " << cputime ;

        mpqslog.open(errfile,ios::out|ios::app) ;
        mpqslog << endl << " ========================= by SIQS " ;
        mpqslog << endl << "#FB " << nX ;
        mpqslog << endl << "(f)  type " << nLPP[0] << " total " << nF ;
        mpqslog << endl << "(p)  type " << nLPP[1]
         << "( p_p " << npp[0] << ", p_p_pp " << npp[1]
         << ",p_p_p_p_pp_pp " << npp[4] << ")" ;
        mpqslog << endl << "(pp) type " << nLPP[2]
             << "( pp_pp " << npp[2] << ", pp_pp_pp " << npp[3] << ",others " << npp[9] << ")" ;
        N/=d ;

        char PC[2] ;

        mpqslog <<  endl << input_str <<" = " ;
        if( d>N ) Swap(d,N) ;                            // sort
        // primality test
        if( APRCL(d) ) { PC[0]='P' ; mpqslog << "P" ; }
        else         { PC[0]='C' ; mpqslog << "C" ; }
        mpqslog << len10(d) << " * " ;
        if( APRCL(N) ) { PC[1]='P' ; mpqslog << "P" ; }
        else           { PC[1]='C' ; mpqslog << "C" ; }
        mpqslog << len10(N)  ;

        mpqslog << endl << PC[0] << len10(d) <<" = " << d ;
        mpqslog << endl << PC[1] << len10(N) <<" = " << N ;

        mpqslog << endl << "cputime " << cputime ;
#ifdef PROFILE_TEST
mpqslog << endl <<"select_a      " << cpu_sela ;
mpqslog << endl <<"sieve_all     " << cpu_sieve_all ;
mpqslog << endl <<"sieve         " << cpu_sieve ;
mpqslog << endl <<"sieve_init    " << cpu_sieve_init ;
mpqslog << endl <<"sieve_shift   " << cpu_shift ;
mpqslog << endl <<"try_div       " << cpu_try_div ;
mpqslog << endl << "#curve       " << count_of_a ;
mpqslog << endl << "#combination " << ncomb ;
    if( len(pN)<266 ) (Success=(0.15000-0.00019*len(pN)))/=1.05 ;
//  if( len(pN)<256 ) Success=0.825*(0.15000-0.00019*len(pN)) ;
//  if( len(pN)<256 ) Success=(0.15000-0.00019*len(pN))/1.4 ;
    else (Success=0.11450-0.000143*len(pN))/=1.1 ;

mpqslog << endl << "gessed success   " << Success ;
mpqslog << endl << "real success     " << RealSuccess  << " " << len(pN) ;
mpqslog << endl << "real success/len " << RealSuccess/len(pN) ;
#endif
#ifdef PROFILE_TEST
//mpqslog << endl << "ndiv " << ndiv ;
#endif
nX=nLPP[0] ;
nX+=nLPP[1] ;
nX+=nLPP[2] ;
//unsigned long count_of_a(0) ;
#ifdef PROFILE_TEST
mpqslog << endl << "parm   " << cpu_sieve_init.Time()/count_of_a ;
mpqslog << endl << "sieve  " << cpu_sieve.Time()/count_of_a/SieveWidth ;
mpqslog << endl << "data   " << double(nX)/count_of_a/SieveWidth ;
mpqslog << endl << "duplicate  " << duplicate ;
mpqslog << endl << "cycle  " ;
for( j=2 ; j<=MaxCycle ; j++ ) {
    if( cycles[j]==0 ) break ;
    if( (j-2)%5==0 ) mpqslog << endl ;
    mpqslog << j << " : " << cycles[j] << " " ;
}
#endif

//mpqslog << endl <<"set_table " << cpu_table ;
//mpqslog << endl <<"shift " << cpu_shift ;
#ifdef PROFILE_TEST
mpqslog << endl << "Width Unit " << SieveWidth << " " << SieveUnit ;
#endif
//mpqslog << endl <<    "Lnaczos " << CPULanczos ;
#ifdef PROFILE_TEST
cout << endl <<"sieve_all  " << cpu_sieve_all ;
cout << endl <<"sieve      " << cpu_sieve ;
cout << endl <<"sieve_init " << cpu_sieve_init ;
cout << endl <<"sieve_shift   " << cpu_shift ;
cout << endl << "parm  " << cpu_sieve_init.Time()/count_of_a ;
cout << endl << "sieve " << cpu_sieve.Time()/count_of_a/SieveWidth ;
cout << endl << "data  " << double(nX)/count_of_a/SieveWidth ;
cout << endl << "#curve " << count_of_a ;
cout << endl <<"try_div " << cpu_try_div ;
#endif
#ifdef PROFILE_TEST
cout << endl << "ndiv " << ndiv ;
#endif

//cout << endl <<"set_tabke " << cpu_table ;
//cout << endl <<"shift " << cpu_shift ;
        mpqslog.close() ;
        delete [] input_str ;
        goto Input_Number ;
    }
    else {
        delete [] input_str ;
    }

    return 0 ;

}

///////////////////////////////////////////
//
//  merge duplicate data
//
//  ꂽ (f) ^f[^ Ɋւă\[g
//
static long MergeDuplicatedData(long nF)
{
    LONG3* fp=NULL ;
    long j,k ;
    long fxpos ;

    fp=new LONG3[nF] ;
    if( !fp ) Error(allocate error) ;
    fxpos=mpqsfx.tellp() ;
    mpqsfxix.seekp(0) ;
    for( j=0 ; j<nF ; j++ ) {
        mpqsfxix.read((char*)&(fp[j][2]),sizeof(long)) ;     // file pointer
        mpqsfx.seekp(fp[j][2]) ;
        read(mpqsfx,_fx) ;
        fp[j][1]=(unsigned long)_fx ;       // fx ̈ꌅڂL[ɂ
        fp[j][0]=len(_fx) ;                         //  2 iWJł̒
    }
    QSortL2A(nF,12,0,4,(long*)fp) ;

    for( k=0,j=1 ; j<nF ; j++ ) {           // df[^̂Ă
        if( fp[j-1][1]==fp[j][1] ) {
            mpqsfx.seekp(fp[j-1][2]) ;
            read(mpqsfx,_fx) ;
            if( sign(_fx)<0 ) ~_fx ;
            mpqsfx.seekp(fp[j][2]) ;
            read(mpqsfx,_t) ;
            if( sign(_t)<0 ) ~_t ;
            if( _fx==_t ) {                      // ꌅڂv̂őŜr
                 k++ ;
                 fp[j-1][0]=0 ;                 // discard
            }
        }
    }
    if( k>0 ) {
//posy++ ;
        cout << endl << k << " duplicated data merged " << flush ;
//      QSortL2A(nF,12,0,8,(long*)fp) ;
    }

//  QSortL2A(nF-k,12,0,8,(long*)fp+3*k) ;    // sort on length
    QSortL2A(nF,12,0,8,(long*)fp) ;  // sort on length
#if 1
    mpqsfxix.seekp(0) ;
    for( j=k ; j<nF ; j++ )              // }[W̌ʁA\[ǧʂo
        mpqsfxix.write((char*)&fp[j][2],sizeof(long)) ;
#endif

    if( fp!=NULL ) delete [] fp ;
    mpqsfx.seekp(fxpos) ;
    return( nF-k ) ;

}

///////////////////////////////////////////
//
//  generate (F)-type data
//    from (P)-type data and (PP)-type data
//
static long LargePrime(long needF,long& nLPP)
{
    Mint fx ;
    LONG3* ppp3 ;
    long* ppp ;
    long j,k,l,l1,l2 ;
    long k0(0),k1 ;
    unsigned long p1,p2 ;
    long newFF(0) ;
    long ifp[MaxCycle+100] ;

    if( nLPP==0 ) return 0 ;
    ppp3=new LONG3[nLPP] ;
    if( !ppp3 ) Error( ppp3 allocate error) ;

    ppp=(long*)ppp3 ;
    long ppfxpos ;

    ppfxpos=mpqsppfx.tellp() ;
    long ppppos=mpqsppp.tellp() ;

    mpqsppp.seekp(0) ;
    mpqsppp.read((char*)ppp,12*nLPP) ;
#ifdef PROFILE_TEST
    memset(cycles,0x00,sizeof(cycles)) ;
#endif

#if 1
         // sort on first key p2 and second key p1
    QSortL2A(nLPP-k0,12,4,0,ppp+3*k0) ;
    do {
        for( k=0,j=k0+1 ; j<nLPP ; j++ ) {    // discard duplicate data
            if( ppp3[j][0]==ppp3[j-1][0] && ppp3[j][1] ==ppp3[j-1][1] ) {

                l1=ppp3[j-1][2] ;
                if( l1<0 ) l1=-l1 ;
                else continue ;         // cannot discard because this data is already combined.

                l2=ppp3[j][2] ;
                if( l2<0 ) l2=-l2 ;
                else continue ;        // same as above

                mpqsppfx.seekp(l1) ;
                mpqsppfx.read((char*)&p1,sizeof(p1)) ;
                mpqsppfx.read((char*)&p1,sizeof(p1)) ;
                read(mpqsppfx,_fx) ;
                if( sign(_fx)<0 ) ~_fx ;

                mpqsppfx.seekp(l2) ;
                mpqsppfx.read((char*)&p1,sizeof(p1)) ;
                mpqsppfx.read((char*)&p1,sizeof(p1)) ;
                read(mpqsppfx,_t) ;
                if( sign(_t)<0 ) ~_t ;
//cout << endl << _fx << " " <<ppp3[j][0] << " " << ppp3[j][1] << " " << l1 << " " << l2 ;
                if( _fx==_t ) {
                    ppp3[j-1][1]=0 ;
                    k++ ;
                }
            }
        }
        if( k>0 ) {
            QSortL2A(nLPP-k0,12,4,0,ppp+3*k0) ;
            k0 += k ;
//posy++ ;
        }
    } while( k!=0 ) ;
    if( k0>0 ) cout << endl << k0 << " duplicated data merged in Large Prime" ;
#ifdef PROFILE_TEST
    duplicate=k0 ;
#endif
#endif
    k1=k0 ;

    for( l1=0 ; l1<MaxSearchLoop ; l1++ ) {
        for( j=k0 ; j<nLPP-1 ; j++ ) {
            if( ppp3[j][1]<=1 ) continue ;      // already combined
            if( ppp3[j+1][1]==ppp3[j][1] ) {    // try to comibine

                l=mpqspair.tellp() ;
                                     // indices of combined data
                mpqspair.write((char*)(ppp+3*j+2),sizeof(long)) ;
                mpqspair.write((char*)(ppp+3*(j+1)+2),sizeof(long)) ;
                ppp3[j][2]=l ;   // index to combination file
                ppp3[j][1]=ppp3[j+1][0] ;
                if( ppp3[j][0]==ppp3[j][1] ) ppp3[j][1]=1 ;
            }
        }

        QSortL2A(nLPP-k0,12,4,0,ppp+3*k0) ;

        if( ppp3[k0][1]!=1 ) break ;         // new pairing is not found any more
        else {                               // repeat algorithm
            for( j=k0 ; j<nLPP ; j++ ) {
                if( ppp3[j][1]>1 ) { k0=j ; break ; }
            }
        }

    }

    long pairpos=mpqspair.tellp() ;          // save file pointer

    int is,ie ;
    int pptype ;


    for( j=k1 ; j<k0 ; j++ ) {
        is=ie=0 ;                            // make node list empty
        if( ppp3[j][2]>0 ) {
            mpqspair.seekp(ppp3[j][2]) ;     // see combination file
            mpqspair.read((char*)&l1,sizeof(l1)) ;
            mpqspair.read((char*)&l2,sizeof(l2)) ;
            ifp[ie++]=l1 ;
            ifp[ie++]=l2 ;
        }
        else ifp[ie++]=ppp3[j][2] ; // maybe error
                                    // combination file ɂ͑Ή鍀͂Ȃ
                                    // Ƃ͂Ȃ͂.
                                    // ,p1=p2=1 ^̂̂܂߂Ƃ
                                    // ꍇ
        for( k=is ; k<ie ; ) {      // envelope nodes to real nodes
            if( ifp[k]<=0 ) k++ ;   // this node is a leaf.
            else {                  // this node has children.
                mpqspair.seekp(ifp[k])  ;   // add children to node list
                mpqspair.read((char*)&l1,sizeof(l1)) ;
                mpqspair.read((char*)&l2,sizeof(l2)) ;
                ifp[k]=l1 ;         // replace parent node by its child
                ifp[ie++]=l2 ;      // add another child to node list
            }
            if( ie>MaxCycle ) break ;  //  long cycle
        }
        if( ie>MaxCycle ) continue ;  // discard long cycle
        if( --needF<0 ) break ;
        newFF++ ;
    }
    if( needF>0 ) {                 // enough data have not been gained.
        mpqsppfx.seekp(ppfxpos) ;   // reset file pointer
        mpqspair.seekp(pairpos) ;
        mpqsppp.seekp(ppppos) ;
        delete [] ppp3 ;
        return -newFF ;             // try to sieve more
    }
    else {
        newFF=0 ;
    }


    for( j=k1 ; j<k0 ; j++ ) { // generate f-type from p-type and pp-type
        is=ie=0 ;              // make node list  empty
        pptype=0 ;
        if( ppp3[j][2]>0 ) {   // combination data
            mpqspair.seekp(ppp3[j][2]) ;     // see combination file
            mpqspair.read((char*)&l1,sizeof(l1)) ;
            mpqspair.read((char*)&l2,sizeof(l2)) ;
            ifp[ie++]=l1 ;
            ifp[ie++]=l2 ;
        }
        else ifp[ie++]=ppp3[j][2] ;

        for( k=is ; k<ie ; ) {
            if( ifp[k]<=0 ) k++ ;
            else {                  // in case combination data
                mpqspair.seekp(ifp[k])  ;
                mpqspair.read((char*)&l1,sizeof(l1)) ;  // read its children
                mpqspair.read((char*)&l2,sizeof(l2)) ;
                ifp[k]=l1 ;                        // replace by child
                ifp[ie++]=l2 ;
            }
            if( ie>MaxCycle ) break ;
        }
        if( ie>MaxCycle ) continue ;
        pp=1 ;
        _fx=1 ;
        sq=1 ;                           //  (ss'..)^2*(pp'..)^-2=ff'.../(pp'..) mod pN

        for( k=0 ; k<ie ; k++ ) {
            if( ifp[k]>0 ) continue ;

            mpqsppfx.seekp(-ifp[k]) ;
            mpqsppfx.read((char*)&p1,sizeof(p1)) ;
            if( p1!=1 ) pptype++ ;
            pp*=p1 ;                                 // product of large primes
            mpqsppfx.read((char*)&p2,sizeof(p2)) ;
            pp*=p2 ;                                 // product of large primes pp'..
            read(mpqsppfx,_t) ;           // value of quadaratic polynomial f
            _fx*=_t ;                     // product of f
            if( p1!=1 ) {                 // divide by large prime
                _fx/=p1 ;
            }
//cout << endl << _fx ;
            _fx/=p2 ;
            read(mpqsppfx,_t) ;           // square part
            (sq*=_t)%=pN    ;             // ss'..
        }
        if( ie<=0 ) continue ;
        pp=mod(isqrt(pp),pN) ;
        sq*=invmodp(pp,pN) ;
        sq=mod(sq,pN) ;                  // product by inverse of square part
#ifdef SIQS_DEBUG
{
        unsigned long nPrimes ;
        unsigned long* primes ;
        ReadPrimes(nPrimes,primes)  ;
        fx=_fx ;
        if( !BitDecomp(nPrimes,primes,fx,v2) ) {
            BitDecomp(nPrimes,primes,fx,v2) ;
            cout << endl << "Bit Decomp Error in Large Prime Variation " ;
            delete [] primes ;
            Error(de) ;
            continue ;
        }
        delete [] primes ;
}
#endif


        l=mpqsfx.tellp() ;
        mpqsfxix.write((char*)&l,sizeof(l)) ;

        write(mpqsfx,_fx) ;         // _fx is fully factored data
        write(mpqsfx,sq) ;

        if( pptype<10 ) npp[pptype]++ ;
        else npp[9]++ ;
#ifdef PROFILE_TEST
        cycles[ie]++ ;
#endif

        newFF++ ;
    }

    if( k0>0 )  {

        mpqsppp.seekp(0) ;
                               // write data which are not combined.
        mpqsppp.write((char*)(ppp+3*k0),12*(nLPP-k0)) ;

        nLPP-=k0 ;
    }

    mpqsppfx.seekp(ppfxpos) ;    // reset file pointer
    mpqspair.seekp(pairpos) ;

    delete [] ppp3 ;            // delete work triplets

    return newFF ;
}

////////////////////////////////////////
//
//  make matrix
//
void SetMatrix(::fstream& mpqsfx,unsigned long* ifx,     // (I) file pointer
               unsigned long nX,                         // (I) number of relations to be needed
               unsigned long nF,                         // (I) number of smooth data
               unsigned long nPrimes,                    // (I) size of factor base
               unsigned long* prime)                     // (I) primes of factor base
{
    unsigned long* warray=new unsigned long[nPrimes] ;   // allocate work array
    unsigned long j,k,nbit,ir ;
    Mint fx ;
    TBit V ;
    unsigned long* v=new unsigned long[(nPrimes+31)/32] ; // bit vector

    V.SetBitString(nPrimes,v) ;

    elem.open(elemfile,ios::binary|ios::out) ;
    elem.seekp(0) ;
    for( ir=j=0 ; j<nF && ir<nX ; j++ ) {
        mpqsfx.seekp(ifx[j]) ;
        read(mpqsfx,fx) ;

        if( !BitDecomp(nPrimes,prime,fx,v) ) {   // decompose by the factor base
#if 1
            Error(decomp error in SetMatix) ;
#else
            mpqsfx.seekp(ifx[j]) ;
            read(mpqsfx,fx) ;
            mpqsfx.seekp(ifx[j]) ;
            read(mpqsfx,fx) ;
            continue ;
#endif
        }
        for( nbit=0,k=0 ; k<nPrimes ; k++ ) {
            if( V[k] ) warray[nbit++]=k ;       // set indices with non-zero bit
        }
        ir++ ;
                                                // write indices of non-zero entries
        elem.write((char*)&nbit,sizeof(unsigned long)) ;   // number of non-zero bits
        if( nbit!=0 ) elem.write((char*)warray,nbit*sizeof(unsigned long)) ; // write elements
    }

    elem.close() ;

    delete [] warray ;
    delete [] v ;
}


#if 1
////////////////////////////////////////////
//
//  Search coefficient of a qudaratic polynomial
//
//
bool select_a(Mint&a,Mint& mina,Mint& maxa,Mint& N,unsigned long ncomb,unsigned long ment)
{
    unsigned long r ;
    int j ;
    int k ;
    int l ;
    int i ;
#if 0
    cout << endl << endl ;
    for( j=0 ; j<ncomb ; j++ ) {
        cout << "\t" << ip[j] ;
    }
    cout << endl  ;
    for( j=0 ; j<ncomb ; j++ ) {
        cout << "\t" << gap[j+1] ;
    }
#endif
#if 0
    unsigned long wpq[MAX_COMB] ;
    memcpy(wpq,pq,ncomb*sizeof(unsigned long)) ;
#endif

//  for(;;) {
        for( k=0 ; k<COMB_DIFF ; k++ ) {
            for( j=1 ; j<ncomb ; j++ ) {
//              if( ip[j-1]<ip[j]-1 ) {
                if( gap[j]>step[j-1] ) {
                    for( l=0 ; l<step[j-1] ; l++ ) {
//                        a/=pq[j-1] ;
                        ip[j-1]++ ;
                        pq[j-1]+=(unsigned long)Primes[ip[j-1]].diff ;
//                        a*=pq[j-1] ;
                        gap[j]-- ;
                        gap[j-1]++ ;
                    }
#if 0
cout << endl << endl << "inc " << j ;
#endif
#if 0
if( a>maxa ) {
    return false ;
    j=ncomb+1 ;
}
#endif
    #if 1
                    if( j!=1 )  {
                        ip[0]=base ;
//    pbase+=(unsigned long)Primes[ip[0]].diff ;
                        pq[0]=pbase ;
                    }
                    {
                        for( i=1 ; i<j-1 ; i++ ) {
                            gap[i]=1 ;
                            ip[i]=ip[i-1]+1 ;
                            pq[i]=pq[i-1]+(unsigned long)Primes[ip[i]].diff ;
                        }
                        gap[j]=ip[j]-ip[j-1] ;
                    }
//  gap[ncomb]=ip[ncomb]-ip[ncomb-1] ;
                    a=pq[0] ;
                    for( i=1 ; i<ncomb ; i++ ) a*=pq[i] ;
//                  j=0 ;
    #endif


                    break ;
                }
            }
#if 0
            if( j==ncomb-1 ) {
                if( gap[j+1]>step[ncomb-1] ) {
                    for( l=0 ; l<step[ncomb-1] ; l++ ) {
                        a/=pq[j-1] ;
                        ip[j-1]++ ;
                        pq[j-1]+=(unsigned long)Primes[ip[j-1]].diff ;
                        a*=pq[j-1] ;
                        gap[j+1]-- ;
                        gap[j]++ ;
                    }

                }
                else j=ncomb ;
            }
#endif
            if( j==ncomb ) {
                if( ip[ncomb-1]<ment-1 ) {
                    for( l=0 ; l<nmachine ; l++ ) {     // Ul
                        for( i=0 ; i<step[ncomb-1] ; i++ ) {
                            ip[ncomb-1]++ ;
                            pq[ncomb-1]+=(unsigned long)Primes[ip[ncomb-1]].diff ;
                            gap[ncomb]-- ;
                            gap[ncomb-1]++ ;
//                          ip[ncomb-1]++ ;
//                          pq[ncomb-1]+=(unsigned long)Primes[ip[ncomb-1]].diff ;
                        }
                    }
//    base++ ;
                    ip[0]=base ;
//    pbase+=(unsigned long)Primes[ip[0]].diff ;
                    pq[0]=pbase ;
                    for( j=1 ; j<ncomb-1 ; j++ ) {
                        gap[j]=1 ;
                        ip[j]=ip[j-1]+1 ;
                        pq[j]=pq[j-1]+(unsigned long)Primes[ip[j]].diff ;
                    }
                    gap[ncomb-1]=ip[ncomb-1]-ip[ncomb-2] ;
  gap[ncomb]=ip[ncomb]-ip[ncomb-1] ;
                    a=pq[0] ;
                    for( j=1 ; j<ncomb ; j++ ) a*=pq[j] ;
//                  j=0 ;
                }
                else return false ;
            }
            else {
//              if( j!=1 ) j-- ;
            }
        }

CHANGE:
#if 0
        a=pq[0] ;
        for( j=1 ; j<ncomb ; j++ ) a*=pq[j] ;
//      cout << endl << "a " << a ;
        if( a<mina ) {
cout << endl << mina  << "a<mina " << a;
//Error(AAAA) ;
            continue ;
        }
        else if( a>maxa ) {
//cout << endl << "a>maxa " << a ;
//for( j=0 ; j<n ; j++ ) cout << " " << ip[j] ;
            return false ;
            ip[0]=base+1 ;
            for( j=1 ; j<ncomb ; j++ ) ip[j]=ip[j-1]+1 ;
            if( ip[n-1]>=ment ) return false ;
            pq[0]=pbase+(unsigned long)Primes[ip[0]].diff ;
            pbase=pq[0] ;
            for( j=1 ; j<n ; j++ ) {
                pq[j]=pq[j-1]+(unsigned long)Primes[ip[j]].diff ;
            }
            base++ ;
//          continue ;
            goto CHANGE ;
        }
        else break ;
#endif
//      break ;
//  }

    for( j=0 ; j<ncomb ; j++ ) {
//      r=sqrt(N,pq[j]) ;
        r=isMul(Primes[ip[j]].Root,invmodp(BS[j]=a/pq[j],pq[j]),pq[j]) ;
        if( r>(pq[j]>>1) ) {
            r=pq[j]-r ;
        }
//      B[j]=r ;
        BS[j]*=r ;
    }

//cout << endl << " B " ;
//  for( j=0 ; j<n ; j++ ) {
//      B[j]*=a/pq[j] ;
//cout << B[j] << " " ;
//  }
#if 0
cout << endl ;
for( j=0 ; j<ncomb ; j++ ) {
cout << BS[j] << " " ;
    if( ((BS[j]*BS[j]-N)%pq[j])!=0 ) Error(AAAAA) ;
    for( int k=0 ; k<ncomb ; k++ ) {
        if( k==j ) continue ;
        if( (BS[j]%pq[k])!=0 ) Error(BBBBB) ;

    }
}
#endif
#if 0
    for( k=j=0 ; j<ncomb ; j++ ) {
        k+=(pq[j]==wpq[j] ? 0:1) ;
    }
    if( k<COMB_DIFF ) {
        cout << endl << endl ;
        for( k=j=0 ; j<ncomb ; j++ ) {
            cout << "( " << pq[j] << "," << wpq[j] << ")" ;
        }
        cout << " " << k ;
//      Error(10101) ;
    }
#endif
#if 0
    cout << endl << endl ;
    for( j=0 ; j<ncomb ; j++ ) {
        cout << "\t" << ip[j] ;
    }
    cout << endl  ;
    for( j=0 ; j<ncomb ; j++ ) {
        cout << "\t" << gap[j+1] ;
    }
#endif
    return true ;
}

#else
////////////////////////////////////////////
//
//  Search coefficient of a qudaratic polynomial
//
//
bool select_a(Mint&a,Mint& mina,Mint& maxa,Mint& N,unsigned long ncomb,unsigned long ment)
{
    unsigned long r ;
    int j ;
    int k ;
    int l ;
#if 1
    unsigned long wpq[MAX_COMB] ;
    memcpy(wpq,pq,ncomb*sizeof(unsigned long)) ;
#endif

//  for(;;) {
#if 0
        for( j=1 ; j<ncomb ; j++ ) {
//          if( ip[j-1]<ip[j]-1 ) {
            if( gap[j]>2 ) {
                ip[j-1]++ ;
                a/=pq[j-1] ;
                pq[j-1]+=(unsigned long)Primes[ip[j-1]].diff ;
                ip[j-1]++ ;
                pq[j-1]+=(unsigned long)Primes[ip[j-1]].diff ;
                a*=pq[j-1] ;
                gap[j]-=2  ;
                gap[j-1]+=2 ;
//              goto CHANGE ;
            }
        }
#endif
        for( k=0 ; k<COMB_DIFF ; k++ ) {
#if 0
            for( j=1 ; j<ncomb ; j++ ) {
#endif
            for( j=1 ; j<ncomb-1 ; j++ ) {
//              if( ip[j-1]<ip[j]-1 ) {
                if( gap[j]!=1 ) {
                    a/=pq[j-1] ;
                    ip[j-1]++ ;
                    pq[j-1]+=(unsigned long)Primes[ip[j-1]].diff ;
                    a*=pq[j-1] ;
                    gap[j]-- ;
                    gap[j-1]++ ;
#if 0
cout << endl << endl << "inc " << j ;
#endif
#if 0
if( a>maxa ) {
    j=ncomb+1 ;
}
#endif
                    break ;
                }
            }
#if 1
            if( j==ncomb-1 ) {
                if( gap[j]>2 ) {
                    a/=pq[j-1] ;
                    ip[j-1]++ ;
                    pq[j-1]+=(unsigned long)Primes[ip[j-1]].diff ;
                    ip[j-1]++ ;
                    pq[j-1]+=(unsigned long)Primes[ip[j-1]].diff ;
                    a*=pq[j-1] ;
                    gap[j]-=2 ;
                    gap[j-1]+=2 ;

                }
                else j=ncomb ;
            }
#endif
            if( j>=ncomb ) {
                if( ip[ncomb-1]<ment-1 ) {
                    for( l=0 ; l<nmachine ; l++ ) {     // Ul
                        ip[ncomb-1]++ ;
                        pq[ncomb-1]+=(unsigned long)Primes[ip[ncomb-1]].diff ;
                        ip[ncomb-1]++ ;
                        pq[ncomb-1]+=(unsigned long)Primes[ip[ncomb-1]].diff ;
                    }
                    ip[0]=base ;
                    pq[0]=pbase ;
                    for( j=1 ; j<ncomb-1 ; j++ ) {
                        gap[j]=1 ;
                        ip[j]=ip[j-1]+1 ;
                        pq[j]=pq[j-1]+(unsigned long)Primes[ip[j]].diff ;
                    }
                    gap[ncomb-1]=ip[ncomb-1]-ip[ncomb-2] ;
                    a=pq[0] ;
                    for( j=1 ; j<ncomb ; j++ ) a*=pq[j] ;
//                  j=0 ;
                }
                else return false ;
            }
            else {
//              if( j!=1 ) j-- ;
            }
        }

CHANGE:
#if 0
        a=pq[0] ;
        for( j=1 ; j<ncomb ; j++ ) a*=pq[j] ;
//      cout << endl << "a " << a ;
        if( a<mina ) {
cout << endl << mina  << "a<mina " << a;
//Error(AAAA) ;
            continue ;
        }
        else if( a>maxa ) {
//cout << endl << "a>maxa " << a ;
//for( j=0 ; j<n ; j++ ) cout << " " << ip[j] ;
            return false ;
            ip[0]=base+1 ;
            for( j=1 ; j<ncomb ; j++ ) ip[j]=ip[j-1]+1 ;
            if( ip[n-1]>=ment ) return false ;
            pq[0]=pbase+(unsigned long)Primes[ip[0]].diff ;
            pbase=pq[0] ;
            for( j=1 ; j<n ; j++ ) {
                pq[j]=pq[j-1]+(unsigned long)Primes[ip[j]].diff ;
            }
            base++ ;
//          continue ;
            goto CHANGE ;
        }
        else break ;
#endif
//      break ;
//  }

    for( j=0 ; j<ncomb ; j++ ) {
//      r=sqrt(N,pq[j]) ;
        r=isMul(Primes[ip[j]].Root,invmodp(BS[j]=a/pq[j],pq[j]),pq[j]) ;
        if( r>(pq[j]>>1) ) {
            r=pq[j]-r ;
        }
//      B[j]=r ;
        BS[j]*=r ;
    }

//cout << endl << " B " ;
//  for( j=0 ; j<n ; j++ ) {
//      B[j]*=a/pq[j] ;
//cout << B[j] << " " ;
//  }
#if 0
cout << endl ;
for( j=0 ; j<ncomb ; j++ ) {
cout << BS[j] << " " ;
    if( ((BS[j]*BS[j]-N)%pq[j])!=0 ) Error(AAAAA) ;
    for( int k=0 ; k<ncomb ; k++ ) {
        if( k==j ) continue ;
        if( (BS[j]%pq[k])!=0 ) Error(BBBBB) ;

    }
}
#endif
#if 0
    for( k=j=0 ; j<ncomb ; j++ ) {
        k+=(pq[j]==wpq[j] ? 0:1) ;
    }
    if( k<COMB_DIFF ) {
        cout << endl << endl ;
        for( k=j=0 ; j<ncomb ; j++ ) {
            cout << "( " << pq[j] << "," << wpq[j] << ")" ;
        }
        cout << " " << k ;
//      Error(10101) ;
    }
#endif
    return true ;
}

#endif

#if 0
void  nextB(unsigned long n,Mint& b)
{
    unsigned long nu=n<<1 ;
    unsigned long nu2 ;
    unsigned long ind ;
    nu=BitScanForward(&nu,1) ;
    nu2=1<<nu ;
    ind=n/nu2 ;
//  cout << endl << "n nu n/ny " << n << " " << nu << " " << double(n)/nu << " " << n%nu ;
    b=BS[nu-1] ;
    if( (n%nu2)!=0 ) ind++ ;
    if( ind&1 ) ~b ;
}

#endif


